/**
 * 
 */
package edu.unika.aifb.owltools.metrics;

import java.util.Set;

import org.semanticweb.kaon2.api.KAON2Exception;
import org.semanticweb.kaon2.api.Ontology;
import org.semanticweb.kaon2.api.Request;
import org.semanticweb.kaon2.api.owl.elements.Description;
import org.semanticweb.kaon2.api.owl.elements.OWLClass;

/**
 * The sum of the caridinalities of all subsumptionpaths in the ontology.
 * Measure M1, Gangemi et al. TR 9/05
 */
public class AbsoluteDepth extends Metric {
	public AbsoluteDepth(Ontology ontologyToBeMeasured, String URI,	Ontology contextOfMeasurement) { super(ontologyToBeMeasured, URI, contextOfMeasurement); }
	public String metricClassName() { return "Absolute depth"; }
	public String metricClassURI() { return "http://owltools.ontoware.org/metrics/absolutedepth"; }
	public String metricClassDescription() { return "The sum of lengths of all asserted subsumption paths in the class hierarchy."; }

	protected boolean measure() {
		try {
			Request<OWLClass> request = ontology.createEntityRequest(OWLClass.class);
			Set<OWLClass> classes = request.get();
			for (OWLClass c : classes) {
				Set<Description> supers = c.getSuperDescriptions(ontology);
				for (Description sup : supers) {
					if (OWLClass.class.isInstance(sup))
						result++;
				}
			}			
		} catch (KAON2Exception k) {
			error = k.toString();
			return false;
		}
		return true;
	}

}
