package edu.unika.aifb.owltools.metrics;

import java.util.Set;

import org.semanticweb.kaon2.api.*;
import org.semanticweb.kaon2.api.owl.elements.*;

/**
 * Calculates the longest asserted subsumption path in the class hierarchy.
 * Measure M3, Gangemi et al. TR 9/05
 */
public class MaximalDepth extends Metric {
	public MaximalDepth(Ontology ontologyToBeMeasured, String URI, Ontology contextOfMeasurement) { super(ontologyToBeMeasured, URI, contextOfMeasurement); }
	public String metricClassName() { return "Maximal depth"; }
	public String metricClassURI() { return "http://owltools.ontoware.org/metrics/maximaldepth"; }
	public String metricClassDescription() { return "The longest asserted subsumption path in the class hierarchy."; }

	protected boolean measure() {
		try {
			Request<OWLClass> request = ontology.createEntityRequest(OWLClass.class);
			Set<OWLClass> classes = request.get();
			for (OWLClass c : classes) {
				Set<Description> supers = c.getSuperDescriptions(ontology);
				for (Description sup : supers) {
					if (OWLClass.class.isInstance(sup))
						result++;
				}
			}			
		} catch (KAON2Exception k) {
			error = k.toString();
			return false;
		}
		return true;
	}

}
