package edu.unika.aifb.owltools.serializer;

import org.semanticweb.kaon2.api.logic.*;

import edu.unika.aifb.owltools.utils.Printer;

/**
 * Serializes a program into a RuleML file
 */
public class RuleMLSerializer extends ProgramSerializer {
	
	public RuleMLSerializer(Printer printer) { super(printer); }

	/**
	 * How much intentation?
	 */
	private int m_intendation;
	
	/**
	 * Prolog of a RuleML file
	 */
	public void start() {
		m_printer.putline("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
		m_printer.putline("<RuleML xmlns=\"http://www.ruleml.org/0.9/xsd\"");
		m_printer.putline("        xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"");
		m_printer.putline("        xsi:schemaLocation=\"http://www.ruleml.org/0.9/xsd http://www.ruleml.org/0.9/xsd/datalog.xsd\">");
	}
	
	/**
	 * Epilog of a RuleML file
	 */
	public void end() {
		m_printer.putline("</RuleML>");
	}
	
	/**
	 * How to serialize a rule
	 */
	public Object visit(Rule rule) {
		
		int headLength = rule.getHeadLength();
		int bodyLength = rule.getBodyLength();

		if (headLength>1)
			m_DDL = true;
		
		m_intendation = 4;
		if (bodyLength > 0) {
			m_printer.putline("  <Implies>");
			m_printer.putline("    <head>");
		}
		if (headLength > 1) {
			m_printer.putline("     <Or>");
		}
		for (int i=0;i<headLength;i++)
			rule.getHeadLiteral(i).accept(this);
		if (headLength > 1) {
			m_printer.putline("      </Or>");
		}
		if (bodyLength > 0) {
			m_printer.putline("    </head>");
			m_printer.putline("    <body>");
		}
		if (bodyLength > 1) {
	        m_printer.putline("     <And>");
		}
        for (int j=0;j<bodyLength;j++)
            rule.getBodyLiteral(j).accept(this);
		if (bodyLength > 1) {
	        m_printer.putline("     </And>");
		}
		if (bodyLength > 0) {
			m_printer.putline("    </body>");
			m_printer.putline("  </Implies>");
		}
	
		return null;
	}
	
	/**
	 * Make the proper intendation
	 * @param relative_intendation how far is it intended already
	 */	
	protected void intend(int relative_intendation) {
		int intendation = (m_intendation+relative_intendation)*2;
		for (int i=0; i<intendation; ++i)
			m_printer.put(" ");
	}
	
	/**
	 * Serializes a literal
	 */
	public Object visit(Literal literal) {
        if (!literal.isPositive())
        	m_printer.error("Flawed serialization. Negative literal won't be expressed correctly.");
        intend(0);
        m_printer.putline("<Atom>");
        intend(1);
        m_printer.put("<op><Rel>");
        m_printer.put(processName(literal.getPredicate().toString()));
        m_printer.putline("</Rel></op>");
        for (int i=0;i<literal.getArity();i++) {
            Variable variable=literal.getArgumentVariable(i);
            if (variable==null) {
            	Constant constant = literal.getArgumentConstant(i);
            	if (constant==null) {
            		m_printer.error("Serialization flawed at RuleMLSerializer.visit(Literal)");
            	}
                intend(1);
                m_printer.put("<Ind>");
                m_printer.put(processName(constant.toString()));
                m_printer.putline("</Ind>");
                continue;
            }
            intend(1);
            m_printer.put("<Var>");
            m_printer.put(processName(variable.getVariableName()));
            m_printer.putline("</Var>");
        }
        intend(0);
        m_printer.putline("</Atom>");		
        
        return null;
	}
	
}
