package edu.unika.aifb.owltools.server;

import java.util.*;

import javax.servlet.http.*;

import org.semanticweb.kaon2.api.*;
import org.semanticweb.kaon2.api.formatting.*;
import org.semanticweb.kaon2.api.reasoner.*;
import org.semanticweb.kaon2.api.owl.elements.*;

import edu.unika.aifb.owltools.*;

/**
 * Implements a very simple HttpServlet with a very
 * strange interface. Extremely Beta. Peter has a running
 * implementation of the SPARQL protocol, this basically
 * should be replecaed by Peter's implementation
 */
public class KaonServer extends HttpServlet {
	static final long serialVersionUID = 237432874;

	Ontology ontology;
	KAON2Connection m_connection = null;
	
	public void init() {
		if (m_connection == null)
			try {
				m_connection = KAON2Manager.newConnection();
				ontology = createOntology("http://aifb.uka.de/ontonto/x/test");
			} catch (KAON2Exception k) {
				System.err.println(k.toString());
			}
	}
	
	public void destroy() {
		super.destroy();
		try {
		if (m_connection != null)
			m_connection.close(); // GANZ HEFTIGES TODO memory leak?
		} catch (KAON2Exception k) {
			System.err.println(k.toString());
		} finally {
			m_connection = null;		
		}
	}

	protected void doGet(HttpServletRequest req, HttpServletResponse resp) {
		try {
			if (req.getPathInfo()==null)
				resp.getWriter().println("empty");
			else if (req.getPathInfo().equalsIgnoreCase("/test"))
				resp.getWriter().println("beta2");
			else if (req.getPathInfo().equalsIgnoreCase("/name"))
				resp.getWriter().println(ontology.getOntologyURI());
			else if (req.getPathInfo().equalsIgnoreCase("/count"))
				resp.getWriter().println(api.count(ontology));
			else if (req.getPathInfo().equalsIgnoreCase("/load")) {
				System.err.println("Loading...");
				ontology = loadOntology("aifbswrc.owl");
				System.err.println("Loaded.");
			} else if (req.getPathInfo().equalsIgnoreCase("/classes")) {
				List<String> entities = api.dumpEntities(ontology, OWLClass.class);
				for (String e : entities) resp.getWriter().println(e);
			} else if (req.getPathInfo().startsWith("/classify")) {
				System.err.println("enter");
				Ontology tboxontology = loadOntology("author.owl");
				System.err.println("loaded author");
				Ontology tempontology = null;
				String title = req.getParameter("title");
				System.err.println("loading " + title);
				//tempontology = loadOntology("http://localhost/mediawiki-1.5.7/index.php/Special:ExportRDF/" + title);
				tempontology = loadOntology(title + ".rdf");
				System.err.print("Nr of axioms tbox: ");
				System.err.println(api.count(tboxontology));
				System.err.print("Nr of axioms loaded: ");
				System.err.println(api.count(tempontology));
				System.err.println("Now merging");
				api.merge(tboxontology, tempontology, ontology);
				System.err.print("Nr of axioms: ");
				System.err.println(api.count(ontology));
				Reasoner reasoner = ontology.createReasoner();
				Namespaces ns = Namespaces.INSTANCE;
				//ns.registerPrefix("wpcat", "http://nodix.de/smw/mediawiki-2D1.5.7/index.php/_Category-3A");
				ns.registerPrefix("wp", "http://nodix.de/smw/mediawiki-2D1.5.7/index.php/");
				Request<OWLClass> request = ontology.createEntityRequest(OWLClass.class);
				Set<OWLClass> entities = request.get();
				List<String> results = new LinkedList<String>();
				for (OWLClass e : entities) {
					StringBuffer b = new StringBuffer();
					e.toString(b, ns);
					String c = b.toString();
					System.err.println(c);
					try {
						Query query = reasoner.createQuery(ns, "SELECT ?N WHERE { wp:_" + title + " rdf:type " + c + " }");
						query.open();
						if (query.getNumberOfTuples()>0) {
							//resp.getWriter().println(e);
							results.add(c.substring(c.indexOf("-3A")+3, c.length()));
						}
						query.close();
					} catch (KAON2Exception k) {
					}
				}
				for (String result : results)
					resp.getWriter().println(result);
				//resp.getWriter().println(api.count(ontology));					
			} else if (req.getPathInfo().equalsIgnoreCase("/query")) {
				Reasoner reasoner = ontology.createReasoner();
				Namespaces ns = Namespaces.INSTANCE;
				ns.registerPrefix("swrc", "http://swrc.ontoware.org/ontology#");
				Query query = reasoner.createQuery(ns, "SELECT ?Name WHERE { ?x swrc:author ?y . ?y swrc:name ?Name }");
				query.open();
				resp.getWriter().println(query.getNumberOfTuples());
				while (!query.afterLast()) {
					resp.getWriter().println(query.tupleBuffer()[0].toString());
					query.next();
				}
			} else if (req.getPathInfo().equalsIgnoreCase("/ontology")) {
				ontology.saveOntology(OntologyFileFormat.OWL_RDF, resp.getWriter(), "UTF-8");
			} else
				resp.getWriter().println(req.getPathInfo());
		} catch (Exception e) {
			System.err.println(e.toString());
		}
		System.err.println(req.toString());
		System.err.println(resp.toString());
	}

	protected void doPost(HttpServletRequest req, HttpServletResponse resp) {
		try {
			if (req.getPathInfo()==null)
				resp.getWriter().println("empty");
			else if (req.getPathInfo().equalsIgnoreCase("/name"))
				resp.getWriter().println(ontology.getOntologyURI());
			else if (req.getPathInfo().equalsIgnoreCase("/count"))
				resp.getWriter().println(api.count(ontology));
			else if (req.getPathInfo().equalsIgnoreCase("/load"))
				ontology = loadOntology("proton.owl");
			else
				resp.getWriter().println(req.getPathInfo());
		} catch (Exception e) {
			System.err.println(e.toString());
		}
	}
	
	protected Ontology loadOntology(String name) throws KAON2Exception, InterruptedException {
		OntologyResolver resolver = m_connection.getOntologyResolver();
		if (DefaultOntologyResolver.class.isInstance(resolver)) {
			DefaultOntologyResolver res = (DefaultOntologyResolver)resolver;
	        if (!name.startsWith("http:"))
		        if (!name.startsWith("file:"))
		        	res.registerReplacement(name, "file:" + name);
	        resolver = res;
		}
        m_connection.setOntologyResolver(resolver);
		
       	Ontology ontology = m_connection.openOntology(name, new HashMap<String,Object>());
       	return ontology;
	}
	
	Ontology createOntology(String name) throws KAON2Exception {
		OntologyResolver resolver = m_connection.getOntologyResolver();
		if (DefaultOntologyResolver.class.isInstance(resolver)) {
			DefaultOntologyResolver res = (DefaultOntologyResolver)resolver;
			res.registerReplacement(name, "file:temp.owl");
	        resolver = res;
		}
        m_connection.setOntologyResolver(resolver);

        Ontology ontology = m_connection.createOntology(name, new HashMap<String, Object>());
		return ontology;
	}
}
