/**
 * 
 */
package edu.unika.aifb.owltools.shell;

import java.io.*;
import java.util.*;

import org.semanticweb.kaon2.api.*;

import edu.unika.aifb.owltools.*;
import edu.unika.aifb.owltools.tools.KAON2Tool;

/**
 * Implements an experimental shell to work with OWL files (beta)
 *
 */
public class Shell extends KAON2Tool {
	/**
	 * name of the owl tool
	 */
	public String name() { return "shell"; }
	
	/**
	 * if the shell is run from an extra window, this is here
	 */
	ShellWindow window;
	
	/**
	 * Simple means that the shell is run on a normal command
	 * line, not the extra shell
	 */
	boolean simple = false;
	
	/**
	 * End of line string
	 */
	String eol = System.getProperty("line.separator");
	
	/**
	 * mapping of ontology local names to their URIs
	 */
	Map<String, String> ontologies;
	
	/**
	 * mapping of commands to their help texts
	 */
	Map<String, String> commands;
	
	/**
	 * mapping of axioms to their help texts
	 */
	Map<String, String> axioms;
	
	/**
	 * initialises all usual commands and their help texts
	 *
	 */
	void initcommands() {
		commands = new HashMap<String, String>();
		if (!simple) commands.put("cls",
				"cls clears the screen. The command does not work on the simple shell.");
		commands.put("version",
				"version returns the version of the KAON2 OWL tools this shell is based on.");
		commands.put("help",
				"help provides further information on all available commands." + eol +
				"Type 'help <command>' to get these information. Type 'help all' for all helps.");
		commands.put("exit",
				"exit ends the shell and frees all resources. The ontologies are not changed.");
		commands.put("load",
				"load <uri> will load an ontology from the given uri and becomes the" + eol +
				"default ontology. You can also 'load <uri> from <location>', which loads the" + eol +
				"ontology named uri directly from location (it is the same as registering" + eol +
				"the location for the ontology with 'resolve' first and load then). You can also" + eol +
				"set the shortcut during loading with 'load <uri> as <name>' - or combine both.");
		commands.put("count",
				"count returns the number of axioms in the default ontology.");
		commands.put("dump",
				"dump prints the axioms of the default ontology.");
		commands.put("namespaces",
				"namespaces return a table of all registered namespaces and the uris the" + eol +
				"namespaces expand to.");
		commands.put("define",
				"define <ns> <uri> registers the uri for the namespace ns, so you can use it as a" + eol +
				"shortcut. 'define default <uri>' sets the default namespace.");
		commands.put("undefine",
				"unregisters a previously defined namespace.");
		commands.put("resolver",
				"resolver returns a table of all uri that have registered a adresses they" + eol +
				"resolve to.");
		commands.put("resolve",
				"resolve <uri> <location> registers an uri and the adress to resolve to. You can" + eol +
				"also register to local files, like 'resolve http://www.example.com/# ex.owl'");
		commands.put("ontologies",
				"ontologies shows all the opened ontologies and, if you have defined any, their" + eol +
				"respective shortcuts.");
		commands.put("set",
				"set <name> <uri> creates a shortcut to an ontology.");
		commands.put("add",
				"add <axiom> will put the given axiom into the default ontology. <axiom> has" + eol +
				"to be written in KAON2 syntax.");
		commands.put("del",
				"del <axiom> removes the given axiom. <axiom> has to be written in KAON2 syntax.");
 	}

	/**
	 * initialises axioms with their help strings (incomplete) TODO
	 *
	 */
	void initaxioms() {
		axioms = new HashMap<String, String>();
		commands.put("classMember",
				"Two arguments: the first one is the class, the second the individual. Asserts" + eol +
				"that the individual is a member of the given class.");
 	}

	/**
	 * processes command line arguments
	 */
	public boolean processArguments(String[] args) {
		int argc = args.length;
		if (argc == 0) {
			return true;
		}
		String s = args[0].toLowerCase();
		if (argc == 1) {
			if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) || (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) || (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help"))) {
				printUsage();
				return false;
			}
			if ((s.equalsIgnoreCase("shell"))) {
				return true;
			}
		}

		for (int i = 0; i < argc; ++i) {
			boolean unknownArg = true;
			
			String arg = args[i];

			if ((arg.equalsIgnoreCase("-?")) || (arg.equalsIgnoreCase("?")) || (arg.equalsIgnoreCase("-h")) || (arg.equalsIgnoreCase("--help")) || (arg.equalsIgnoreCase("-help")) || (arg.equalsIgnoreCase("help"))) {
				printUsage();
				return false;
			}
			
			if (arg.startsWith("-")) {
				// Simple
				if (arg.equalsIgnoreCase("-simple")) {
					simple = true;
					unknownArg = false;
				}
			} else {
				if (arg.equalsIgnoreCase("shell")) {
					unknownArg = false;
				} else {
				}
			}
			
			if (unknownArg) { printUsageError(); return false; }
		}
		
		return true;
	}

	/**
	 * to be started from the command line as an extra tool
	 * @param args command line arguments
	 */
	public static void main(String[] args) {
		Shell shell = new Shell();
		if (!shell.processArguments(args)) return;
		shell.compile();
		shell.printer.close();
	}

	/**
	 * prints the command line parameters and a short info
	 */
	public static void printUsage() {
		System.out.println("Usage: owl shell");
		System.out.println("");
		System.out.println("Opens a shell where you can quickly edit and explore ontologies");
		System.out.println("Use the parameter -simple to start it in your local shell instead of a");
		System.out.println("new window, but you may lack some features then.");
	}	

	/**
	 * starts the shell after having processed all arguments
	 *
	 */
	public void compile() {
		System.out.println("OWL shell " + OWLTools.version() + ".");
		initcommands();
		initaxioms();
		try {
			m_connection = KAON2Manager.newConnection();
		} catch (KAON2Exception k) {
			System.err.println("Could not create a KAON2 connection. Shell stopped.");
			return;			
		}
		OntologyResolver resolver = m_connection.getOntologyResolver();
		String defaultname = "http://www.aifb.uni-karlsruhe.de/WBS/dvr/owltools/shelldefault#";
		if (DefaultOntologyResolver.class.isInstance(resolver)) {
			DefaultOntologyResolver res = (DefaultOntologyResolver)resolver;
			res.registerReplacement(defaultname, "file:default.owl");
	        resolver = res;
		}
        m_connection.setOntologyResolver(resolver);

        try {
            m_connection.createOntology(defaultname, new HashMap<String, Object>());
            ontologies = new HashMap<String, String>();
            ontologies.put("default", defaultname);
            ontologies.put("owlshell", defaultname);
        } catch (KAON2Exception k) {
        	System.err.println("Could not create default ontology. Please try to load one.");        	
        }

		if (simple) {
			System.out.println("Type 'help' for further instructions.");
			BufferedReader in = new BufferedReader(new InputStreamReader(System.in ));
			boolean cont = true;
			while (cont) {
				System.out.print(">>> ");
				try {
					String cmd = in.readLine();
					if (cmd.trim().equals("exit"))
						cont = false;
					else
						System.out.println(eval(cmd));
				} catch (IOException io) {
					System.out.println("Unhandled IOException: " + io.toString());
				}
			}
		} else {
		    window = new ShellWindow(this); 
			window.setVisible(true);
			window.console.append("OWL shell " + OWLTools.version() + ".");
			window.newline();
			window.console.append("Type 'help' for further instructions.");		
			window.nextCommand();		
		}
	}
	
	/**
	 * returns an ontology by its local name
	 * @param name local name, as used in the shell
	 * @return the referenced ontology
	 * @throws KAON2Exception
	 */
	Ontology getOntology(String name) throws KAON2Exception {
		String uri = name;
		if (ontologies.containsKey(uri))
			uri = ontologies.get(uri);
		return m_connection.getOntology(uri);
	}
	
	/**
	 * evaluates the user input
	 * @param cmd user input
	 * @return the output on the command line
	 */
	public String eval(String cmd) {
		String[] tokens = cmd.trim().split("(\\s)+");
		if (tokens[0].equals("help"))		return help(tokens);
		if (tokens[0].equals("version"))	return OWLTools.version();
		if (tokens[0].equals("load"))		return load(tokens);
		if (tokens[0].equals("count"))		return count(tokens);
		if (tokens[0].equals("dump"))		return dump(tokens);
		if (tokens[0].equals("namespaces"))	return namespaces(tokens);
		if (tokens[0].equals("define"))		return define(tokens);
		if (tokens[0].equals("undefine"))	return undefine(tokens);
		if (tokens[0].equals("resolver"))	return resolver(tokens);
		if (tokens[0].equals("resolve"))	return resolve(tokens);
		if (tokens[0].equals("ontologies"))	return ontologies(tokens);
		if (tokens[0].equals("set"))		return set(tokens);
		if (tokens[0].equals("add"))		return add(tokens);
		if (tokens[0].equals("del"))		return del(tokens);
		return "Unknown command: " + cmd;		
	}
	
	/**
	 * returns help on commands
	 * @param cmd which help is asked?
	 * @return help text
	 */
	String help(String cmd) {
		if (cmd.equals("all")) {
			Set<String> commandset = commands.keySet();
			String rv = "";
			for (String command : commandset)
				rv += help(command) + eol;
			return rv + "The OWL shell is part of the OWL tools.";
		}
		if (commands.containsKey(cmd))
			return commands.get(cmd);
		return "Sorry, no help on " + cmd + " available.";
	}
	
	/**
	 * returns help on a set of commands
	 * @param tokens the set of commands
	 * @return the help strings
	 */
	String help(String[] tokens) {
		if (tokens.length == 1) {
			String rv = "The following commands are available: ";
			Set<String> commandset = commands.keySet();
			int counter = 0;
			for (String command : commandset) {
				rv += command + (++counter==commandset.size()?"":", ") + (counter%10==5?eol:"");
			}
			rv += "." + eol + "Type 'help <command>' for further information.";
			return rv;
		}
		String rv = "";
		for (String token : tokens)
			rv += (token==tokens[0]?"":((rv.length()==0?"":eol)+help(token)));
		return rv;
	}
	
	/**
	 * loads an ontology
	 * @param tokens shell command parameteres
	 * @return shell output
	 */	
	String load(String[] tokens) {
		if ((tokens.length!=2)&&(tokens.length!=4)&&(tokens.length!=6)) return help("load");
		try {
			String from = null;
			String as = "default";
			if ((tokens.length == 4) && (tokens[2].equals("from")))
				from = tokens[3];
			if ((tokens.length == 6) && (tokens[2].equals("from")))
				from = tokens[3];
			if ((tokens.length == 6) && (tokens[4].equals("from")))
				from = tokens[5];
			if ((tokens.length == 4) && (tokens[2].equals("as")))
				as = tokens[3];
			if ((tokens.length == 6) && (tokens[2].equals("as")))
				as = tokens[3];
			if ((tokens.length == 6) && (tokens[4].equals("as")))
				as = tokens[5];
			
			if (from != null) {	
				String[] str = new String[3];
				str[0] = "resolve";
				str[1] = tokens[1];
				str[2] = from;
				resolve(str);
				m_connection.openOntology(tokens[1], new HashMap<String,Object>());
				ontologies.put("default", tokens[1]);
			}
			
			m_connection.openOntology(tokens[1], new HashMap<String,Object>());
			ontologies.put(as, tokens[1]);

			return api.count(getOntology(as)) + " axioms loaded.";
		} catch (InterruptedException i) {
			return i.toString();
		} catch (KAON2Exception k) {
			return k.toString();
		}
	}

	/**
	 * uses the count owl tool
	 * @param tokens shell command tokens
	 * @return shell output
	 */
	String count(String[] tokens) {
		if ((tokens.length!=1)&&(tokens.length!=2)) return help("count");
		try {
			String ont = "default";
			if (tokens.length == 2)
				ont = tokens[1];
			return api.count(getOntology(ont)) + "";
		} catch (KAON2Exception k) {
			return k.toString();
		}
	}

	/**
	 * uses the dump owl tool
	 * @param tokens shell command tokens
	 * @return shell output
	 */
	String dump(String[] tokens) {
		if ((tokens.length!=1)&&(tokens.length!=2)) return help("dump");
		try {
			String ont = "default";
			if (tokens.length == 2)
				ont = tokens[1];
			List<String> onto = api.dump(getOntology(ont));
			String rv = "";
			if (onto.size() == 0) rv = "Empty ontology.";
			for (String axiom : onto) {
				rv += (rv.length()>0?eol:"") + axiom;
			}
			return rv;
		} catch (KAON2Exception k) {
			return k.toString();
		}
	}

	/**
	 * returns the namespaces that are safed locally
	 * @param tokens shell command tokens
	 * @return shell output
	 */
	String namespaces(String[] tokens) {
		if (tokens.length!=1) return help("namespaces");
		String rv = "";
		String defaultns = Namespaces.INSTANCE.getDefaultNamespace();
		if (defaultns == null)
			rv += "default\t\tNone defined";
		else
			rv += "default\t\t" + defaultns;				
		Iterator<String> iter = Namespaces.INSTANCE.prefixes();
		while (iter.hasNext()) {
			String ns = iter.next();
			rv += eol + ns + (ns.length()<8?"\t\t":"\t") + Namespaces.INSTANCE.getNamespaceForPrefix(ns);
		}
		return rv;
	}
	
	/**
	 * defines a new namespace
	 * @param tokens shell command tokens
	 * @return shell output
	 */
	String define(String[] tokens) {
		if (tokens.length!=3) return help("define");
		String rv = "";
		if (tokens[1].equals("default")) {
			Namespaces.INSTANCE.setDefaultNamespace(tokens[2]);
			rv = "Registered default prefix as " + tokens[2];
		} else {
			if (Namespaces.INSTANCE.getNamespaceForPrefix(tokens[1])==null) {
				rv = "Registered prefix " + tokens[1] + ".";
			} else {
				rv = "Changed registration for prefix " + tokens[1] + ".";
			}
		}
		Namespaces.INSTANCE.registerPrefix(tokens[1], tokens[2]);
		return rv;
	}

	/**
	 * forgets about a namespace
	 * @param tokens shell command tokens
	 * @return shell output
	 */
	String undefine(String[] tokens) {
		if (tokens.length!=2) return help("undefine");
		String rv = "";
		if (Namespaces.INSTANCE.getNamespaceForPrefix(tokens[1])==null)
			rv = "The prefix " + tokens[1] + " is not registered.";
		else {
			Namespaces.INSTANCE.unregisterPrefix(tokens[1]);
			rv = "Unregistered prefix " + tokens[1] + ".";
		}
		return rv;
	}

	/**
	 * access to the Ontology resolver
	 * @param tokens shell command parameters
	 * @return shell output
	 */
	String resolver(String[] tokens) {
		if (tokens.length!=1) return help("resolver");
		String rv = "";
		if (DefaultOntologyResolver.class.isInstance(m_connection.getOntologyResolver())) {
			DefaultOntologyResolver res = (DefaultOntologyResolver)m_connection.getOntologyResolver();
			Iterator<String> iter = res.registeredOntologyURIs();
			while (iter.hasNext()) {
				String uri = iter.next();
				rv += uri + "\t" + res.getReplacement(uri) + (iter.hasNext()?eol:"");
			}
		} else 
			rv = "Unknown ontologyresolver registered. Can't answer the request, sorry.";
		return rv;
	}
	
	/**
	 * resolves an ontology
	 * @param tokens shell command parameters
	 * @return shell output
	 */
	String resolve(String[] tokens) {
		if (tokens.length!=3) return help("resolve");
		String rv = "";
		if (DefaultOntologyResolver.class.isInstance(m_connection.getOntologyResolver())) {
			DefaultOntologyResolver res = (DefaultOntologyResolver)m_connection.getOntologyResolver();
			String physical = tokens[2];
			if (!physical.contains(":"))
				physical = "file:" + physical;				
			res.registerReplacement(tokens[1], physical);
		} else 
			rv = "Unknown ontologyresolver registered. Can't fulfil the request, sorry.";
		return rv;
	}

	/**
	 * Returns a list of all ontologies the shell knows about
	 * @param tokens shell command parameters
	 * @return shell output
	 */
	String ontologies(String[] tokens) {
		if (tokens.length!=1) return help("ontologies");
		String rv = "";
		try {
			rv += "default\t\t" + ontologies.get("default");
			Set<String> os = m_connection.getOntologyURIs();
			Set<String> namedos = ontologies.keySet();
			for (String o : namedos)
				if (!o.equals("default"))
					rv += eol + o + (o.length()<8?"\t\t":"\t") + ontologies.get(o); 
			for (String o : os)
				if (!ontologies.containsValue(o))
					rv += eol + o;
		} catch (KAON2Exception k) {
			rv = k.toString();
		}
		return rv;
	}

	/**
	 * sets a shortcut to an ontology
	 * @param tokens shell command parameters
	 * @return shell output
	 */
	String set(String[] tokens) {
		if (tokens.length!=3) return help("set");
		String rv = "";
		try {
			if (!m_connection.getOntologyURIs().contains(tokens[2]))
				rv = "The ontology you want to shortcut to is not open.";
			else {
				if (ontologies.containsKey(tokens[1]))
					rv = "Reassigned";
				else
					rv = "Assigned";
				rv += " shortcut " + tokens[1] + " to " + tokens[2];
				ontologies.put(tokens[1], tokens[2]);
			}
		} catch (KAON2Exception k) {
			rv = k.toString();
		}
		return rv;
	}
	
	/**
	 * adds an axiom to an ontology
	 * @param tokens shell command parameters
	 * @return shell output
	 */
	String add(String[] tokens) {
		String rv = "";
		String axiomstring = "";
		for (String token : tokens)
			axiomstring += " " + token;
		axiomstring = axiomstring.substring(5);
		try {
			Axiom axiom = KAON2Manager.factory().axiom(axiomstring, Namespaces.INSTANCE);
			List<OntologyChangeEvent> changes = new ArrayList<OntologyChangeEvent>();
			changes.add(new OntologyChangeEvent(axiom, OntologyChangeEvent.ChangeType.ADD));
			getOntology("default").applyChanges(changes);
			rv += "Added " + axiom.toString() + " to " + ontologies.get("default");			
		} catch (KAON2Exception k) {
			rv = k.toString();
		}
		return rv;
	}
	
	/**
	 * deletes an axiom from an ontology 
	 * @param tokens shell command parameters
	 * @return shell output
	 */
	String del(String[] tokens) {
		String rv = "";
		String axiomstring = "";
		for (String token : tokens)
			axiomstring += " " + token;
		axiomstring = axiomstring.substring(5);
		try {
			Axiom axiom = KAON2Manager.factory().axiom(axiomstring, Namespaces.INSTANCE);
			List<OntologyChangeEvent> changes = new ArrayList<OntologyChangeEvent>();
			changes.add(new OntologyChangeEvent(axiom, OntologyChangeEvent.ChangeType.REMOVE));
			getOntology("default").applyChanges(changes);
			rv += "Removed " + axiom.toString() + " from " + ontologies.get("default");			
		} catch (KAON2Exception k) {
			rv = k.toString();
		}
		return rv;
	}
}
