/**
 * 
 */
package edu.unika.aifb.owltools.shell;

import java.awt.*; 
import java.awt.event.*; 
 
/**
 * the window with the OWL shell. This is needed because the command line
 * cannot capture key events-
 */
public class ShellWindow extends Frame {
	/**
	 * UID for the window
	 */
	private static final long serialVersionUID = 7598324874938732702L;
	
	/**
	 * the textarea with the content of the console
	 */
	TextArea console;
	
	/**
	 * what does the prompt look like
	 */
	String prompt = ">>> ";
	
	/**
	 * the shell with the intelligence. does all the work
	 */
	Shell shell;

	/**
	 * constructor for the acutal shell window
	 * @param caller
	 */
	public ShellWindow(Shell caller) { 
	    super("OWL Shell"); 
	    shell = caller;
		setSize(570, 380);

		console = new TextArea();
		console.addKeyListener(new ConsoleKeyListener(this));
		add(console);
		
		addWindowListener(new WindowAdapter() { 
			public void windowClosing(WindowEvent e) { 
				setVisible(false);
				dispose(); 
//				System.exit(0); 
		    } 
		});
	}
	
	/**
	 * evaluates the entered command
	 * @param cmd user input
	 */
	public void eval(String cmd) {
		newline();
		boolean unknown = true;
		if (cmd.trim().equals("exit")) {
			unknown = false;
			setVisible(false);
			dispose(); 
		}
		if (cmd.trim().equals("cls")) {
			unknown = false;
			console.setText("");
		}
		if (unknown) console.append(shell.eval(cmd));
		nextCommand();
	}
	
	/**
	 * adds a new line to the console
	 *
	 */
	public void newline() {
		console.append(shell.eol);
	}

	/**
	 * clears the current line
	 *
	 */
	public void clearline() {
		// TODO this needs a much more efficient implementation
		String s = console.getText().substring(0, console.getText().lastIndexOf(prompt) + prompt.length());
		console.setText("");
		console.append(s);
	}
	
	/**
	 * adds a new line
	 * @param cmd
	 */
	public void setline(String cmd) {
		// TODO this needs a much more efficient implementation
		clearline();
		console.append(cmd);
	}
	
	/**
	 * this is needed for the autocompletion
	 * @param word the word to set the last word to
	 */
	public void replaceLastWord(String word) {
		String line = lastline();
		String[] tokens = line.split("\\s");
		String lasttoken = tokens[tokens.length-1];
		if (!line.endsWith(" "))
			line = line.substring(0, line.length()-lasttoken.length());
		line += word;
		setline(line.substring(4));	
	}
	
	/**
	 * a command is finished, now the next
	 *
	 */
	public void nextCommand() {
		newline();
		console.append(prompt);
	}
	
	/**
	 * locates where the line starts
	 * @return which line is it?
	 */
	public int linestart() {
		int linex = console.getText().lastIndexOf(shell.eol) + 2;
		if (linex == 1) linex = 0;
		return linex;
	}
	
	/**
	 * X-Position of the cursor
	 * @return X-Position of the cursor
	 */
	public int cursorX() {
		return console.getCaretPosition() - linestart();		
	}

	/**
	 * what is the last line of the console?
	 * @return last line of the console
	 */
	public String lastline() {
		return console.getText().substring(linestart(), console.getText().length());
	}
}
