package edu.unika.aifb.owltools;

import edu.unika.aifb.owltools.tools.*;
import edu.unika.aifb.owltools.server.Server;
import edu.unika.aifb.owltools.shell.Shell;

/**
 * OWLTools provides the interface to the command line
 */
public class OWLTools {
	private static String version = "0.28";
	private static Boolean developmentVersion = true;
	private static String yearOfRelease = "2006";

	/**
	 * Figures out which owl tool to call and does so
	 * @param args Command line arguments
	 */
	public static void main(String[] args) {
		int argc = args.length;
		if (argc == 0) {
			printUsage();
			return;
		}
		String s = args[0].toLowerCase();
		
		if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) || (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) || (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help")) || (s.equalsIgnoreCase("info"))) {
			printUsage();
			return;
		}
		if ((s.equalsIgnoreCase("-v")) || (s.equalsIgnoreCase("--version")) || (s.equalsIgnoreCase("-version")) || (s.equalsIgnoreCase("version"))) {
			printVersion();
			return;
		}
		if (s.equalsIgnoreCase("count")) {
			boolean doubleSpecified = false;
			for (String arg : args)
				if (arg.equalsIgnoreCase("-count"))
					doubleSpecified = true;
			if (!doubleSpecified) {
				String newArgs[] = new String[argc+1];
				for (int i = 0; i<argc; ++i)
					newArgs[i] = args[i];
				newArgs[argc] = "-count";
				args = newArgs;
			}
			args[0] = "dump";
			s = "dump";
		}
		if (s.equalsIgnoreCase("dump") || s.equalsIgnoreCase("print")) {
			args[0] = "dump";
			Dump dump = new Dump();
			if (!dump.processArguments(args)) return;
			dump.compile();
			dump.printer.close();
			return;
		}
		if (s.equalsIgnoreCase("syntax")) {
			args[0] = "syntax";
			Syntax syntax = new Syntax();
			if (!syntax.processArguments(args)) return;
			syntax.compile();
			syntax.printer.close();
			return;
		}
		if (s.equalsIgnoreCase("owlxml2owlrdf")
    	 || s.equalsIgnoreCase("owlxml2rdf")
    	 || s.equalsIgnoreCase("xml2rdf")
    	 || s.equalsIgnoreCase("xml2owlrdf")
		 || s.equalsIgnoreCase("x2r")
		 || s.equalsIgnoreCase("a2r")
		 || s.equalsIgnoreCase("abstractsyntax2owlrdf")
		 || s.equalsIgnoreCase("abstractsyntax2rdf")
		 || s.equalsIgnoreCase("abstract2owlrdf")
		 || s.equalsIgnoreCase("abstract2rdf")) {
			boolean doubleSpecified = false;
			for (String arg : args)
				if (arg.equalsIgnoreCase("-rdf"))
					doubleSpecified = true;
			if (!doubleSpecified) {
				String newArgs[] = new String[argc+1];
				for (int i = 0; i<argc; ++i)
					newArgs[i] = args[i];
				newArgs[argc] = "-rdf";
				args = newArgs;
			}
			args[0] = "r2x";
			s = "r2x";
		}
		if (s.equalsIgnoreCase("owlxml2abstractsyntax")
		 || s.equalsIgnoreCase("owlxml2abstract")
	  	 || s.equalsIgnoreCase("xml2abstractsyntax")
	   	 || s.equalsIgnoreCase("xml2abstract")
		 || s.equalsIgnoreCase("x2a")
		 || s.equalsIgnoreCase("r2a")
		 || s.equalsIgnoreCase("owlrdf2abstractsyntax")
		 || s.equalsIgnoreCase("owlrdf2abstract")
		 || s.equalsIgnoreCase("rdf2abstractsyntax")
		 || s.equalsIgnoreCase("rdf2abstract")) {
			boolean doubleSpecified = false;
			for (String arg : args)
				if (arg.equalsIgnoreCase("-abstract") || arg.equalsIgnoreCase("-as"))
					doubleSpecified = true;
			if (!doubleSpecified) {
				String newArgs[] = new String[argc+1];
				for (int i = 0; i<argc; ++i)
					newArgs[i] = args[i];
				newArgs[argc] = "-as";
				args = newArgs;
			}
			args[0] = "r2x";
			s = "r2x";
		}
		if (s.equalsIgnoreCase("owlrdf2owlxml")
		 || s.equalsIgnoreCase("owlrdf2xml")
		 || s.equalsIgnoreCase("rdf2owlxml")
		 || s.equalsIgnoreCase("rdf2xml")
		 || s.equalsIgnoreCase("r2x")
		 || s.equalsIgnoreCase("abstractsyntax2owlxml")
		 || s.equalsIgnoreCase("abstractsyntax2xml")
		 || s.equalsIgnoreCase("abstract2owlxml")
		 || s.equalsIgnoreCase("abstract2xml")
		 || s.equalsIgnoreCase("a2x")) {
			Syntax syntax = new Syntax();
			if (!syntax.processArguments(args)) return;
			syntax.compile();
			syntax.printer.close();
			return;
		}
//		if (s.equalsIgnoreCase("rdf2owl") || s.equalsIgnoreCase("r2o")) {
//			rdf2owl r2o = new rdf2owl();
//			if (!r2o.processArguments(args)) return;
//			r2o.compile();
//			r2o.printer.close();
//			return;
//		}
		if (s.equalsIgnoreCase("filter")) {
			Filter filter = new Filter();
			if (!filter.processArguments(args)) return;
			filter.compile();
			filter.printer.close();
			return;
		}
		if (s.equalsIgnoreCase("deo")) {
			Deo d = new Deo();
			if (!d.processArguments(args)) return;
			d.compile();
			d.printer.close();
			return;
		}
		if (s.equalsIgnoreCase("ded")) {
			Ded d = new Ded();
			if (!d.processArguments(args)) return;
			d.compile();
			d.printer.close();
			return;
		}
		if (s.equalsIgnoreCase("pop") || s.equalsIgnoreCase("populate")) {
			Populate pop = new Populate();
			if (!pop.processArguments(args)) return;
			pop.compile();
			pop.printer.close();
			return;
		}
		if (s.equalsIgnoreCase("diff")) {
			Diff d = new Diff();
			if (!d.processArguments(args)) return;
			d.compile();
			d.printer.close();
			return;
		}
		if (s.equalsIgnoreCase("merge")) {
			Merge d = new Merge();
			if (!d.processArguments(args)) return;
			d.compile();
			d.printer.close();
			return;
		}
		if (s.equalsIgnoreCase("satisfiable") || s.equalsIgnoreCase("sat")) {
			Satisfiable sat = new Satisfiable();
			if (!sat.processArguments(args)) return;
			sat.compile();
			sat.printer.close();
			return;
		}
		if (s.equalsIgnoreCase("screech")) {
			String newArgs[] = new String[argc+1];
			for (int i = 0; i<argc; ++i)
				newArgs[i] = args[i];
			newArgs[argc] = "-screech";
			args = newArgs;
			args[0] = "dlp";
			s = "dlp";
		}
		if (s.equalsIgnoreCase("dlpconvert") || s.equalsIgnoreCase("dlp")) {
			DLPConvert dlp = new DLPConvert();
			if (!dlp.processArguments(args)) return;
			dlp.compile();
			dlp.printer.close();
			return;
		}
		if (s.equalsIgnoreCase("demeta")) {
			String newArgs[] = new String[argc+1];
			for (int i = 0; i<argc; ++i)
				newArgs[i] = args[i];
			newArgs[argc] = "-demeta";
			args = newArgs;
			args[0] = "meta";
			s = "meta";
		}
		if (s.equalsIgnoreCase("meta")) {
			Meta meta = new Meta();
			if (!meta.processArguments(args)) return;
			meta.compile();
			meta.printer.close();
			return;
		}
		if (s.equalsIgnoreCase("latex")) {
			Latex latex = new Latex();
			if (!latex.processArguments(args)) return;
			latex.compile();
			latex.printer.close();
			return;
		}
		if (s.equalsIgnoreCase("luna")) {
			Luna luna = new Luna();
			if (!luna.processArguments(args)) return;
			luna.compile();
			luna.printer.close();
			return;
		}
		if (s.equalsIgnoreCase("server")) {
			Server server = new Server();
			if (!server.processArguments(args)) return;
			server.compile();
			server.printer.close();
			return;
		}
		if (s.equalsIgnoreCase("shell")) {
			Shell shell = new Shell();
			if (!shell.processArguments(args)) return;
			shell.compile();
			shell.printer.close();
			return;
		}
		if (s.equalsIgnoreCase("eval") || s.equalsIgnoreCase("evaluate")) {
			Evaluate eval = new Evaluate();
			if (!eval.processArguments(args)) return;
			eval.compile();
			eval.printer.close();
			return;
		}
		if (s.equalsIgnoreCase("metrics") || s.equalsIgnoreCase("measure")) {
			Metrics metrics = new Metrics();
			if (!metrics.processArguments(args)) return;
			metrics.compile();
			metrics.printer.close();
			return;
		}
		if (s.equalsIgnoreCase("x")) {
			Script x = new Script();
			if (!x.processArguments(args)) return;
			x.compile();
			x.printer.close();
			return;
		}
		
		printUsageError();
	}
	
	/**
	 * prints the command line parameters and a short info
	 */
	public static void printUsage() {
		System.out.println("Usage: owl <command> [<commandparameter>]");
		System.out.println("");
		System.out.println("The KAON2 OWL Tools are a set of tools based on KAON2 to work with OWL files.");
		System.out.println("To get more information on a specific command, enter \"owl <command> help\". The");
		System.out.println("following commands are available in this version:");
		System.out.println("");
		System.out.println("count          counts the number of axioms/entities (same as dump -count)");
		System.out.println("ded            removing all concrete domains from the ontology");
		//System.out.println("demeta         turns a metaontology back to an ontology");
		System.out.println("deo            weakens the ontology by replacing nominals with simple classes");
		System.out.println("diff           gets all axioms in one ontology missing from the other");
		System.out.println("dlpconvert     (short: dlp) converts a DL knowledge base into logic rules");
		System.out.println("dump           (or print) prints the content of an ontology (may be filtered)");
		//System.out.println("evaluate       (or eval) accesses the evaluation routines, with an own help");
		System.out.println("filter         deletes all axioms of a specific kind (or retains only those)");
		System.out.println("latex          prints out the ontology in latex-processable DL syntax");
		System.out.println("luna           makes all individuals in an ontology different from each other");
		System.out.println("merge          creates a new ontology with the axioms of two ontologies");
		//System.out.println("meta           creates a metaontology out of an existing one");
		System.out.println("populate       (short: pop) populates an ontology randomly with instances");
		System.out.println("satisfiable    (short: sat) checks if an ontology is satisfiable or not");		
		System.out.println("screech        creates a split program based on the ontologie");
		System.out.println("shell          (beta) a shell to work with OWL, with an own help system");		
		//System.out.println("server         (beta) serves browsable information via a http server");		
		System.out.println("syntax         converts an ontology from one serialization to another");
		System.out.println("version        returns the version of this tool suite");
	}

	/**
	 * Returns the version string.
	 */
	public static String version() {
		if (developmentVersion)
			return version + " " + "beta";
		else
			return version;
	}
	
	/**
	 * Is this a development version?
	 */
	public static Boolean developmentVersion() {
		return developmentVersion;
	}

	/**
	 * prints a short excuse after not understandable rambling
	 */
	public static void printUsageError() {
		System.out.println("Sorry, didn't understand the command line parameters. Please try entering:");
		System.out.println("owl help");
	}	

	/**
	 * prints the version and copyright information
	 */
	public static void printVersion() {
		System.out.println("KAON2 OWL Tools Version " + version());
		System.out.println("(c) 2005-" + yearOfRelease + " AIFB, University of Karlsruhe");
		System.out.println("see http://owltools.ontoware.org for more information");
	}

}
