package edu.unika.aifb.owltools.serializer;

import java.util.*;

import org.semanticweb.kaon2.api.logic.*;

import edu.unika.aifb.owltools.tools.DLPConvert.IntegrityConstraintHandling;
import edu.unika.aifb.owltools.utils.Printer;


/**
 * Serializes a program in F-Logic.
 */
public class FLogicSerializer extends ProgramSerializer {
	
	public FLogicSerializer(Printer printer) { super(printer); }

	public void start() {
		m_printer.putline("<ns ontons=\"http://www.example.com/\">");
	}
	
	public void end() {
		m_printer.putline("</ns>");
		m_printer.close();
	}
	
	
	public Object visit(Rule rule) {
		//Printer.verbose("rule: " + rule.toString());
		
		int headLength = rule.getHeadLength();
		int bodyLength = rule.getBodyLength();
		
		if (headLength>1)
			m_DDL = true;
		
		boolean done = false;
				
		// Instances
		if (bodyLength==0) {
			if (headLength == 1) {
				Literal literal = rule.getHeadLiteral(0);

				if (!literal.isPositive()) {
					m_negation = true;
					m_printer.error("Negation unhandled.");
				}
				
				literal.accept(this);
				done = true;
				m_printer.putline(".");
				
			} else {
				m_printer.error("Head longer than 1, but no body.");
				done = true;
			}
		}
		
		// Subsumption
		if (bodyLength==1)
			if (headLength==1)
				if (rule.getHeadLiteral(0).getArity()==1)
					if (rule.getBodyLiteral(0).getArity()==1) {
						m_printer.put(processName(rule.getBodyLiteral(0).getPredicate().toString()));
						m_printer.put("::");
						m_printer.put(processName(rule.getHeadLiteral(0).getPredicate().toString()));
						done = true;
						m_printer.putline(".");
					}
		
		if (headLength!=1) {
			if (!done)
				if ((headLength==0) && (m_ichandling == IntegrityConstraintHandling.ICpredicate)) {
					m_printer.put("Inconsistent");
					m_printer.put(" <- ");
					// write the body literals, seperated by and
					for (int i = 0; i < bodyLength; i++) {
						if (i!=0) m_printer.put(" and ");
						rule.getBodyLiteral(i).accept(this);
					}
					done = true;
					m_printer.putline(".");					
				} else {
					m_printer.error("Head not of length 1.");
					done = true;
				}
		}
					
		if (!done) {
			m_printer.put("FORALL ");
			// gather variables and write them in a list
			// make a set
			Set<String> variables = new HashSet<String>();
			variables.addAll(getVariable(rule.getHeadLiteral(0)));
			for (int i = 0; i < bodyLength; i++) {
				variables.addAll(getVariable(rule.getBodyLiteral(i)));
			}
			// print out set names seperated with ,
			boolean first = true;
			for (String var : variables) {
				if (!first) m_printer.put(","); 
				m_printer.put(processName(var));
				first = false;
			}
			
			m_printer.put(" ");
			// write the one head (and one head only)
			rule.getHeadLiteral(0).accept(this);
			m_printer.put(" <- ");
			// write the body literals, seperated by and
			for (int i = 0; i < bodyLength; i++) {
				if (i!=0) m_printer.put(" and ");
				rule.getBodyLiteral(i).accept(this);
			}
			m_printer.putline(".");
		}
		
		
		return null;
	}
	
	public Object visit(Literal literal) {
		int arity = literal.getArity();
		if (arity == 1) {
			m_printer.put(processName(literal.getArgument(0).toString()));
			m_printer.put(":");
			m_printer.put(processName(literal.getPredicate().toString()));
		} else if (arity == 2) {
			m_printer.put(processName(literal.getArgument(0).toString()));
			m_printer.put("[");
			m_printer.put(processName(literal.getPredicate().toString()));
			m_printer.put("->>");
			m_printer.put(processName(literal.getArgument(1).toString()));			
			m_printer.put("]");
		} else {
			m_printer.error("Arity > 2. What to do?");
		}
		return null;
	}

	public Set<String> getVariable(Literal literal) {
		Set<String> set = new HashSet<String>();
		int arity = literal.getArity();
		for (int i = 0; i < arity; i++)
			if (Variable.class.isInstance(literal.getArgument(i)))
					set.add(literal.getArgument(i).toString());
		return set;
	}
}
