package edu.unika.aifb.owltools.serializer;

import java.util.*;
import org.semanticweb.kaon2.api.owl.axioms.*;
import org.semanticweb.kaon2.api.owl.elements.*;
import org.semanticweb.kaon2.api.logic.*; // TODO KAON2 Version

import edu.unika.aifb.owltools.utils.KAON2VisitorAdapter;

/**
 * Serializes an ontology in latex code
 */
public class LatexSerializer extends KAON2VisitorAdapter {

	boolean m_simple = false;
	
	String t1 = "\\axiom{";
	String t2 = "}{";
	String t3 = "}{";
	String t4 = "}";
	String a1 = "\\fact{";
	String a2 = "}";
	String c = "concept";
	String p = "property";
	String i = "individual";
	String v = "variable";
	
	public void setSimple(boolean to) { m_simple = to; if (to) { t1 = t2 = t3 = t4 = a1 = a2 = ""; } }
	
	public String toName(String s) {
		// different Naming conventions! TODO
		String label;
		if (s.lastIndexOf('#') > -1) {
			label = s.substring(s.lastIndexOf('#') + 1);
		} else {
			// return last slash-part
			if (s.lastIndexOf("/") > 0) {
				label = s.substring(s.lastIndexOf('/') + 1);
			} else {
				if (s.lastIndexOf(":") > 0) {
					label = s.substring(s.lastIndexOf(':') + 1);
				} else {
					// fallback return full uri
					label = s;
				}
			}
		}
		label = label.replace("_", "\\_");
		//label = "\\mathsf{" + label + "}";
		return label;
	}

	public String processLiteral(Object o) {
		return "\\textsf{\"" + o.toString() + "\"}";
	}
	
	public String processAnnotation(Object o) {
		return "\\textit{``" + o.toString() + "''}";
	}
	
	public String visit(DataAll da) {
		List<DataProperty> dps = da.getDataProperties();
		if (dps.size() == 1)
			return "\\forall " + dps.get(0).accept(this) + "." + visit(da.getDataRange());
		String retval = "\\forall {";
		boolean first = true;
		for (DataProperty dp : dps) {
			if (!first) retval += ", "; else first = false;
			retval += dp.accept(this);
		}
		return retval;
	}
	
	public String visit(DataCardinality dc) {
		int max = dc.getMaximumCardinality();
		int min = dc.getMinimumCardinality();
		if (min==max) return "=" + min + "~" + dc.getDataProperty().accept(this) + ".\\top";
		if (min>0) return "\\geq" + min + "~" + dc.getDataProperty().accept(this) + ".\\top";
		return "\\leq" + max + "~" + dc.getDataProperty().accept(this) + ".\\top" ;
	}
		
	public String visit(DataHasValue dhv) {
		return "\\exists " + dhv.getDataProperty().accept(this) + ".\\{" + processLiteral(dhv.getLiteralValue()) + "\\}";
	}
	
	public String visit(DataSome ds) {
		if (ds.getDataProperties().size()>1)
			return "DataSome not implemented yet for multiple properties \\owltoolserror";
		return "\\exists " + ds.getDataProperties().get(0).accept(this) + "." + ds.getDataRange().accept(this);
	}
	
	public String visit(ObjectAll oa) {
		return "\\forall " + oa.getObjectProperty().accept(this) + "." + oa.getDescription().accept(this);
	}
	
	public String visit(ObjectAnd oa) {
		Set<Description> ds = oa.getDescriptions();
		boolean first = true;
		String retval = "";
		for (Description d : ds) {
			if (!first) retval += " \\sqcap "; else first = false;
			retval += d.accept(this);
		}
		return retval;
	}

	public String visit(ObjectCardinality oc) {
		int max = oc.getMaximumCardinality();
		int min = oc.getMinimumCardinality();
		if (min==max) return "=" + min + "~" + oc.getObjectProperty().accept(this) + ".\\top";
		if (min>0) return "\\geq" + min + "~" + oc.getObjectProperty().accept(this) + ".\\top";
		return "\\leq" + max + "~" + oc.getObjectProperty().accept(this) + ".\\top" ;
	}
	
	public String visit(ObjectHasValue ohv) {
		return "\\exists " + ohv.getObjectProperty().accept(this) + ".\\{" + ohv.getIndividual().accept(this) + "\\}" ;
	}
	
	public String visit(ObjectNot on) {
		return "\\neg " + on.getDescription().accept(this);
	}

	public String visit(ObjectOneOf ooo) {
		Set<Individual> is = ooo.getIndividuals();
		boolean first = true;
		String retval = "\\{";
		for (Individual i : is) {
			if (!first) retval += ", "; else first = false;
			retval += i.accept(this);
		}
		return retval + "\\}";
	}

	public String visit(ObjectOr oo) {
		Set<Description> ds = oo.getDescriptions();
		boolean first = true;
		String retval = "";
		for (Description d : ds) {
			if (!first) retval += " \\sqcup "; else first = false;
			retval += d.accept(this);
		}
		return retval;
	}

	public String visit(ObjectSome os) {
		return "\\exists " + os.getObjectProperty().accept(this) + "." + os.getDescription().accept(this);
	}
	
	public String visit(OWLClass oc) {
		if (oc.toString().equals("owl:Thing")) return "\\top";
		if (oc.toString().equals("owl:Nothing")) return "\\bottom";
		return "\\concept{"+  toName(oc.toString()) + "}";
	}

	public String visit(DataProperty dp) {
		return "\\property{" + toName(dp.toString()) + "}";
	}

	public String visit(AnnotationProperty ap) {
		return "\\property{" + toName(ap.toString()) + "}";
	}

	public String visit(Individual i) {
		return "\\individual{" + toName(i.toString()) + "}";
	}
	
	public String visit(ObjectProperty op) {
		return "\\property{" + toName(op.toString()) + "}";
	}

	public String visit(Rule r) {
		int headLength = r.getHeadLength();
		int bodyLength = r.getBodyLength();
		String head ="";
		String body = "";
		for (int i = 0; i < headLength; ++i) {
			if (i!=0) head += " \\vee  ";
			head += r.getHeadLiteral(i).accept(this);
		}
		for (int i = 0; i < bodyLength; ++i) {
			if (i!=0) body += " \\wedge  ";
			body += r.getBodyLiteral(i).accept(this);
		}
		return t1 + head + t2 + "\\leftarrow" + t3 + body + t4;
	}
	
	public String visit(Literal l) {
		String rv = "";
		if (!l.isPositive()) rv += "\\not ";
		String terms = "";
		int arity = l.getArity();
		for (int i = 0; i < arity; ++i) {
			if (i!=0) terms += ", ";
			terms += l.getArgument(i).accept(this);
		}
		return rv + l.getPredicate().accept(this) + "(" + terms + ")";
	}
	
	public String visit(Variable var) {
		return "\\" + v + "{" + var.getVariableName() + "}";
	}
	
	public String visit(DataRange dr) {
		return "\\property{" + dr.toString() + "}";
	}

	public String visit(Annotation a) {
		return a1 + a.getAnnotationProperty().accept(this) + "(" + a.getEntity().accept(this) + ", " + processAnnotation(a.getAnnotationValue()) + ")" + a2;
	}
	
	public String visit(ClassMember m) {
		return a1 + m.getDescription().accept(this) + "(" + m.getIndividual().accept(this) + ")" + a2;
	}
	
	public String visit(DataPropertyAttribute dpa) {
		if (dpa.getAttribute()==DataPropertyAttribute.DATA_PROPERTY_FUNCTIONAL)
			return t1 + "\\top" + t2 + "\\sqsubseteq" + t3 + "\\leq 1 " + dpa.getDataProperty().accept(this) + ".\\top" + t4;
		return "UnknownAttribute \\owltoolserror"; // TODO should never happen
	}
	
	public String visit(DataPropertyDomain dpd) {
		return t1 + "\\top" + t2 + "\\sqsubseteq" + t3 + "\\forall " + dpd.getDataProperty().accept(this) + "^{-1}." + dpd.getDomain().accept(this) + t4;
	}
	
	public String visit(DataPropertyMember dpm) {
		return a1 + dpm.getDataProperty().accept(this) + "(" + dpm.getSourceIndividual().accept(this) + ", " + processLiteral(dpm.getTargetValue()) + ")" + a2;
	}
	
	public String visit(DataPropertyRange dpr) {
		return t1 + "\\top" + t2 + "\\sqsubseteq" + t3 + "\\forall " + dpr.getDataProperty().accept(this) + "." + visit(dpr.getRange()) + t4;
	}

	public String visit(DifferentIndividuals di) {
		Set<Individual> is = di.getIndividuals();
		if (is.size()==2) return t1 + ((Individual)is.toArray()[0]).accept(this) + t2 + "\\not\\approx" + t3 + ((Individual)is.toArray()[1]).accept(this) + t4;
		String retval = "";
		boolean first = true;
		for (Individual i : is) {
			if (first) first = false; else retval += " \\not\\approx ";
			retval += i.accept(this);
		}
		return a1 + retval + a2;
	}

	public String visit(DisjointClasses dc) { // TODO Check Syntax!
		Set<Description> dcs = dc.getDescriptions();
		if (dcs.size()==2) return t1 + ((Description)dcs.toArray()[0]).accept(this) + "\\sqcap" + ((Description)dcs.toArray()[1]).accept(this) + t2 + "\\sqsubseteq" + t3 + "\\bot" + t4;
		String retval = "\\concept{AllDisjoint}(";
		boolean first = true;
		for (Description d : dc.getDescriptions()) {
			if (first) first = false; else retval += ", ";
			retval += d.accept(this);
		}
		retval += ")";
		return a1 + retval + a2;
	}
	
	public String visit(EquivalentClasses ec) {
		Set<Description> ecs = ec.getDescriptions();
		if (ecs.size()==2) return t1 + ((Description)ecs.toArray()[0]).accept(this) + t2 + "\\equiv" + t3 + ((Description)ecs.toArray()[1]).accept(this) + t4;
		String retval = "";
		boolean first = true;
		for (Description d : ecs) {
			if (first) first = false; else retval += " \\equiv ";
			retval += d.accept(this);
		}
		return a1 + retval + a2;
	}	
	
	public String visit(EquivalentDataProperties edp) {
		Set<DataProperty> edps = edp.getDataProperties();
		if (edps.size()==2) return t1 + ((DataProperty)edps.toArray()[0]).accept(this) + t2 + "\\equiv" + t3 + ((DataProperty)edps.toArray()[1]).accept(this) + t4;
		String retval = "";
		boolean first = true;
		for (DataProperty dp : edps) {
			if (first) first = false; else retval += " \\equiv ";
			retval += dp.accept(this);
		}
		return a1 + retval + a2;
	}	
	
	public String visit(EquivalentObjectProperties eop) {
		Set<ObjectProperty> eops = eop.getObjectProperties();
		if (eops.size()==2) return t1 + ((ObjectProperty)eops.toArray()[0]).accept(this) + t2 + "\\equiv" + t3 + ((ObjectProperty)eops.toArray()[1]).accept(this) + t4;
		String retval = "";
		boolean first = true;
		for (ObjectProperty op : eops) {
			if (first) first = false; else retval += " \\equiv ";
			retval += op.accept(this);
		}
		return a1 + retval + a2;
	}	
	
	public String visit(InverseObjectProperties iop) {
		return t1 + iop.getFirst().accept(this) + t2 + " \\equiv " + t3 + iop.getSecond().accept(this) + "^{-1}" + t4;
	}	
	
	public String visit(ObjectPropertyAttribute opa) { // TODO Syntax!
		if (opa.getAttribute()==ObjectPropertyAttribute.OBJECT_PROPERTY_FUNCTIONAL)
			return t1 + "\\top " + t2 + " \\sqsubseteq " + t3 + " \\leq1~" + opa.getObjectProperty().accept(this) + ".\\top" + t4;
		if (opa.getAttribute()==ObjectPropertyAttribute.OBJECT_PROPERTY_INVERSE_FUNCTIONAL)
			return t1 + "\\top " + t2 + " \\sqsubseteq " + t3 + " \\leq1~" + opa.getObjectProperty().accept(this) + "^{-1}.\\top" + t4;
		if (opa.getAttribute()==ObjectPropertyAttribute.OBJECT_PROPERTY_SYMMETRIC)
			return t1 + opa.getObjectProperty().accept(this) + t2 + " \\equiv " + t3 + opa.getObjectProperty().accept(this) + "^{-1}" + t4;
		if (opa.getAttribute()==ObjectPropertyAttribute.OBJECT_PROPERTY_TRANSITIVE)
			return a1 + "\\concept{Trans}(" + opa.getObjectProperty().accept(this) + ")" + a2;
		return "UnknownAttribute \\owltoolserror"; // TODO should never happen
	}
	
	public String visit(ObjectPropertyDomain opd) {
		return t1 + "\\top " + t2 + " \\sqsubseteq " + t3 + " \\forall " + opd.getObjectProperty().accept(this) + "^{-1}." + opd.getDomain().accept(this) + t4;
	}
	
	public String visit(ObjectPropertyMember opm) {
		return a1 + opm.getObjectProperty().accept(this) + "(" + opm.getSourceIndividual().accept(this) + ", " + opm.getTargetIndividual().accept(this) + ")" + a2;
	}
	
	public String visit(ObjectPropertyRange opr) {
		return t1 + "\\top " + t2 + " \\sqsubseteq " + t3 + " \\forall " + opr.getObjectProperty().accept(this) + "." + opr.getRange().accept(this) + t4;
	}

	public String visit(SameIndividual si) {
		Set<Individual> is = si.getIndividuals();
		if (is.size()==2) return t1 + ((Individual)is.toArray()[0]).accept(this) + t2 + "\\approx" + t3 + ((Individual)is.toArray()[1]).accept(this) + t4;
		String retval = "";
		boolean first = true;
		for (Individual i : is) {
			if (first) first = false; else retval += " \\approx ";
			retval += i.accept(this);
		}
		return a1 + retval + a2;
	}

	public String visit(SubClassOf sco) {
		return t1 + sco.getSubDescription().accept(this) + t2 + " \\sqsubseteq " + t3 + sco.getSuperDescription().accept(this) + t4;
	}
	
	public String visit(SubDataPropertyOf sdpo) {
		return t1 + sdpo.getSubDataProperty().accept(this) + t2 + " \\sqsubseteq " + t3 + sdpo.getSuperDataProperty().accept(this) + t4;
	}
	
	public String visit(SubObjectPropertyOf sopo) {
		return t1 + sopo.getSubObjectProperty().accept(this) + t2 + " \\sqsubseteq " + t3 + sopo.getSuperObjectProperty().accept(this) + t4;
	}	

}
