package edu.unika.aifb.owltools.serializer;

import java.util.*;
import java.lang.Character;

import org.semanticweb.kaon2.api.logic.*; // TODO KAON2 Version

import edu.unika.aifb.owltools.tools.DLPConvert.IntegrityConstraintHandling;
import edu.unika.aifb.owltools.utils.KAON2VisitorAdapter;
import edu.unika.aifb.owltools.utils.Printer;

/**
 * An abstract serializer for programs. Provides the name serialization algorithms
 * common to most of the serializer implementations.
 */
abstract public class ProgramSerializer extends KAON2VisitorAdapter {
	
	protected Printer m_printer;
	protected boolean m_DDL = false;
	protected boolean m_screech = true;
	protected boolean m_stat = false;
	protected boolean m_emptyHead = false;
	protected boolean m_negation = false;
	protected IntegrityConstraintHandling m_ichandling = IntegrityConstraintHandling.asIC;
	
	public ProgramSerializer(Printer printer) {
		m_printer = printer;
	}
	
	public void setScreech(boolean to) {
		m_screech = to;
	}
	
	public boolean getScreech() {
		return m_screech;
	}
	
	public void setStat(boolean to) {
		m_stat = to;
	}
	
	public void setIC(IntegrityConstraintHandling ic) {
		m_ichandling = ic;
	}

	public boolean encounteredDDL() {
		return m_DDL;
	}
	
	public boolean emptyHead() {
		return m_emptyHead;
	}
	
	public boolean hasNegation() {
		return m_negation;
	}

	public enum NameSerializationStyle { uri, fragment, atom, ns };
	
	protected NameSerializationStyle m_nameStyle; 

	public void setStyle(NameSerializationStyle style) {
		m_nameStyle = style;
	}
	
	/**
	 * Return first letter uppercase, all other letters lowercase Remove all
	 * non-alphanumeric
	 * 
	 * Taken from RDFReactor, with eternal gratefulness to mvo
	 * 
	 * @param s
	 * @return first letter.toUpperCase() + rest.toLowerCase()
	 */
	public static String beanify(String s, boolean atom) {
		//s = s.substring(0, 1).toUpperCase() + s.substring(1);
		if (Character.isDigit(s.charAt(0)))
			s = "_" + s;
		s = s.replaceAll("[\\-=?]", "_");
		return s.replaceAll("[\\.\\[\\]\\/\\!\\\"\\:\\-\\#\\\\$\\%\\&\\/\\(\\)=?]", (atom?"_":""));
	}

	/**
	 * toName gives a resource a readable and prologable name
	 * 
	 * Based on RDFReactors code, with eternal gratefulness to mvo
	 * 
	 * @param uriString	the uri of the resource to be named
	 * @return	the name of the returned resource
	 */
	public static String toName(String uriString) {
		String label;
		if (uriString.lastIndexOf('#') > 0) {
			label = uriString.substring(uriString.lastIndexOf('#') + 1);
		} else {
			// return last slash-part
			if (uriString.lastIndexOf("/") > 0) {
				label = uriString.substring(uriString.lastIndexOf('/') + 1);
			} else {
				// fallback return full uri
				label = uriString;
			}
		}
		// post process uri
		label = label.toLowerCase();

		// make label bean-style
		return beanify(label, false);
	}
	
	
	protected String processName(String name) {
		switch (m_nameStyle) {
			case fragment :
				return toName(name);
			case atom :
				return beanify(name, true);
			case ns :
				return name;
			default   :
				return name;
		}
	}
	
	public void start(String filename) {
		if (filename.equals(""))
			start();
		else
			m_printer.setFile(filename);
	}
	
	public void start() {
	}
	
	public void end() {
		m_printer.close();
	}
	
	public Object visit(List<Rule> program) {
		for (Rule rule : program) {
			if (rule.getHeadLength()==0) {
				if (m_ichandling == IntegrityConstraintHandling.noIC)
					continue;
				rule.accept(this);
			}
		}
		return null;
	}

}
