package edu.unika.aifb.owltools.serializer;

import org.semanticweb.kaon2.api.logic.*;
import org.semanticweb.kaon2.api.owl.elements.*;

import edu.unika.aifb.owltools.tools.DLPConvert.IntegrityConstraintHandling;
import edu.unika.aifb.owltools.utils.Printer;

/**
 * Serializes a set of rules to an XSB Prolog program.
 */
public class PrologSerializer extends ProgramSerializer {
	
	private boolean m_head = false;
	private boolean m_emptyBody = false;
	private int m_body = 0;
	
	public PrologSerializer(Printer printer) {
		super(printer);
	}
	
	public Object visit(Rule rule) {
		//Printer.verbose("rule: " + rule.toString());
		int headLength = rule.getHeadLength();
		int bodyLength = rule.getBodyLength();
		m_emptyBody = (bodyLength==0);
		
		if (headLength == 0)
			m_emptyHead = true;
		
		if (headLength > 1)
			m_DDL = true;
		
		if (m_screech) {
			if (headLength == 0) {
				if (!m_emptyBody) {
					if (m_ichandling == IntegrityConstraintHandling.ICpredicate)
						m_printer.put("Inconsistent");
					m_printer.put(" :- ");
					for (int j = 0; j < bodyLength; ++j) {
						m_printer.put(j==0 ? "" : ", ");
						m_body = 0;
						rule.getBodyLiteral(j).accept(this);
					}					
					m_printer.putline(".");
				}				
			} else {
				for (int i = 0; i < headLength; ++i) {
					m_head = true;
					rule.getHeadLiteral(i).accept(this);
					m_head = false;
					if (!m_emptyBody) {
						m_printer.put(" :- ");
						for (int j = 0; j < bodyLength; ++j) {
							m_printer.put(j==0 ? "" : ", ");
							m_body = 0;
							rule.getBodyLiteral(j).accept(this);
							if (m_stat) {
								m_printer.put(", N is ");
								for (int k = 0; k < m_body; ++k) {
									if (k != 0)
										m_printer.put(" + ");
									m_printer.put("N" + k);
								}
								if (headLength > 1)
									m_printer.put(" + 1");
							}
						}					
					}
					m_printer.putline(".");
				}				
			}
		} else {
			for (int i = 0; i < headLength; ++i) {
				m_printer.put(i==0 ? "" : ", ");
				rule.getHeadLiteral(i).accept(this);
			}
			
			if (!m_emptyBody) {
				m_printer.put(" :- ");
				for (int i = 0; i < bodyLength; ++i) {
					m_printer.put(i==0 ? "" : ", ");
					rule.getBodyLiteral(i).accept(this);
				}
			}
			m_printer.putline(".");
		}
				
		return null;
	}
	
	public Object visit(Literal literal) {
		
		// negation
		if (!literal.isPositive()) {
			m_printer.put("NOT ");
			m_negation = true;
		}
		
		// predicate
		//literal.getPredicate().accept(this);
		m_printer.put(processName(literal.getPredicate().toString()));
		
		// brackets
		m_printer.put("(");
		int arity = literal.getArity();
		for (int i = 0; i < arity; ++i) {
			m_printer.put(i==0 ? "" : ", ");
			literal.getArgument(i).accept(this);
		}
		if (m_stat) {
			if (m_head) {
				if (m_emptyBody)
					m_printer.put(", 0");
				else
					m_printer.put(", N");
			} else {
				m_printer.put(", N" + (m_body++));
			}
		}
		m_printer.put(")");
		return null;
	}
	
	public Object visit(Constant constant) {
		m_printer.put(processName(constant.toString()));
		return null;
	}

	public Object visit(Individual individual) {
		m_printer.put(processName(individual.toString()));
		return null;
	}

	public Object visit(Variable variable) {
		m_printer.put(variable.toString());
		return null;
	}

}
