/*
 * dump just dumps what it understands of a read ontology.
 */
package edu.unika.aifb.owltools.tools;

import java.util.*;

import org.semanticweb.kaon2.api.*;
import org.semanticweb.kaon2.api.owl.axioms.*;
import org.semanticweb.kaon2.api.owl.elements.*;
import org.semanticweb.kaon2.api.logic.*; // TODO KAON2 Version


/**
 * dump is used to serialize the ontology, or filtered parts of it, or to count it 
 */
public class Dump extends KAON2Tool {
	/**
	 * Name of the tool
	 */
	public String name() { return "dump"; }
	
	/**
	 * Input filename
	 */
	private String m_fileuri = "";
	
	/**
	 * Name of the output file
	 */
	private String m_output = "";
	
	/**
	 * Is this Dump called for counting or for dumping?
	 */
	private boolean m_count = false;
	
	/**
	 * Is this Dump called for the entites or for the axioms?
	 */
	private boolean m_entities = false;
	
	/**
	 * Filter to be used for counting or dumping
	 */
	private Class m_filter = null;
	
	/**
	 * If the tool is being called from the outside, some basic processing
	 * is being done
	 * @param args Command line parameters. The same as for the OWLtools call
	 */
	public static void main(String[] args) {
		Dump dump = new Dump();
		if (!dump.processArguments(args)) return;
		dump.compile();
		dump.printer.close();
	}
	
	/**
	 * processes the command line parameteres
	 */
	public boolean processArguments(String[] args) {
		// process command line arguments
		
		int argc = args.length;
		if (argc == 0) {
			printUsage();
			return false;
		}
		String s = args[0].toLowerCase();
		if (argc == 1) {
			if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) || (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) || (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help")) || (s.equalsIgnoreCase("dump")) || (s.equalsIgnoreCase("count"))) {
				printUsage();
				return false;
			}
		}

		boolean hasFileName = false;
		boolean hasOutput = false;
		boolean hasCount = false;
		boolean hasFilter = false;
		boolean hasEntities = false;
		
		for (int i = 0; i < argc; ++i) {
			boolean unknownArg = true;
			
			String arg = args[i];

			if ((arg.equalsIgnoreCase("-?")) || (arg.equalsIgnoreCase("?")) || (arg.equalsIgnoreCase("-h")) || (arg.equalsIgnoreCase("--help")) || (arg.equalsIgnoreCase("-help")) || (arg.equalsIgnoreCase("help"))) {
				printUsage();
				return false;
			}

			if (arg.startsWith("-")) {
				
				if (arg.equalsIgnoreCase("-count")) {
					if (hasCount) { printUsageError(); return false; }
					m_count = true;
					hasCount = true;
					unknownArg = false;
				}
				
				if (arg.equalsIgnoreCase("-entities")||arg.equalsIgnoreCase("-e")) {
					if (hasEntities) { printUsageError(); return false; }
					m_entities = true;
					hasEntities = true;
					unknownArg = false;
				}
				
				// filter (Entities)
				if (arg.equalsIgnoreCase("-AnnotationProperty")||arg.equalsIgnoreCase("-ap")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = AnnotationProperty.class;
					m_entities = true;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-DataProperty")||arg.equalsIgnoreCase("-dp")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = DataProperty.class;
					m_entities = true;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-DataType")||arg.equalsIgnoreCase("-dt")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = Datatype.class;
					m_entities = true;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-Individual")||arg.equalsIgnoreCase("-i")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = Individual.class;
					m_entities = true;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-NonOwlPredicate")||arg.equalsIgnoreCase("-nop")) {
					if (hasFilter) { printUsageError(); return false; }
					//m_filter = NonOWLPredicate.class;
					m_entities = true;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-owlClass")||arg.equalsIgnoreCase("-c")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = OWLClass.class;
					m_entities = true;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-OwlEntity")||arg.equalsIgnoreCase("-oe")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = OWLEntity.class;
					m_entities = true;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-ObjectProperty")||arg.equalsIgnoreCase("-op")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = ObjectProperty.class;
					m_entities = true;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-ObjectCardinality")||arg.equalsIgnoreCase("-oc")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = ObjectCardinality.class;
					m_entities = true;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-ObjectSome")||arg.equalsIgnoreCase("-oc")) {
					if (hasFilter) { printUsageError(); return false; }
					m_filter = ObjectSome.class;
					m_entities = true;
					hasFilter = true;
					unknownArg = false;
				}
				
				
				// filter (Axioms)
				if (arg.equalsIgnoreCase("-Annotation")||arg.equalsIgnoreCase("-a")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = Annotation.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-ClassMember")||arg.equalsIgnoreCase("-cm")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = ClassMember.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-DataPropertyAttribute")||arg.equalsIgnoreCase("-dpa")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = DataPropertyAttribute.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-DataPropertyDomain")||arg.equalsIgnoreCase("-dpd")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = DataPropertyDomain.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-DataPropertyMember")||arg.equalsIgnoreCase("-dpm")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = DataPropertyMember.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-DataPropertyRange")||arg.equalsIgnoreCase("-dpr")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = DataPropertyRange.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-Deprecation")||arg.equalsIgnoreCase("-d")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = Deprecation.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-DifferentIndividuals")||arg.equalsIgnoreCase("-di")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = DifferentIndividuals.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-DisjointClasses")||arg.equalsIgnoreCase("-dc")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = DisjointClasses.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-EquivalentClasses")||arg.equalsIgnoreCase("-ec")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = EquivalentClasses.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-EquivalentDataProperties")||arg.equalsIgnoreCase("-edp")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = EquivalentDataProperties.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-EquivalentObjectProperties")||arg.equalsIgnoreCase("-eop")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = EquivalentObjectProperties.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-InverseObjectProperties")||arg.equalsIgnoreCase("-iop")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = InverseObjectProperties.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-ObjectPropertyAttribute")||arg.equalsIgnoreCase("-opa")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = ObjectPropertyAttribute.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-ObjectPropertyDomain")||arg.equalsIgnoreCase("-opd")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = ObjectPropertyDomain.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-ObjectPropertyMember")||arg.equalsIgnoreCase("-opm")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = ObjectPropertyMember.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-ObjectPropertyRange")||arg.equalsIgnoreCase("-opr")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = ObjectPropertyRange.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-Rule")||arg.equalsIgnoreCase("-r")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = Rule.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-SameIndividual")||arg.equalsIgnoreCase("-si")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = SameIndividual.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-SubClassOf")||arg.equalsIgnoreCase("-sco")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = SubClassOf.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-SubDataPropertyOf")||arg.equalsIgnoreCase("-sdpo")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = SubDataPropertyOf.class;
					hasFilter = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-SubObjectPropertyOf")||arg.equalsIgnoreCase("-sopo")) {
					if (hasFilter||hasEntities) { printUsageError(); return false; }
					m_filter = SubObjectPropertyOf.class;
					hasFilter = true;
					unknownArg = false;
				}
				
			} else {
				if (arg.equalsIgnoreCase("dump")) {
					unknownArg = false;
				} else {
					if (hasFileName) {
						if (hasOutput) { printUsageError(); return false; }
						m_output = arg;
						hasOutput = true;
						unknownArg = false;
					} else {
						m_fileuri = arg;
						hasFileName = true;					
						unknownArg = false;
					}
				}
			}
			
			if (unknownArg) { printUsageError(); return false; } 
		}
		if (!hasFileName) { printUsageError(); return false; }
		return true;
	}
	
	/**
	 * Lists all the axioms of the ontology
	 * @param ontology The ontology to be listed
	 * @return A list of strings, each one representing one axiom
	 * @throws KAON2Exception
	 */
	public List<String> dump(Ontology ontology) throws KAON2Exception {
		List<String> retval = new LinkedList<String>();
		Request<Axiom> request = ontology.createAxiomRequest();
		
		Set<Axiom> axioms = request.get();
		for (Axiom axiom : axioms)
			retval.add(axiom.toString());

		return retval;
	}

	/**
	 * Lists all the axioms of a special kind
	 * @param ontology The ontology to be listed from
	 * @param filter The type of axiom. Must be one of the KAON2 axiom sublcass
	 * @return A list of strings, each one representing one axiom of the looked for type
	 * @throws KAON2Exception
	 */
	public List<String> dump(Ontology ontology, Class filter) throws KAON2Exception {
		List<String> retval = new LinkedList<String>();
		Request<Axiom> request = ontology.createAxiomRequest(filter);
		
		Set<Axiom> axioms = request.get();
		for (Axiom axiom : axioms)
			retval.add(axiom.toString());

		return retval;
	}
	
	/**
	 * Lists all entities in the ontology
	 * @param ontology The ontology to be listed from
	 * @return A list of strings, each one being an URI of an entity in the ontology
	 * @throws KAON2Exception
	 */
	public List<String> dumpEntities(Ontology ontology) throws KAON2Exception {
		List<String> retval = new LinkedList<String>();
		Request<Entity> request = ontology.createEntityRequest();
		
		Set<Entity> entities = request.get();
		for (Entity entity : entities)
			retval.add(entity.toString());

		return retval;
	}

	/**
	 * Lists all entities of the specified kind in the ontology
	 * @param ontology The ontology to be listed from
	 * @param filter The kind of entities to be looked for
	 * @return A list of strings, each one being an URI of an entity of the specified kind
	 * @throws KAON2Exception
	 */
	public List<String> dumpEntities(Ontology ontology, Class filter) throws KAON2Exception {
		List<String> retval = new LinkedList<String>();
		Request<Entity> request = ontology.createEntityRequest(filter);
		
		Set<Entity> entities = request.get();
		for (Entity entity : entities)
			retval.add(entity.toString());

		return retval;
	}
	
	/**
	 * Counts the number of axioms in an ontology
	 * @param ontology The ontology to be counted
	 * @return The number of axioms in the ontology
	 * @throws KAON2Exception
	 */
	public int count(Ontology ontology) throws KAON2Exception {
		int retval;
		Request<Axiom> request = ontology.createAxiomRequest();
		retval = request.size();
		return retval;
	}
	
	/**
	 * Counts the number of axioms of the specified kind in an ontology
	 * @param ontology The ontology to be counted
	 * @param filter The kind of axioms to be counted
	 * @return The number of axioms of the specified kind in the ontology
	 * @throws KAON2Exception
	 */
	public int count(Ontology ontology, Class filter) throws KAON2Exception {
		int retval;
		Request<Axiom> request = ontology.createAxiomRequest(filter);
		retval = request.size();
		return retval;
	}

	/**
	 * Counts the number of entities mentioned in an ontology
	 * @param ontology The ontology to be counted
	 * @return The number of entities in the ontology
	 * @throws KAON2Exception
	 */
	public int countEntities(Ontology ontology) throws KAON2Exception {
		int retval;
		Request<Entity> request = ontology.createEntityRequest();
		retval = request.size();
		return retval;
	}
	
	/**
	 * Counts the number of entities of the specified kind in an ontology
	 * @param ontology The ontology to be counted
	 * @param filter The kind of entities to be counted
	 * @return The number of entities of the specified kind in the ontology
	 * @throws KAON2Exception
	 */
	public int countEntities(Ontology ontology, Class filter) throws KAON2Exception {
		int retval;
		Request<Entity> request = ontology.createEntityRequest(filter);
		retval = request.size();
		return retval;
	}

	/**
	 * Entry point for the command line based usage of the Dump owl tool
	 *
	 */
	public void compile() {
		try {
			if (!m_output.equals("")) printer.setFile(m_output);
			Ontology ontology = loadOntology(m_fileuri);

	        try {
	        	if (m_count)
	        		if (m_entities)
	        			if (m_filter == null)
	        				printer.putline(countEntities(ontology));
	        			else
	        				printer.putline(countEntities(ontology, m_filter));
	        		else
	        			if (m_filter == null)
	        				printer.putline(count(ontology));
	        			else
	        				printer.putline(count(ontology, m_filter));
	        	else {
	        		List<String> messages;
	        		if (m_entities)
	        			if (m_filter == null)
	        				messages = dumpEntities(ontology);
	        			else
	        				messages = dumpEntities(ontology, m_filter);
	        		else
	        			if (m_filter == null)
	        				messages = dump(ontology);
	        			else
	        				messages = dump(ontology, m_filter);
	        		for (String message : messages )
	        			printer.putline(message);
	        	}
	        } catch (KAON2Exception e) {
				printer.error("Could not open ontology " + m_fileuri);
				printer.error(e.toString());
			}
			finally {
				endConnection();
			}
		} catch (KAON2Exception e) {
			printer.error("Could not open or close KAON2 connection.");
			printer.error(e.toString());
		} catch (InterruptedException e) {
			printer.error("Interrupted opening KAON2 connection.");
			printer.error(e.toString());
		}
	}

	/**
	 * prints the command line parameters and a short info
	 */
	public static void printUsage() {
		System.out.println("Usage: owl dump <input_url> [<output_filename>] [-count] [-{filter}]");
		System.out.println("dump reads an OWL ontology and returns in its own syntax what it understood.");
		System.out.println("");
		System.out.println(" <input_url>");
		System.out.println("     The location of the ontology to be read.");
		System.out.println(" <output_filename>");     
		System.out.println("     The name of the file where the dump will be written to.");
		System.out.println("     If the filename is omitted, the output is send to the console.");
		System.out.println(" -count    returns the number of axioms or entities");
		System.out.println(" -entities (or just -e) returns entities instead of axioms");
		System.out.println(" -{filter} returns axioms or entities of the given kind. Available filters:");
		System.out.println("           Annotation, ClassMember, DataPropertyAttribute, DataPropertyDomain,");
		System.out.println("           DataPropertyMember, DataPropertyRange, Deprecation,");
		System.out.println("           DifferentIndividuals, DisjointClasses, EquivalentClasses,");
		System.out.println("           EquivalentDataProperties, EquivalentObjectProperties,");
		System.out.println("           InverseObjectProperties, ObjectPropertyAttribute,");
		System.out.println("           ObjectPropertyDomain, ObjectPropertyMember, ObjectPropertyRange,");
		System.out.println("           Rule, SameIndividual, SubClassOf, SubDataPropertyOf,");
		System.out.println("           SubObjectPropertyOf. AnnotationProperty, DataProperty, DataType,");
		System.out.println("           Individual, NonOwlPredicate, ObjectProperty, owlClass. All kinds may");
		System.out.println("           be abbreviated with the capital letters (e.g. DT instead DataType)");
	}

}
