/*
 * meta makes a meta-ontology from an ontology, and back again
 */
package edu.unika.aifb.owltools.tools;

import java.io.*;
import java.util.*;

import org.semanticweb.kaon2.api.*;
import org.semanticweb.kaon2.api.reasoner.*;
import org.semanticweb.kaon2.api.owl.axioms.*;
import org.semanticweb.kaon2.api.owl.elements.*;
import org.semanticweb.kaon2.api.logic.*;
import org.semanticweb.kaon2.api.formatting.*;

import edu.unika.aifb.owltools.api;

/**
 * meta makes a meta-ontology from an ontology, and back again
 */
public class Meta extends KAON2Tool {
	/**
	 * the name of the owl tool
	 */
	public String name() { return "meta"; }
	
	/**
	 * the URI of the ontology to be meta'd
	 */
	private String m_fileuri = "";
	
	/**
	 * the filename where to save the result
	 */
	private String m_output = "";
	
	/**
	 * The format in which to serialize the results
	 */
	OntologyFileFormat m_format = OntologyFileFormat.OWL_RDF;
	
	/**
	 * If the meta-ontology is to be saved in the same ontology
	 * as the original ontology (or vice versa)
	 */
	private boolean m_same = false;
	
	/**
	 * If backlinking annotation properties are to be added to
	 * the metaontology
	 */
	private boolean m_backref = true;
	
	/**
	 * If the ontology is not to be meta'd, but rather a metaontology
	 * being turned concrete again
	 */
	private boolean m_demeta = false;

	/**
	 * Random variable. Needed for the very sophisticated naming
	 * algorithm for new entities
	 */
	static Random m_random = new Random();

	/**
	 * Namespace for the metaontology entities
	 */
	static String meta_ns = "http://ontoware.org/meta/2006/04#";
	static String meta_entity = meta_ns + "Entity";
	static String meta_class = meta_ns + "Class";
	static String meta_individual = meta_ns + "Individual";
	static String meta_objectproperty = meta_ns + "ObjectProperty";
	static String meta_dataproperty = meta_ns + "DataProperty";
	static String meta_functionaldataproperty = meta_ns + "FunctionalDataProperty";
	static String meta_datatype = meta_ns + "Datatype";
	static String meta_annotationproperty = meta_ns + "AnnotationProperty";
	static String meta_nonowlpredicate = meta_ns + "NonOWLPredicate";

	static String meta_description = meta_ns + "Description";
	static String meta_dataall = meta_ns + "DataAll";
	static String meta_datacardinality = meta_ns + "DataCardinality";
	static String meta_datahasvalue = meta_ns + "DataHasValue";
	static String meta_datasome = meta_ns + "DataSome";
	static String meta_objectall = meta_ns + "ObjectAll";
	static String meta_objectand = meta_ns + "ObjectAnd";
	static String meta_objectcardinality = meta_ns + "ObjectCardinality";
	static String meta_objecthasvalue = meta_ns + "ObjectHasValue";
	static String meta_objectnot = meta_ns + "ObjectNot";
	static String meta_objectoneof = meta_ns + "ObjectOneOf";
	static String meta_objector = meta_ns + "ObjectOr";
	static String meta_objectsome = meta_ns + "ObjectSome";

	static String meta_axiom = meta_ns + "Axiom";
	static String meta_annotation = meta_ns + "Annotation";
	static String meta_classmember = meta_ns + "ClassMember";
	static String meta_datapropertyattribute = meta_ns + "DataPropertyAttribute";
	static String meta_datapropertydomain = meta_ns + "DataPropertyDomain";
	static String meta_datapropertymember = meta_ns + "DataPropertyMember";
	static String meta_datapropertyrange = meta_ns + "DataPropertyRange";
	static String meta_deprecation = meta_ns + "Deprecation";
	static String meta_differentindividuals = meta_ns + "DifferentIndividuals";
	static String meta_disjointclasses = meta_ns + "DisjointClasses";
	static String meta_equivalentclasses = meta_ns + "EquivalentClasses";
	static String meta_equivalentdataproperties = meta_ns + "EquivalentDataProperties";
	static String meta_equivalentobjectproperties = meta_ns + "EquivalentObjectProperties";
	static String meta_inverseobjectproperties = meta_ns + "InverseObjectProperties";
	static String meta_objectpropertyattribute = meta_ns + "ObjectPropertyAttribute";
	static String meta_objectpropertydomain = meta_ns + "ObjectPropertyDomain";
	static String meta_objectpropertymember = meta_ns + "ObjectPropertyMember";
	static String meta_objectpropertyrange = meta_ns + "ObjectPropertyRange";
	static String meta_sameindividual = meta_ns + "SameIndividual";
	static String meta_subclassof = meta_ns + "SubClassOf";
	static String meta_subdatapropertyof = meta_ns + "SubDataPropertyOf";
	static String meta_subobjectpropertyof = meta_ns + "SubObjectPropertyOf";

	static String meta_annotationpropertyinstance = meta_ns + "AnnotationPropertyInstance";
	
	static String meta_classmemberclass = meta_ns + "ClassMemberClass";
	static String meta_classmemberindividual = meta_ns + "ClassMemberIndividual";
	static String meta_ismemberof = meta_ns + "isMemberOf";
	
	static String meta_datapropertydomaindataproperty = meta_ns + "DataPropertyDomainDataProperty";
	static String meta_datapropertydomaindomain = meta_ns + "DataPropertyDomainDomain";
	static String meta_datapropertyrangedataproperty = meta_ns + "DataPropertyRangeDataProperty";
	static String meta_datapropertyrangerange = meta_ns + "DataPropertyRangeRange";
	
	static String meta_objectpropertydomainobjectproperty = meta_ns + "ObjectPropertyDomainObjectProperty";
	static String meta_objectpropertydomaindomain = meta_ns + "ObjectPropertyDomainDomain";
	static String meta_objectpropertyhasdomain = meta_ns + "objectPropertyHasDomain";
	static String meta_objectpropertyrangeobjectproperty = meta_ns + "ObjectPropertyRangeObjectProperty";
	static String meta_objectpropertyrangerange = meta_ns + "ObjectPropertyRangeRange";
	static String meta_objectpropertyhasrange = meta_ns + "objectPropertyHasRange";
	
	static String meta_subclassofsuperclass = meta_ns + "SubClassOfSuperClass";
	static String meta_subclassofsubclass = meta_ns + "SubClassOfSubClass";
	static String meta_issubclassof = meta_ns + "isSubClassOf";

	/**
	 * an empty axiom
	 */
	static Axiom emptyAxiom;
	
	/**
	 * used for backlinks from metainstances to the concrete ontology entities 
	 */
	static String meta_refers = meta_ns + "refers";

	static OWLClass metaentity = KAON2Manager.factory().owlClass(meta_entity);
	static OWLClass metaclass = KAON2Manager.factory().owlClass(meta_class);
	static OWLClass metaindividual = KAON2Manager.factory().owlClass(meta_individual);
	static OWLClass metaobjectproperty = KAON2Manager.factory().owlClass(meta_objectproperty);
	static OWLClass metadataproperty = KAON2Manager.factory().owlClass(meta_dataproperty);
	static OWLClass metafunctionaldataproperty = KAON2Manager.factory().owlClass(meta_functionaldataproperty);
	static OWLClass metadatatype = KAON2Manager.factory().owlClass(meta_datatype);
	static OWLClass metaannotationproperty = KAON2Manager.factory().owlClass(meta_annotationproperty);
	static OWLClass metanonowlpredicate = KAON2Manager.factory().owlClass(meta_nonowlpredicate);

	static OWLClass metadescription = KAON2Manager.factory().owlClass(meta_description);
	static OWLClass metadataall = KAON2Manager.factory().owlClass(meta_dataall);
	static OWLClass metadatacardinality = KAON2Manager.factory().owlClass(meta_datacardinality);
	static OWLClass metadatahasvalue = KAON2Manager.factory().owlClass(meta_datahasvalue);
	static OWLClass metadatasome = KAON2Manager.factory().owlClass(meta_datasome);
	static OWLClass metaobjectall = KAON2Manager.factory().owlClass(meta_objectall);
	static OWLClass metaobjectand = KAON2Manager.factory().owlClass(meta_objectand);
	static OWLClass metaobjectcardinality = KAON2Manager.factory().owlClass(meta_objectcardinality);
	static OWLClass metaobjecthasvalue = KAON2Manager.factory().owlClass(meta_objecthasvalue);
	static OWLClass metaobjectnot = KAON2Manager.factory().owlClass(meta_objectnot);
	static OWLClass metaobjectoneof = KAON2Manager.factory().owlClass(meta_objectoneof);
	static OWLClass metaobjector = KAON2Manager.factory().owlClass(meta_objector);
	static OWLClass metaobjectsome = KAON2Manager.factory().owlClass(meta_objectsome);

	static OWLClass metaaxiom = KAON2Manager.factory().owlClass(meta_axiom);
	static OWLClass metaannotation = KAON2Manager.factory().owlClass(meta_annotation);
	static OWLClass metaclassmember = KAON2Manager.factory().owlClass(meta_classmember);
	static OWLClass metadatapropertyattribute = KAON2Manager.factory().owlClass(meta_datapropertyattribute);
	static OWLClass metadatapropertydomain = KAON2Manager.factory().owlClass(meta_datapropertydomain);
	static OWLClass metadatapropertymember = KAON2Manager.factory().owlClass(meta_datapropertymember);
	static OWLClass metadatapropertyrange = KAON2Manager.factory().owlClass(meta_datapropertyrange);
	static OWLClass metadeprecation = KAON2Manager.factory().owlClass(meta_deprecation);
	static OWLClass metadifferentindividuals = KAON2Manager.factory().owlClass(meta_differentindividuals);
	static OWLClass metadisjointclasses = KAON2Manager.factory().owlClass(meta_disjointclasses);
	static OWLClass metaequivalentclasses = KAON2Manager.factory().owlClass(meta_equivalentclasses);
	static OWLClass metaequivalentdataproperties = KAON2Manager.factory().owlClass(meta_equivalentdataproperties);
	static OWLClass metaequivalentobjectproperties = KAON2Manager.factory().owlClass(meta_equivalentobjectproperties);
	static OWLClass metainverseobjectproperties = KAON2Manager.factory().owlClass(meta_inverseobjectproperties);
	static OWLClass metaobjectpropertyattribute = KAON2Manager.factory().owlClass(meta_objectpropertyattribute);
	static OWLClass metaobjectpropertydomain = KAON2Manager.factory().owlClass(meta_objectpropertydomain);
	static OWLClass metaobjectpropertymember = KAON2Manager.factory().owlClass(meta_objectpropertymember);
	static OWLClass metaobjectpropertyrange = KAON2Manager.factory().owlClass(meta_objectpropertyrange);
	static OWLClass metasameindividual = KAON2Manager.factory().owlClass(meta_sameindividual);
	static OWLClass metasubclassof = KAON2Manager.factory().owlClass(meta_subclassof);
	static OWLClass metasubdatapropertyof = KAON2Manager.factory().owlClass(meta_subdatapropertyof);
	static OWLClass metasubobjectpropertyof = KAON2Manager.factory().owlClass(meta_subobjectpropertyof);

	static ObjectProperty metaannotationpropertyinstance = KAON2Manager.factory().objectProperty(meta_annotationpropertyinstance);

	static ObjectProperty metaclassmemberclass = KAON2Manager.factory().objectProperty(meta_classmemberclass);
	static ObjectProperty metaclassmemberindividual = KAON2Manager.factory().objectProperty(meta_classmemberindividual);
	static ObjectProperty metaismemberof = KAON2Manager.factory().objectProperty(meta_ismemberof);
	
	static ObjectProperty metadatapropertydomaindataproperty = KAON2Manager.factory().objectProperty(meta_datapropertydomaindataproperty);
	static ObjectProperty metadatapropertydomaindomain = KAON2Manager.factory().objectProperty(meta_datapropertydomaindomain);
	static ObjectProperty metadatapropertyrangedataproperty = KAON2Manager.factory().objectProperty(meta_datapropertyrangedataproperty);
	static ObjectProperty metadatapropertyrangerange = KAON2Manager.factory().objectProperty(meta_datapropertyrangerange);
	
	static ObjectProperty metaobjectpropertydomainobjectproperty = KAON2Manager.factory().objectProperty(meta_objectpropertydomainobjectproperty);
	static ObjectProperty metaobjectpropertydomaindomain = KAON2Manager.factory().objectProperty(meta_objectpropertydomaindomain);
	static ObjectProperty metaobjectpropertyhasdomain = KAON2Manager.factory().objectProperty(meta_objectpropertyhasdomain);
	static ObjectProperty metaobjectpropertyrangeobjectproperty = KAON2Manager.factory().objectProperty(meta_objectpropertyrangeobjectproperty);
	static ObjectProperty metaobjectpropertyrangerange = KAON2Manager.factory().objectProperty(meta_objectpropertyrangerange);
	static ObjectProperty metaobjectpropertyhasrange = KAON2Manager.factory().objectProperty(meta_objectpropertyhasrange);
	
	static ObjectProperty metasubclassofsuperclass = KAON2Manager.factory().objectProperty(meta_subclassofsuperclass);
	static ObjectProperty metasubclassofsubclass = KAON2Manager.factory().objectProperty(meta_subclassofsubclass);
	static ObjectProperty metaissubclassof = KAON2Manager.factory().objectProperty(meta_issubclassof);

	static AnnotationProperty refers = KAON2Manager.factory().annotationProperty(meta_refers);

	/**
	 * To be called when used from a standalone command line tool
	 * @param args Command line parameters
	 */
	public static void main(String[] args) {
		Meta m = new Meta();
		if (!m.processArguments(args)) return;
		m.compile();
		m.printer.close();
	}

	/**
	 *  process command line arguments
	 */
	public boolean processArguments(String[] args) {
		
		int argc = args.length;
		if (argc == 0) {
			printUsage();
			return false;
		}
		String s = args[0].toLowerCase();
		if (argc == 1) {
			if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) || (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) || (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help")) || (s.equalsIgnoreCase("ded"))) {
				printUsage();
				return false;
			}
		}

		boolean hasFileName = false;
		boolean hasOutput = false;
		boolean hasVerbose = false;
		boolean hasFormat = false;
		boolean hasResult = false;
		boolean hasRef = false;
		
		for (int i = 0; i < argc; ++i) {
			boolean unknownArg = true;
			
			String arg = args[i];

			if ((arg.equalsIgnoreCase("-?")) || (arg.equalsIgnoreCase("?")) || (arg.equalsIgnoreCase("-h")) || (arg.equalsIgnoreCase("--help")) || (arg.equalsIgnoreCase("-help")) || (arg.equalsIgnoreCase("help"))) {
				printUsage();
				return false;
			}
			
			if (arg.startsWith("-")) {
				if ((arg.equalsIgnoreCase("-verbose")) || (arg.equalsIgnoreCase("-verb")) || (arg.equalsIgnoreCase("-v"))) {
					if (hasVerbose) { printUsageError(); return false; }
					printer.verbose = true;
					hasVerbose = true;
					unknownArg = false;
				}					

				if ((arg.equalsIgnoreCase("-demeta")) || (arg.equalsIgnoreCase("-de"))) {
					if (hasResult) { printUsageError(); return false; }
					m_demeta = true;
					hasResult = true;
					unknownArg = false;
				}					
				
				if (arg.equalsIgnoreCase("-noref")) {
					if (hasRef) { printUsageError(); return false; }
					m_backref = false;
					hasRef = true;
					unknownArg = false;
				}					
				if (arg.equalsIgnoreCase("-ref")) {
					if (hasRef) { printUsageError(); return false; }
					m_backref = true;
					hasRef = true;
					unknownArg = false;
				}					
				
				// Format
				if (arg.equalsIgnoreCase("-rdf")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.OWL_RDF;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-xml")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.OWL_XML;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-as") || arg.equalsIgnoreCase("-abstract")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.ABSTRACT_SYNTAX;
					hasFormat = true;
					unknownArg = false;
				}
			} else {
				if (arg.equalsIgnoreCase("meta")) {
					unknownArg = false;
				} else {
					if (hasFileName) {
						if (hasOutput) { printUsageError(); return false; }
						m_output = arg;
						hasOutput = true;
						unknownArg = false;
					} else {
						m_fileuri = arg;
						hasFileName = true;					
						unknownArg = false;
					}
				}
			}
			
			if (unknownArg) { printUsageError(); return false; } 
		}
		if (!hasFileName) { printUsageError(); return false; }
		return true;
	}

	/**
	 * Entry point for the command line based usage of the Filter owl tool
	 *
	 */
	public void compile() {
		try {
	        m_same = (m_output.equals(""));

	        try {
	        	Ontology input = loadOntology(m_fileuri);
	        	Ontology output = null;
	        	if (m_demeta) {
		        	output = demeta(input); 
	        	} else {
		        	output = meta(input, m_backref);
	        	}
       	
        		Ontology result = null;
	        	if (m_same) {
	        		printer.setFile(m_fileuri);
	        		result = createOntology("http://www.aifb.uni-karlsruhe.de/WBS/dvr/owltools/meta/mergedresult");
	        		api.merge(output, input, result);
	        	} else {
	        		result = output;
	        		printer.setFile(m_output);
	        	}
	        	result.saveOntology(m_format, printer.writer, "UTF-8");

	        } catch (KAON2Exception e) {
				printer.error("Could not open ontology " + m_fileuri);
				printer.error(e.toString());
			} catch (IOException io) {
				printer.error("Could not save resulting ontology " + m_output);
				printer.error(io.toString());
			} catch (InterruptedException i) {
				printer.error("Could not save resulting ontology " + m_fileuri);
				printer.error(i.toString());				
			}
			finally {
				endConnection();
			}
		} catch (KAON2Exception e) {
			printer.error("Could not open or close KAON2 connection.");
			printer.error(e.toString());
		}
	}
	
	/**
	 * Creates a metaontology out of an ontology
	 * @param ontology The ontology to be meta'd
	 * @param backref If backreferences should be included to the original ontology entities
	 * @return The meta-ontology
	 * @throws KAON2Exception
	 */
	public Ontology meta(Ontology ontology, boolean backref) throws KAON2Exception {
		Ontology metaontology = createOntology("http://www.aifb.uni-karlsruhe.de/WBS/dvr/owltools/meta/result");
		meta(ontology, metaontology, backref);
		return metaontology;
	}
	
	/**
	 * Takes a metaontology and extracts the concrete ontology out of it
	 * @param metaontology The metaontology that hols the ontology
	 * @return The concrete ontology
	 * @throws KAON2Exception
	 */
	public Ontology demeta(Ontology metaontology) throws KAON2Exception,InterruptedException {
		Ontology ontology = createOntology("http://www.aifb.uni-karlsruhe.de/WBS/dvr/owltools/meta/result");
		demeta(metaontology, ontology);
		return ontology;
	}
	
	/**
	 * Createas a metaontology out of an ontology
	 * @param ontology The ontology to be meta'd
	 * @param metaontology The created meta-ontology
	 * @param backref If backreferences should be included to the original ontology entities
	 * @throws KAON2Exception
	 */
	public void meta(Ontology ontology, Ontology metaontology, boolean backref) throws KAON2Exception {
  		List<OntologyChangeEvent> changes = new ArrayList<OntologyChangeEvent>();

		Request<Entity> request = ontology.createEntityRequest();
		Set<Entity> entities = request.get();
		
		for (Entity entity : entities) {			
			Individual individual = KAON2Manager.factory().individual(metaURI(entity.getURI()));
			Axiom maxiom = KAON2Manager.factory().classMember(chooseClass(entity), individual); 
			changes.add(new OntologyChangeEvent(maxiom, OntologyChangeEvent.ChangeType.ADD));
			if (backref) {
				Constant constant = KAON2Manager.factory().constant(entity);
				maxiom = KAON2Manager.factory().annotation(refers, individual, constant);
				changes.add(new OntologyChangeEvent(maxiom, OntologyChangeEvent.ChangeType.ADD));
			}
		}
		metaontology.applyChanges(changes);
		
		Request<Axiom> requestAxioms = ontology.createAxiomRequest();
		Set<Axiom> axioms = requestAxioms.get();
		
		for (Axiom axiom : axioms)
			metaontology.applyChanges(meta(axiom));
	}
	
	/**
	 * Takes a metaontology and extracts the concrete ontology out of it
	 * @param metaontology The metaontology that hols the ontology
	 * @param ontology  The concrete ontology
	 * @throws KAON2Exception
	 */
	public void demeta(Ontology metaontology, Ontology ontology) throws KAON2Exception,InterruptedException {
		Reasoner reasoner = metaontology.createReasoner();

		Query query = reasoner.createQuery(metaclassmember);
		query.open();
		printer.verbose(query.getNumberOfTuples());
		while (!query.afterLast()) {
			Term[] answer = query.tupleBuffer();
			for (Term a : answer) {
				if (Individual.class.isInstance(a)) {
					Individual axiom = (Individual)a;
					ontology.addAxioms(demetaClassMember(axiom, reasoner));
				}
			}
			query.next();
		}
		query.close();
		query.dispose();

		query = reasoner.createQuery(metasubclassof);
		query.open();
		printer.verbose(query.getNumberOfTuples());
		while (!query.afterLast()) {
			Term[] answer = query.tupleBuffer();
			for (Term a : answer) {
				if (Individual.class.isInstance(a)) {
					Individual axiom = (Individual)a;
					ontology.addAxioms(demetaSubClassOf(axiom, reasoner));
				}
			}
			query.next();
		}
		query.close();
		query.dispose();
		
		Namespaces ns = new Namespaces(Namespaces.INSTANCE);
		ns.registerPrefix("meta", "http://ontoware.org/meta/2006/04#");
		query = reasoner.createQuery(ns, "SELECT ?x ?y WHERE { ?x meta:isSubClassOf ?y }");
		query.open();
		printer.verbose(query.getNumberOfTuples());
		while (!query.afterLast()) {
			Term[] answer = query.tupleBuffer();
			Term subclass = answer[0];
			Term superclass = answer[1];
			if ((Individual.class.isInstance(subclass)) && (Individual.class.isInstance(superclass))) {
				Individual sub = (Individual)subclass;
				Individual sup = (Individual)superclass;
				String ssubclass = demetaURI(sub, metaontology);
				String ssuperclass = demetaURI(sup, metaontology);
		  		Axiom axiom = KAON2Manager.factory().subClassOf(KAON2Manager.factory().owlClass(ssubclass), KAON2Manager.factory().owlClass(ssuperclass));
		  		printer.verbose("Added axiom: " + axiom.toString());
		  		ontology.addAxiom(axiom);				
			}
			printer.verbose("-");
			printer.verbose(subclass.toString());
			printer.verbose(superclass.toString());
			printer.verbose(".");
			
			query.next();
		}
		query.close();
		query.dispose();

		query = reasoner.createQuery(metaobjectpropertydomain);
		query.open();
		printer.verbose(query.getNumberOfTuples());
		while (!query.afterLast()) {
			Term[] answer = query.tupleBuffer();
			for (Term a : answer) {
				if (Individual.class.isInstance(a)) {
					Individual axiom = (Individual)a;
					ontology.addAxioms(demetaObjectPropertyDomain(axiom, reasoner));
				}
			}
			query.next();
		}
		query.close();
		query.dispose();

		query = reasoner.createQuery(metaobjectpropertyrange);
		query.open();
		printer.verbose(query.getNumberOfTuples());
		while (!query.afterLast()) {
			Term[] answer = query.tupleBuffer();
			for (Term a : answer) {
				if (Individual.class.isInstance(a)) {
					Individual axiom = (Individual)a;
					ontology.addAxioms(demetaObjectPropertyRange(axiom, reasoner));
				}
			}
			query.next();
		}
		query.close();
		query.dispose();

		reasoner.dispose();
	}
	
	/**
	 * Chooses the appropriate meta-class for a given concrete entity
	 * @param entity Concrete entity
	 * @return Meta-Class
	 */
	private OWLClass chooseClass(Entity entity) {
		if (OWLClass.class.isInstance(entity)) return metaclass;
		if (Individual.class.isInstance(entity)) return metaindividual;
		if (DataProperty.class.isInstance(entity)) return metadataproperty;
		if (ObjectProperty.class.isInstance(entity)) return metaobjectproperty;
		if (AnnotationProperty.class.isInstance(entity)) return metaannotationproperty;
		if (Datatype.class.isInstance(entity)) return metadatatype;
		//if (NonOWLPredicate.class.isInstance(entity)) return metanonowlpredicate;
		return metaentity;
	}
	
	/**
	 * Chooses an appropriate meta-class for a given concrete description
	 * @param description Concrete description
	 * @return Meta-Class
	 */
	private OWLClass chooseClass(Description description) {
  		List<OntologyChangeEvent> changes = new ArrayList<OntologyChangeEvent>();
		//Axiom maxiom = KAON2Manager.factory().classMember(axiomclass, individual); 
		//changes.add(new OntologyChangeEvent(maxiom, OntologyChangeEvent.ChangeType.ADD));
		if (DataAll.class.isInstance(description)) return metadataall;
		if (DataCardinality.class.isInstance(description)) return metadatacardinality;
		if (DataHasValue.class.isInstance(description)) return metadatahasvalue;
		if (DataSome.class.isInstance(description)) return metadatasome;
		if (ObjectAll.class.isInstance(description)) return metaobjectall;
		if (ObjectAnd.class.isInstance(description)) return metaobjectand;
		if (ObjectCardinality.class.isInstance(description)) return metaobjectcardinality;
		if (ObjectHasValue.class.isInstance(description)) return metaobjecthasvalue;
		if (ObjectNot.class.isInstance(description)) return metaobjectnot;
		if (ObjectOneOf.class.isInstance(description)) return metaobjectoneof;
		if (ObjectOr.class.isInstance(description)) return metaobjector;
		if (ObjectSome.class.isInstance(description)) return metaobjectsome;
		if (Entity.class.isInstance(description)) return chooseClass((Entity)description);
		return metadescription;
	}
	
	/**
	 * The axioms to be created from one concrete axiom
	 * @param axiom Concrete axiom
	 * @return Meta-axioms
	 */
	private List<OntologyChangeEvent> meta(Axiom axiom) {
		Individual individual = KAON2Manager.factory().individual(newURI());
  		List<OntologyChangeEvent> changes = new ArrayList<OntologyChangeEvent>();

  		if (Annotation.class.isInstance(axiom)) {
  			Annotation annotation = (Annotation)axiom;
  			Axiom maxiom = KAON2Manager.factory().classMember(metaannotation, individual); 
  			changes.add(new OntologyChangeEvent(maxiom, OntologyChangeEvent.ChangeType.ADD));
  			// annotation property
			Individual annotationproperty = KAON2Manager.factory().individual(metaURI(annotation.getAnnotationProperty().getURI()));
  			maxiom = KAON2Manager.factory().objectPropertyMember(metaannotationpropertyinstance, annotationproperty, individual);
  			changes.add(new OntologyChangeEvent(maxiom, OntologyChangeEvent.ChangeType.ADD));
  			// TODO ann entity
  			// TODO ann value
  			return changes;
  		}
		if (ClassMember.class.isInstance(axiom)) {
			ClassMember cm = (ClassMember)axiom;
			Axiom maxiom = KAON2Manager.factory().classMember(metaclassmember, individual);
			changes.add(new OntologyChangeEvent(maxiom, OntologyChangeEvent.ChangeType.ADD));
			Description d = cm.getDescription();
			Axiom maxiomclass = KAON2Manager.factory().objectPropertyMember(metaclassmemberclass, individual, meta(d));
			changes.add(new OntologyChangeEvent(maxiomclass, OntologyChangeEvent.ChangeType.ADD));
			Individual i = cm.getIndividual();
			Axiom maxiomindividual = KAON2Manager.factory().objectPropertyMember(metaclassmemberindividual, individual, meta(i));
			changes.add(new OntologyChangeEvent(maxiomindividual, OntologyChangeEvent.ChangeType.ADD));
			Axiom maxiomhasclassmember = KAON2Manager.factory().objectPropertyMember(metaismemberof, meta(i), meta(d));
			changes.add(new OntologyChangeEvent(maxiomhasclassmember, OntologyChangeEvent.ChangeType.ADD));
			return changes;
		}
		if (DataPropertyAttribute.class.isInstance(axiom)) {			
			DataPropertyAttribute dpa = (DataPropertyAttribute)axiom;
			Individual dataproperty = meta(dpa.getDataProperty());
			Axiom maxiom = KAON2Manager.factory().classMember(metafunctionaldataproperty,dataproperty);
			changes.add(new OntologyChangeEvent(maxiom, OntologyChangeEvent.ChangeType.ADD));			
			return changes;
		}
		if (DataPropertyDomain.class.isInstance(axiom)) {
			DataPropertyDomain dpd = (DataPropertyDomain)axiom;
			Axiom maxiom = KAON2Manager.factory().classMember(metadatapropertydomain, individual);
			changes.add(new OntologyChangeEvent(maxiom, OntologyChangeEvent.ChangeType.ADD));
			Individual dataproperty = meta(dpd.getDataProperty());
			Axiom maxiomdataproperty = KAON2Manager.factory().objectPropertyMember(metadatapropertydomaindataproperty, individual, dataproperty);
			changes.add(new OntologyChangeEvent(maxiomdataproperty, OntologyChangeEvent.ChangeType.ADD));
			Description d = dpd.getDomain();
			Axiom maxiomdomain = KAON2Manager.factory().objectPropertyMember(metadatapropertydomaindomain, individual, meta(d));
			changes.add(new OntologyChangeEvent(maxiomdomain, OntologyChangeEvent.ChangeType.ADD));
			return changes;
		}
// TODO		if (DataPropertyMember.class.isInstance(axiom)) return metadatapropertymember;
		if (DataPropertyRange.class.isInstance(axiom)) {
			DataPropertyRange dpr = (DataPropertyRange)axiom;
			Axiom maxiom = KAON2Manager.factory().classMember(metadatapropertyrange, individual);
			changes.add(new OntologyChangeEvent(maxiom, OntologyChangeEvent.ChangeType.ADD));
			Individual dataproperty = meta(dpr.getDataProperty());
			Axiom maxiomdataproperty = KAON2Manager.factory().objectPropertyMember(metadatapropertyrangedataproperty, individual, dataproperty);
			changes.add(new OntologyChangeEvent(maxiomdataproperty, OntologyChangeEvent.ChangeType.ADD));
			DataRange datarange = dpr.getRange();
			if (DataNot.class.isInstance(datarange)) {
				DataNot datanot = (DataNot)datarange;
				Individual negatedDatarange = KAON2Manager.factory().individual(newURI());
				Axiom maxiomrange = KAON2Manager.factory().objectPropertyMember(metadatapropertyrangerange, individual, negatedDatarange);
				changes.add(new OntologyChangeEvent(maxiomrange, OntologyChangeEvent.ChangeType.ADD));
				// TODO negation of the datarange
				return changes;
			}
			if (Datatype.class.isInstance(datarange)) {
				Datatype datatype = (Datatype)datarange;
				Axiom maxiomrange = KAON2Manager.factory().objectPropertyMember(metadatapropertyrangerange, individual, meta(datatype));
				changes.add(new OntologyChangeEvent(maxiomrange, OntologyChangeEvent.ChangeType.ADD));
				return changes;
			}
			if (DataOneOf.class.isInstance(datarange)) {
				// TODO define one of datarange
				return changes;
			}
		}
// TODO		if (Deprecation.class.isInstance(axiom)) return metadeprecation;
// TODO		if (DifferentIndividuals.class.isInstance(axiom)) return metadifferentindividuals;
// TODO 	if (DisjointClasses.class.isInstance(axiom)) return metadisjointclasses;
// TODO		if (EquivalentClasses.class.isInstance(axiom)) return metaequivalentclasses;
// TODO		if (EquivalentDataProperties.class.isInstance(axiom)) return metaequivalentdataproperties;
// TODO		if (EquivalentObjectProperties.class.isInstance(axiom)) return metaequivalentobjectproperties;
// TODO		if (InverseObjectProperties.class.isInstance(axiom)) return metainverseobjectproperties;
// TODO		if (ObjectPropertyAttribute.class.isInstance(axiom)) return metaobjectpropertyattribute;
		if (ObjectPropertyDomain.class.isInstance(axiom)) {
			ObjectPropertyDomain opd = (ObjectPropertyDomain)axiom;
			Axiom maxiom = KAON2Manager.factory().classMember(metaobjectpropertydomain, individual);
			changes.add(new OntologyChangeEvent(maxiom, OntologyChangeEvent.ChangeType.ADD));
			Individual objectproperty = meta(opd.getObjectProperty());
			Axiom maxiomobjectproperty = KAON2Manager.factory().objectPropertyMember(metaobjectpropertydomainobjectproperty, individual, objectproperty);
			changes.add(new OntologyChangeEvent(maxiomobjectproperty, OntologyChangeEvent.ChangeType.ADD));
			Description d = opd.getDomain();
			Axiom maxiomdomain = KAON2Manager.factory().objectPropertyMember(metaobjectpropertydomaindomain, individual, meta(d));
			changes.add(new OntologyChangeEvent(maxiomdomain, OntologyChangeEvent.ChangeType.ADD));
			return changes;
		}
// TODO		if (ObjectPropertyMember.class.isInstance(axiom)) return metaobjectpropertymember;
// TODO		if (ObjectPropertyRange.class.isInstance(axiom)) return metaobjectpropertyrange;
// TODO		if (SameIndividual.class.isInstance(axiom)) return metasameindividual;
		if (SubClassOf.class.isInstance(axiom)) {
			SubClassOf sco = (SubClassOf)axiom;
			Axiom maxiom = KAON2Manager.factory().classMember(metasubclassof, individual);
			changes.add(new OntologyChangeEvent(maxiom, OntologyChangeEvent.ChangeType.ADD));
			Description sup = sco.getSuperDescription();
			Axiom maxiomsup = KAON2Manager.factory().objectPropertyMember(metasubclassofsuperclass, individual, meta(sup));
			changes.add(new OntologyChangeEvent(maxiomsup, OntologyChangeEvent.ChangeType.ADD));
			Description sub = sco.getSubDescription();
			Axiom maxiomsub = KAON2Manager.factory().objectPropertyMember(metasubclassofsubclass, individual, meta(sub));
			changes.add(new OntologyChangeEvent(maxiomsub, OntologyChangeEvent.ChangeType.ADD));
			return changes;
		}
// TODO		if (SubDataPropertyOf.class.isInstance(axiom)) return metasubdatapropertyof;
// TODO		if (SubObjectPropertyOf.class.isInstance(axiom)) return metasubobjectpropertyof;

		// TODO warning should be issued default faultback
		Axiom maxiom = KAON2Manager.factory().classMember(metaaxiom, individual); 
		changes.add(new OntologyChangeEvent(maxiom, OntologyChangeEvent.ChangeType.ADD));
		return changes;
	}
	
	private Individual meta(Description description) {
		if (OWLClass.class.isInstance(description)) {
			OWLClass clazz = (OWLClass)description;
			return KAON2Manager.factory().individual(metaURI(clazz.getURI()));
		}
		return KAON2Manager.factory().individual(newURI()); // TODO
	}
	
	private Individual meta(DataProperty dataproperty) {
		return KAON2Manager.factory().individual(metaURI(dataproperty.getURI()));
	}
	
	private Individual meta(Datatype datatype) {
		return KAON2Manager.factory().individual(metaURI(datatype.getURI()));
	}
	
	private Individual meta(Individual individual) {
		return KAON2Manager.factory().individual(metaURI(individual.getURI()));
	}
	
	private Individual meta(ObjectProperty objectproperty) {
		return KAON2Manager.factory().individual(metaURI(objectproperty.getURI()));
	}
	
	/**
	 * Demetas a single meta-axiom. Needs the metaontology as a context.
	 * @param metaaxiom name of the meta-axiom to be demeta'd
	 * @param reasoner reasoner over the meta-ontology
	 * @return axioms that represent the de-meta'd meta-axioms
	 */
	private List<Axiom> demeta(Individual maxiom, Reasoner reasoner) throws KAON2Exception {
  		List<Axiom> axioms = new ArrayList<Axiom>();
  		Axiom axiom = KAON2Manager.factory().classMember(metaaxiom, maxiom);
  		if (reasoner.getOntology().containsAxiom(axiom, true))
  			axioms.addAll(demetaClassMember(maxiom, reasoner));
  		return axioms;
	}
	
	private List<Axiom> demetaClassMember(Individual maxiom, Reasoner reasoner) throws KAON2Exception {
		Ontology o = reasoner.getOntology();
  		List<Axiom> axioms = new ArrayList<Axiom>();
  		Individual mclass = maxiom.getObjectPropertyValue(o, metaclassmemberclass);
  		if (mclass == null) {
  			printer.warning("ClassMember axiom " + maxiom.toString() + " without class.");
  			return axioms;
  		}
  		String sclass = mclass.getAnnotationValue(o, refers).toString();
  		Individual mindividual = maxiom.getObjectPropertyValue(o, metaclassmemberindividual);
  		if (mindividual == null) {
  			printer.warning("ClassMember axiom " + maxiom.toString() + " without class.");
  			return axioms;
  		}
  		String sindividual = mindividual.getAnnotationValue(o, refers).toString();
  		Axiom axiom = KAON2Manager.factory().classMember(KAON2Manager.factory().owlClass(sclass), KAON2Manager.factory().individual(sindividual));
  		printer.verbose("Added axiom: " + axiom.toString());
  		axioms.add(axiom);
  		return axioms;
	}

	private List<Axiom> demetaSubClassOf(Individual maxiom, Reasoner reasoner) throws KAON2Exception {
		Ontology o = reasoner.getOntology();
  		List<Axiom> axioms = new ArrayList<Axiom>();
  		Individual msubclass = maxiom.getObjectPropertyValue(o, metasubclassofsubclass);
  		if (msubclass == null) {
  			printer.warning("SubClassOf axiom " + maxiom.toString() + " without subclass.");
  			return axioms;
  		}
  		String ssubclass = demetaURI(msubclass, o);
  		Individual msuperclass = maxiom.getObjectPropertyValue(o, metasubclassofsuperclass);
  		if (msuperclass == null) {
  			printer.warning("SubClassOf axiom " + maxiom.toString() + " without superclass.");
  			return axioms;
  		}
  		String ssuperclass = demetaURI(msuperclass, o);
  		Axiom axiom = KAON2Manager.factory().subClassOf(KAON2Manager.factory().owlClass(ssubclass), KAON2Manager.factory().owlClass(ssuperclass));
  		printer.verbose("Added axiom: " + axiom.toString());
  		axioms.add(axiom);
  		return axioms;
	}

	private List<Axiom> demetaObjectPropertyDomain(Individual maxiom, Reasoner reasoner) throws KAON2Exception {
		Ontology o = reasoner.getOntology();
  		List<Axiom> axioms = new ArrayList<Axiom>();
  		Individual mproperty = maxiom.getObjectPropertyValue(o, metaobjectpropertydomainobjectproperty);
  		if (mproperty == null) {
  			printer.warning("ObjectPropertyDomain axiom " + maxiom.toString() + " without object property.");
  			return axioms;
  		}
  		String sproperty = mproperty.getAnnotationValue(o, refers).toString();
  		Individual mdomain = maxiom.getObjectPropertyValue(o, metaobjectpropertydomaindomain);
  		if (mdomain == null) {
  			printer.warning("ObjectPropertyDomain axiom " + maxiom.toString() + " without domain.");
  			return axioms;
  		}
  		String sdomain = mdomain.getAnnotationValue(o, refers).toString();
  		Axiom axiom = KAON2Manager.factory().objectPropertyDomain(KAON2Manager.factory().objectProperty(sproperty), KAON2Manager.factory().owlClass(sdomain));
  		printer.verbose("Added axiom: " + axiom.toString());
  		axioms.add(axiom);
  		return axioms;
	}

	private List<Axiom> demetaObjectPropertyRange(Individual maxiom, Reasoner reasoner) throws KAON2Exception {
		Ontology o = reasoner.getOntology();
  		List<Axiom> axioms = new ArrayList<Axiom>();
  		Individual mproperty = maxiom.getObjectPropertyValue(o, metaobjectpropertyrangeobjectproperty);
  		if (mproperty == null) {
  			printer.warning("ObjectPropertyRange axiom " + maxiom.toString() + " without object property.");
  			return axioms;
  		}
  		String sproperty = mproperty.getAnnotationValue(o, refers).toString();
  		Individual mrange = maxiom.getObjectPropertyValue(o, metaobjectpropertyrangerange);
  		if (mrange == null) {
  			printer.warning("ObjectPropertyRange axiom " + maxiom.toString() + " without range.");
  			return axioms;
  		}
  		String srange = mrange.getAnnotationValue(o, refers).toString();
  		Axiom axiom = KAON2Manager.factory().objectPropertyRange(KAON2Manager.factory().objectProperty(sproperty), KAON2Manager.factory().owlClass(srange));
  		printer.verbose("Added axiom: " + axiom.toString());
  		axioms.add(axiom);
  		return axioms;
	}

	/**
	 * Creates a meta-URI out of a concrete one
	 * @param uri Concrete URI
	 * @return Meta-URI
	 */
	private String metaURI(String uri) {
		return uri.substring(0, 7) + "meta." + uri.substring(7);
	}
	
	/**
	 * Demetas a URI
	 */
	private String demetaURI(OWLEntity thing, Ontology o) throws KAON2Exception {
		Object refer = thing.getAnnotationValue(o, refers);
		if (refer != null)
			return refer.toString();
		else
			return thing.getURI();		
	}
	
	/**
	 * Createas a new URI
	 * @return new URI
	 */
	static String newURI() {
		// TODO make a more elaborate algorithm
		return "http://www.example.com/axiom#a" + Math.abs(m_random.nextInt());
	}
		
	/**
	 * prints the command line parameters and a short info
	 */
	public static void printUsage() {
		System.out.println("Usage: owl meta <input_url> [<output_filename>] [-verbose] [-rdf|-xml] [-no]");
		System.out.println("");
		System.out.println("meta makes a meta-ontology out of an ontology, and the other way around.");
		System.out.println("This tool is still beta and only paritall finished"); // TODO
		System.out.println("");
		System.out.println(" <input_url>");
		System.out.println("     The location of the ontology to be read.");
		System.out.println("");
		System.out.println(" <output_filename>");     
		System.out.println("     The name of the file where the meta-ontology will be written to.");
		System.out.println("     If the filename is omitted, the meta-ontology is added to the ontology.");
		System.out.println("");
		System.out.println(" -verbose or simply -verb");     
		System.out.println("     If this switch is set, you will get additional output while processing.");
		System.out.println("");
		System.out.println(" -rdf / -xml / -as");     
		System.out.println("     These switches decide the format of the output ontology: either");
		System.out.println("     RDF/XML-serialisation (default), OWL/XML Presentation Syntax or");
		System.out.println("     Abstract Syntax");
		System.out.println("");
		System.out.println(" -demeta or simply -de");     
		System.out.println("     Turn the metaontology into an ontology.");
		System.out.println("");
		System.out.println(" -ref / -noref");     
		System.out.println("     Adds annotations from metaelements to concrete ones (or not)");
	}	
	
}
