/*
 * owlrdf2owlxml is a small and simple tool to use KAON2 in order
 * to change an OWL ontology from XML/RDF-serialisation into one in
 * OWL/XML syntax (and back)
 */
package edu.unika.aifb.owltools.tools;

import java.io.*;
import java.io.StringReader;
import java.util.*;
import org.semanticweb.kaon2.api.*;

import edu.unika.aifb.owltools.utils.Printer;

/**
 * owlrdf2owlxml calls KAON2 in order to load an OWL ontology
 * in XML/RDF-serialisation and to export it in OWL/XML syntax
 * (or the other way around)
 */
public class Owlrdf2owlxml extends KAON2Tool {
	public String name() { return "owlrdf2owlxml"; }

	String m_fileuri = "";
	String m_output = "";
	org.semanticweb.kaon2.api.OntologyFileFormat m_format = org.semanticweb.kaon2.api.OntologyFileFormat.OWL_XML; 

	/**
	 * only used in the Online-Version
	 */
	static int count = 0;
	
	public static void main(String[] args) {
		Owlrdf2owlxml r2x = new Owlrdf2owlxml();
		if (!r2x.processArguments(args)) return;
		r2x.compile();
		r2x.printer.close();
	}

	/**
	 * Only used in the online version
	 */
	static String createReturnString(Printer printer) {
		String s = printer.getOutput();
		if (printer.error.equals(""))
			return s;
		String errorheader = "<html><h2>Error</h2><pre>";
		errorheader += printer.error;
		errorheader += "</pre><br />An error occured. Maybe you should try to <a href=\"http://phoebus.cs.man.ac.uk:9999/OWL/Validator\">validate your OWL ontology</a>, to see if it is faulty. If it is not, please contact <a href=\"mailto:denny@aifb.uni-karlsruhe.de\">denny</a> with a copy of your ontology.<br />";
		return errorheader + "</html>"; 
	}
	
	/**
	 * Only used in the online version
	 */
	public static String compileUri(String uri, String dir) {
		Printer printer = new Printer();
		printer.setString();
		printer.html = true;
		count += 1;
		try {
			String filename = dir + "/uri" + String.valueOf(count) + ".uri";
			BufferedWriter out = new BufferedWriter(new FileWriter(filename, false));
			out.write(uri);
			out.close();
		} catch(IOException e) {
			printer.error(e.toString());
		}
		Owlrdf2owlxml r2x = new Owlrdf2owlxml();
		r2x.m_format = org.semanticweb.kaon2.api.OntologyFileFormat.OWL_XML;
		r2x.m_fileuri = uri;
		r2x.compile();
		printer.close();
		return createReturnString(printer);
	}
	
	/**
	 * Only used in the online version
	 */
	public static String compileString(String string, String dir) {
        Printer printer = new Printer();
        printer.html = true;
        printer.setString();
        count += 1;
		try {
			String filename = dir + "/owl" + String.valueOf(count) + ".xml";
			BufferedWriter out = new BufferedWriter(new FileWriter(filename, false));
			out.write(string);
			out.close();
		} catch(IOException e) {
			printer.error(e.toString());
		}
		try {
			// construct ontology from string
			KAON2Connection connection = KAON2Manager.newConnection();
			String uri = "http://www.aifb.uni-karlsruhe.de/WBD/dvr/internal";
	        DefaultOntologyResolver resolver=new DefaultOntologyResolver();
        	resolver.registerReplacement(uri, "file:temp.owl");
	        connection.setOntologyResolver(resolver);
			Ontology ontology = connection.createOntology(uri, new HashMap<String,Object>());
			
			// import String into ontology
			Reader text = new StringReader(string);
			ontology.importContentsFrom(text, uri);
			
			// compile it the usual way
			ontology.saveOntology(org.semanticweb.kaon2.api.OntologyFileFormat.OWL_XML, printer.writer, "UTF-8");
			
			connection.close();
		} catch (KAON2Exception e) {
			printer.error(e.toString());
		} catch (InterruptedException i) {
			printer.error(i.toString());
		} catch (IOException io) {
			printer.error(io.toString());
		}
		// TODO refactor this method and the compile method
		return createReturnString(printer);
	}

	public boolean processArguments(String[] args) {
		// process command line arguments
		
		int argc = args.length;
		if (argc == 0) {
			printUsage();
			return false;
		}
		String s = args[0].toLowerCase();
		if (argc == 1) {
			if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) || (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) || (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help")) || (s.equalsIgnoreCase("owlrdf2owlxml")) || (s.equalsIgnoreCase("r2x"))) {
				printUsage();
				return false;
			}
		}

		boolean hasFileName = false;
		boolean hasOutput = false;
		boolean hasFormat = false;
		
		for (int i = 0; i < argc; ++i) {
			boolean unknownArg = true;
			
			String arg = args[i];

			if ((arg.equalsIgnoreCase("-?")) || (arg.equalsIgnoreCase("?")) || (arg.equalsIgnoreCase("-h")) || (arg.equalsIgnoreCase("--help")) || (arg.equalsIgnoreCase("-help")) || (arg.equalsIgnoreCase("help"))) {
				printUsage();
				return false;
			}

			if (arg.startsWith("-")) {
				
				// Format
				if (arg.equalsIgnoreCase("-rdf")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = org.semanticweb.kaon2.api.OntologyFileFormat.OWL_RDF;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-xml")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = org.semanticweb.kaon2.api.OntologyFileFormat.OWL_XML;
					hasFormat = true;
					unknownArg = false;
				}
			} else {
				if (arg.equalsIgnoreCase("owlrdf2owlxml") || arg.equalsIgnoreCase("r2x")) {
					unknownArg = false;
				} else {
					if (hasFileName) {
						if (hasOutput) { printUsageError(); return false; }
						m_output = arg;
						hasOutput = true;
						unknownArg = false;
					} else {
						m_fileuri = arg;
						hasFileName = true;					
						unknownArg = false;
					}
				}
			}
			
			if (unknownArg) { printUsageError(); return false; } 
		}
		return true;
	}
	
	public void compile() {
		try {
	        if (!m_output.equals("")) printer.setFile(m_output);
	        try {
	        	Ontology ontology = loadOntology(m_fileuri);
    			ontology.saveOntology(m_format, printer.writer, "UTF-8");
	        } catch (KAON2Exception e) {
				printer.error("Could not open ontology " + m_fileuri);
				printer.error(e.toString());
			} catch (InterruptedException i) {
				printer.error("Could not export ontology " + m_fileuri);
				printer.error(i.toString());
			} catch (IOException io) {
				printer.error("Could not export ontology " + m_fileuri);
				printer.error(io.toString());
			}
			finally {
				endConnection();
			}
		} catch (KAON2Exception e) {
			printer.error("Could not open or close KAON2 connection.");
			printer.error(e.toString());
		}
	}

	/**
	 * prints the command line parameters and a short info
	 */
	public static void printUsage() {
		System.out.println("Usage: owl r2x <input_url> [<output_filename>] [-rdf|-xml]");
		System.out.println("");
		System.out.println("owlrdf2owlxml converts OWL ontologies from XML/RDF-serialisation into");
		System.out.println("OWL XML presentation syntax and back.");
		System.out.println("");
		System.out.println(" <input_url>");
		System.out.println("     The location of the ontology to be read.");
		System.out.println("");
		System.out.println(" <output_filename>");     
		System.out.println("     The name of the file where the transformation will be written to.");
		System.out.println("     If the filename is omitted, the output is send to the console.");
		System.out.println("");
		System.out.println(" -rdf   the output is in OWL XML presentation syntax (default)");
		System.out.println(" -xml   the output is in RDF/XML-serialisation syntax");
	}
}
