/*
 * Created on 24.03.2005
 *
 * Merge adds all axioms of the second ontology to the first one.
 */
package edu.unika.aifb.owltools.tools;

import java.io.*;
import java.util.*;
import org.semanticweb.kaon2.api.*;
import org.semanticweb.kaon2.api.formatting.*;

/**
 * Returns an ontology conaining all axioms from both ontologies
 */
public class Merge extends KAON2Tool {
	/**
	 * the name of the tool
	 */
	public String name() { return "merge"; }

	/**
	 * The URI of the first ontology
	 */
	private String m_fileuriOne = "";
	
	/**
	 * The URI of the second ontology
	 */
	private String m_fileuriTwo = "";
	
	/**
	 * The filename of the output ontology
	 */
	private String m_output = "";
	
	/**
	 * The serialization format for the output ontology
	 */
	private String m_format = OntologyFileFormat.OWL_RDF;
	
	/**
	 * If false, no result will be generated 
	 */
	private boolean m_result = true;

	/**
	 * To be called when used from a standalone command line tool
	 * @param args Command line parameters
	 */
	public static void main(String[] args) {
		Merge d = new Merge();
		if (!d.processArguments(args)) return;
		d.compile();
		d.printer.close();
	}

	/**
	 * process command line arguments
	 */
	public boolean processArguments(String[] args) {
		
		int argc = args.length;
		if (argc == 0) {
			printUsage();
			return false;
		}
		String s = args[0].toLowerCase();
		if (argc == 1) {
		if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) || (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) || (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help")) || (s.equalsIgnoreCase("merge"))) {
				printUsage();
				return false;
			}
		}

		boolean hasFileName1 = false;
		boolean hasFileName2 = false;
		boolean hasOutput = false;
		boolean hasVerbose = false;
		boolean hasFormat = false;
		boolean hasResult = false;
		
		for (int i = 0; i < argc; ++i) {
			boolean unknownArg = true;
			
			String arg = args[i];

			if ((arg.equalsIgnoreCase("-?")) || (arg.equalsIgnoreCase("?")) || (arg.equalsIgnoreCase("-h")) || (arg.equalsIgnoreCase("--help")) || (arg.equalsIgnoreCase("-help")) || (arg.equalsIgnoreCase("help"))) {
				printUsage();
				return false;
			}
			
			if (arg.startsWith("-")) {
				if ((arg.equalsIgnoreCase("-verbose")) || (arg.equalsIgnoreCase("-verb"))) {
					if (hasVerbose) { printUsageError(); return false; }
					printer.verbose = true;
					hasVerbose = true;
					unknownArg = false;
				}					

				if ((arg.equalsIgnoreCase("-noresult")) || (arg.equalsIgnoreCase("-no"))) {
					if (hasResult) { printUsageError(); return false; }
					m_result = false;
					hasResult = true;
					unknownArg = false;
				}					
				
				// Format
				if (arg.equalsIgnoreCase("-rdf")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.OWL_RDF;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-xml")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.OWL_XML;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-as") || arg.equalsIgnoreCase("-abstract")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.FUNCTIONAL_SYNTAX;
					hasFormat = true;
					unknownArg = false;
				}
			} else {
				if (arg.equalsIgnoreCase("merge")) {
					unknownArg = false;
				} else {
					if (hasFileName1) {
						if (hasFileName2) {
							if (hasOutput) { printUsageError(); return false; }
							m_output = arg;
							hasOutput = true;
							unknownArg = false;
						} else {
							m_fileuriTwo = arg;
							hasFileName2 = true;					
							unknownArg = false;
						}
					} else {
						m_fileuriOne = arg;
						hasFileName1 = true;					
						unknownArg = false;
					}
				}
			}
			
			if (unknownArg) { printUsageError(); return false; }
		}
		
		if (!hasFileName2) { printUsageError(); return false; }
		return true;
	}
	
	/**
	 * Entry point for the command line based usage of the Filter owl tool
	 *
	 */
	public void compile() {
		try {
	        if (!m_output.equals("")) printer.setFile(m_output);

	        try {
	        	Ontology base = loadOntology(m_fileuriOne);
	        	Ontology diff = loadOntology(m_fileuriTwo);
	        	Ontology result = merge(base, diff);
	        	
	        	if (m_result)
	        		result.saveOntology(m_format, printer.writer, "UTF-8");

	        } catch (KAON2Exception e) {
				printer.error("Could not open ontology.");
				printer.error(e.toString());
			} catch (IOException io) {
				printer.error("Could not save resulting ontology " + m_output);
				printer.error(io.toString());
			} catch (InterruptedException i) {
				printer.error("Could not save resulting ontology.");
				printer.error(i.toString());				
			}
			finally {
				endConnection();
			}
		} catch (KAON2Exception e) {
			printer.error("Could not open or close KAON2 connection.");
			printer.error(e.toString());
		}
	}
	

	/**
	 * Merges two ontologies
	 * @param one One of the ontologies to be merged
	 * @param two The second ontology to be merged
	 * @return The merged ontology
	 * @throws KAON2Exception
	 */
	public Ontology merge(Ontology one, Ontology two) throws KAON2Exception {
		Ontology result = null;
		result = createOntology("http://www.aifb.uni-karlsruhe.de/WBS/dvr/owltools/merge/result");
		merge(one, two, result);
		return result;
	}
	
	/**
	 * Merges two ontologies by simply taking each axiom and adding it to
	 * a resulting ontology
	 * @param one The first ontology
	 * @param two The second ontology
	 * @param result The resulting, merged ontology
	 * @throws KAON2Exception
	 */
	public void merge(Ontology one, Ontology two, Ontology result) throws KAON2Exception {
  		List<OntologyChangeEvent> changes = new ArrayList<OntologyChangeEvent>();
  		
  		Cursor<Axiom> cursor = one.createAxiomRequest().openCursor();
  		while (cursor.hasNext()) {
  			Axiom axiom = cursor.next();
  			if (!result.containsAxiom(axiom, false)) {
		        changes.add(new OntologyChangeEvent(axiom, OntologyChangeEvent.ChangeType.ADD));	  				
  				printer.verbose("Added: " + axiom.toString());
  			} else {
  				printer.verbose("Already included: " + axiom.toString());	  				
  			}
  		}
  		result.applyChanges(changes);
  		changes = new ArrayList<OntologyChangeEvent>();
  		cursor = two.createAxiomRequest().openCursor();
  		while (cursor.hasNext()) {
  			Axiom axiom = cursor.next();
  			if (!result.containsAxiom(axiom, false)) {
		       changes.add(new OntologyChangeEvent(axiom, OntologyChangeEvent.ChangeType.ADD));
		       printer.verbose("Added: " + axiom.toString());
  			} else {
  				printer.verbose("Already included: " + axiom.toString());	  				
  			}
  		}
  		
  		result.applyChanges(changes);
	}
	
	/**
	 * prints the command line parameters and a short info
	 */
	public static void printUsage() {
		System.out.println("Usage: owl merge <url1> <url2> [<output_filename>] [-verbose] [-rdf|-xml] [-no]");
		System.out.println("");
		System.out.println("merge creates one ontology with the axioms of the two given ontologies");
		System.out.println("");
		System.out.println(" <url1>");
		System.out.println("     The location of the base ontology");
		System.out.println(" <url2>");
		System.out.println("     The location of the compared ontology");
		System.out.println(" <output_filename>");     
		System.out.println("     The name of the file where the merged ontology will be written to.");
		System.out.println("     If omitted, the result is sent to the console.");
		System.out.println("");
		System.out.println(" -verbose or simply -verb");     
		System.out.println("     If this switch is set, you will get additional output while processing.");
		System.out.println("");
		System.out.println(" -rdf / -xml / -as");     
		System.out.println("     These switches decide the format of the output ontology: either");
		System.out.println("     RDF/XML-serialisation (default), OWL/XML Presentation Syntax or");
		System.out.println("     Abstract Syntax");
		System.out.println("");
		System.out.println(" -noresult or simply -no");     
		System.out.println("     Doesn't save or show the result.");
	}	
}
