package edu.unika.aifb.owltools.tools;

import java.util.*;
import java.util.concurrent.TimeUnit;
import java.io.File;

import edu.unika.aifb.foam.rules.feature.objectproperty.ObjectPropertyInverseProperties;
import edu.unika.aifb.owltools.*;
import edu.unika.aifb.owltools.utils.*;

import org.semanticweb.kaon2.api.*;
import org.semanticweb.kaon2.api.owl.axioms.*;
import org.semanticweb.kaon2.api.owl.elements.*;
import org.semanticweb.kaon2.api.reasoner.*;
import org.semanticweb.kaon2.api.logic.*;
import org.semanticweb.kaon2.api.formatting.*;
// import org.semanticweb.kaon2.util.trace.*;
//import edu.unika.aifb.owltools.evaluation.*;
import java.io.*;

/**
 * This is an experimental class. I use it only for my own scripts.
 * If you like, look into it, but as a user -- just forget about it.
 */
public class Script extends KAON2Tool {
	public String name() { return "Script"; }

	static int m_counter = 0;
	static String ns = "http://aifb.de/";
	
	String m_fileuri = "";
	String m_output = "";
	private String m_format = OntologyFileFormat.OWL_RDF;
	boolean m_result = true;
	boolean m_screech = false;
	boolean m_checksat = false;
	boolean m_instances = false;
	int m_arg = -1;
	
	Predicate getPredicate(Ontology ontology) throws KAON2Exception {
  		Set<OWLClass> predicates = ontology.createEntityRequest(OWLClass.class).get();
  		int n = predicates.size();
  		System.out.println("Number of classes: " + n);
  		Predicate[] pres = new Predicate[n];
  		
  		pres = predicates.toArray(pres);
  		Random random = new Random();
		if (m_arg == -1) m_arg =  random.nextInt(n);
		System.out.println("Choosen class #: " + m_arg);
  		return pres[m_arg];
	}
	
	public Ontology script(Ontology ontology) throws KAON2Exception, InterruptedException {
		Reasoner reasoner = ontology.createReasoner();
		long start;
		long end;
		if (m_checksat) {
			System.out.print("Satisfiability: ");
			boolean sat = true;
			start = System.currentTimeMillis();
			sat = reasoner.isSatisfiable();
			end = System.currentTimeMillis();
			if (sat)
				System.out.print("yes, ");
			else
				System.out.print("no, ");
			System.out.println((end-start) + " ms");
		}
		
		Predicate p = getPredicate(ontology);
  		
  		System.out.println("Choosen class name: " + p.toString());
  		Query query;
  		//reasoner.setParameter("datalogMonitor", new SaveExecutionPlan(new File("C:\\temp\\fastafter.plan")));
		

  		try {
  	  		query = reasoner.createQuery(p);
  			System.out.print("Normal query: ");
  			start = System.currentTimeMillis();
  	  		query.open();
  			end = System.currentTimeMillis();
  			System.out.print(end-start);
  			System.out.println(" ms");
  			int n = query.getNumberOfTuples();
  			System.out.println("Found instances: " + n);
  			if (m_instances) {
  				while (!query.afterLast()) {
  		            Object[] tupleBuffer=query.tupleBuffer();
  		            printer.putline(tupleBuffer[0].toString());
  		            query.next();
  		        }  				
  			}
  			query.close();
  			query.dispose();
  		} finally {
  		}
		return ontology;
	}
	
	public Ontology script1(Ontology ontology) throws KAON2Exception, InterruptedException {
		long start;
		long end;
		Ontology screechont;
		System.out.print("Screech Conversion: ");
		DLPConvert dlp = new DLPConvert();
		start = System.currentTimeMillis();
		screechont = dlp.screech(ontology);
		end = System.currentTimeMillis();
		System.out.println(", " + (end-start) + " ms");
		Reasoner screechreasoner = screechont.createReasoner();
		if (m_checksat) {
			System.out.print("Screech Satisfiability: ");
			boolean sat;
			start = System.currentTimeMillis();
			sat = screechreasoner.isSatisfiable();
			end = System.currentTimeMillis();
			if (sat)
				System.out.print("yes, ");
			else
				System.out.print("no, ");
			System.out.println((end-start) + " ms");
		}
  		Predicate p = getPredicate(ontology);
  		
  		System.out.println("Choosen class name: " + p.toString());
  		Query query;

  		try {
  	  		query = screechreasoner.createQuery(p);
  			System.out.print("Screech query: ");
  			start = System.currentTimeMillis();
  	  		query.open();
  			end = System.currentTimeMillis();
  			System.out.print(end-start);
  			System.out.println(" ms");
  			int n = query.getNumberOfTuples();
  			System.out.println("Found instances: " + n);
  			if (m_instances) {
  				while (!query.afterLast()) {
  		            Object[] tupleBuffer=query.tupleBuffer();
  		            printer.putline(tupleBuffer[0].toString());
  		            query.next();
  		        }  				
  			}
  			query.close();
  			query.dispose();
  		} finally {
  		}
		return ontology;
	}
	
	public int getLevel(Ontology ontology, OWLClass clazz) throws KAON2Exception {
		int level = 1;
		Set<Description> superclasses = clazz.getSuperDescriptions(ontology);
		for (Description superclass : superclasses) {
			if (OWLClass.OWL_THING == superclass.toString()) return 1;
			if (OWLClass.class.isInstance(superclass)) {
				int l = getLevel(ontology, (OWLClass)superclass);
				if ((1==level)||((l+1)<level))
					level = l+1;
			}
			//System.out.println(superclass.toString());
		}
		return level;
	}
	
	public void makeTaxon(Map<String,Integer> taxon, Ontology ontology, OWLClass clazz, int level) throws KAON2Exception {
		if (taxon.containsKey(clazz.getURI()))
			if (level > taxon.get(clazz.getURI()))
				return;				
		taxon.put(clazz.getURI(), level);
		Set<Description> superclasses = clazz.getSuperDescriptions(ontology);
		for (Description superclass : superclasses) {
			if (OWLClass.class.isInstance(superclass)) {
				makeTaxon(taxon, ontology, (OWLClass)superclass, level+1);
			}
		}
		
	}
	
	boolean m_ancestor = false;
	
	public int getDistance(Ontology ontology, OWLClass start, OWLClass goal) throws KAON2Exception {
		m_ancestor = false;
		Map<String, Integer> taxon1 = new HashMap<String,Integer>();
		makeTaxon(taxon1, ontology, start, 0);
		Map<String, Integer> taxon2 = new HashMap<String,Integer>();
		makeTaxon(taxon2, ontology, goal, 0);
		int distance = 999;
		if (taxon2.containsKey(start.toString()))
				m_ancestor = true;
		if (taxon1.containsKey(goal.toString()))
				m_ancestor = true;
		for (String key : taxon1.keySet()) {
			if (taxon2.containsKey(key)) {
				int d = taxon1.get(key) + taxon2.get(key);
				if (d<distance) distance = d;
			}
		}
		//for (String key : taxon1.keySet()) System.out.println(key + taxon1.get(key));
		//for (String key : taxon2.keySet()) System.out.println(key + taxon2.get(key));
		
		return distance;
	}
	
	int m_level[] = new int[20];
	int m_distance[] = new int[20];
	int m_siblings = 0;
	int m_nonsiblings = 0;
	int m_ancestors = 0;
	int m_nonancestors = 0;
	
	public boolean getPair(Ontology ontology) throws KAON2Exception {
		boolean counts = true;
  		Set<OWLClass> predicates = ontology.createEntityRequest(OWLClass.class).get();
  		int n = predicates.size();
  		//System.out.println("Number of classes: " + n);
  		Predicate[] pres = new Predicate[n];
  		pres = predicates.toArray(pres);
  		Random random = new Random();
		int firstnr =  random.nextInt(n);
		int secondnr =  random.nextInt(n);
		while (firstnr == secondnr)
			secondnr =  random.nextInt(n);
		int l1 = getLevel(ontology, (OWLClass)pres[firstnr]);
		int l2 = getLevel(ontology, (OWLClass)pres[secondnr]);
		int d = getDistance(ontology, (OWLClass)pres[firstnr], (OWLClass)pres[secondnr]);
		if (999==d)
			d = l1 + l2 + 2;
		//counts = ((d==2)&&!m_ancestor);
		if (counts) {
			printer.putline("URI: " + pres[firstnr].toString());
			printer.putline("Level: " + l1);
			printer.putline("URI: " + pres[secondnr].toString());
			printer.putline("Level: " + l2);
			printer.putline("Distance: " + d);
			printer.put("Ancestor: ");
			if (m_ancestor) {
				printer.putline("true");
				m_ancestors++;
			} else {
				printer.putline("false");
				m_nonancestors++;
			}
			printer.put("Sibling: ");
			if ((d==2)&&!m_ancestor) {
				printer.putline("true");
				m_siblings++;
			} else {
				printer.putline("false");
				m_nonsiblings++;
			}
			if (l1<20) m_level[l1]++;
			if (l2<20) m_level[l2]++;
			if (d<20) m_distance[d]++;
		}
		
		return counts;
		
	}
	
	public void queryTop(Ontology ontology) throws KAON2Exception, InterruptedException {
		Reasoner reasoner = ontology.createReasoner();
		System.out.print("Satisfiability: ");
		boolean sat = reasoner.isSatisfiable();
		if (sat)
			System.out.println("yes");
		else
			System.out.println("no");

		Predicate p = KAON2Manager.factory().thing();
  		
  		System.out.println("Choosen class name: " + p.toString());
  		Query query;

  		query = reasoner.createQuery(p);
  		query.open();
		int n = query.getNumberOfTuples();
		System.out.println("Found instances: " + n);
		while (!query.afterLast()) {
            Object[] tupleBuffer=query.tupleBuffer();
            System.out.println(tupleBuffer[0].toString());
            query.next();
        }  				
		query.close();
		query.dispose();
	}

	public boolean processArguments(String[] args) {
		// process command line arguments
		
		int argc = args.length;
		if (argc == 0) {
			printUsage();
			return false;
		}
		String s = args[0].toLowerCase();
		if (argc == 1) {
			if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) || (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) || (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help")) || (s.equalsIgnoreCase("deo"))) {
				printUsage();
				return false;
			}
		}

		boolean hasFileName = false;
		boolean hasOutput = false;
		boolean hasVerbose = false;
		boolean hasFormat = false;
		boolean hasResult = false;
		
		for (int i = 0; i < argc; ++i) {
			boolean unknownArg = true;
			
			String arg = args[i];
			
			if ((arg.equalsIgnoreCase("-?")) || (arg.equalsIgnoreCase("?")) || (arg.equalsIgnoreCase("-h")) || (arg.equalsIgnoreCase("--help")) || (arg.equalsIgnoreCase("-help")) || (arg.equalsIgnoreCase("help"))) {
				printUsage();
				return false;
			}

			if (arg.startsWith("-")) {
				if (java.lang.Character.isDigit(arg.charAt(1))) {
					try {
						m_arg = java.lang.Integer.valueOf(arg.substring(1));
					} catch (NumberFormatException n) {
						printUsageError();
						return false;
					}
					unknownArg = false;
				}
				
				if ((arg.equalsIgnoreCase("-verbose")) || (arg.equalsIgnoreCase("-verb"))) {
					if (hasVerbose) { printUsageError(); return false; }
					printer.verbose = true;
					hasVerbose = true;
					unknownArg = false;
				}					

				if ((arg.equalsIgnoreCase("-noresult")) || (arg.equalsIgnoreCase("-no"))) {
					if (hasResult) { printUsageError(); return false; }
					m_result = false;
					hasResult = true;
					unknownArg = false;
				}
				
				if (arg.equalsIgnoreCase("-sat")) {
					m_checksat = true;
					unknownArg = false;
				}
				
				if (arg.equalsIgnoreCase("-screech")) {
					m_screech = true;
					unknownArg = false;
				}
				
				if (arg.equalsIgnoreCase("-instances")) {
					m_instances = true;
					unknownArg = false;
				}

				// Format
				if (arg.equalsIgnoreCase("-rdf")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.OWL_RDF;
					hasFormat = true;
					unknownArg = false;
				}
				if (arg.equalsIgnoreCase("-xml")) {
					if (hasFormat) { printUsageError(); return false; }
					m_format = OntologyFileFormat.OWL_XML;
					hasFormat = true;
					unknownArg = false;
				}
			} else {
				if (arg.equalsIgnoreCase("x")) {
					unknownArg = false;
				} else {
					if (hasFileName) {
						if (hasOutput) { printUsageError(); return false; }
						m_output = arg;
						hasOutput = true;
						unknownArg = false;
					} else {
						m_fileuri = arg;
						hasFileName = true;					
						unknownArg = false;
					}
				}
			}
			
			if (unknownArg) { printUsageError(); return false; } 
		}
		return true;
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		Script x = new Script();
		if (!x.processArguments(args)) return;
		x.compile();
		x.printer.close();
	}

	public Ontology scriptNominal(Ontology o) throws KAON2Exception {
		return deo(o);
	}

	static void verbose(String message) {
		System.out.println(message);
//		printer.verbose(message);
	}

	static Description processDescription(Description description, Ontology ontology, List<OntologyChangeEvent> changes, OWLClass owlClass) throws KAON2Exception {
		Description result = null;
		
		if (ObjectOneOf.class.isInstance(description)) {
			ObjectOneOf nominal = (ObjectOneOf)description;
			if (owlClass == null)
				owlClass = KAON2Manager.factory().owlClass(ns + "Nominal" );
			
			OWLClass proxy = KAON2Manager.factory().owlClass(ns + "Proxy");
			
			Set<Individual> individuals = nominal.getIndividuals();
			int number = individuals.size();

			OWLClass thing = KAON2Manager.factory().thing();
			
			Description[] ecs = new Description[2];
			ecs[0] = proxy;
			ecs[1] = thing;
			EquivalentClasses ec = KAON2Manager.factory().equivalentClasses(ecs);
			verbose("Adding: " + ec.toString());
	        changes.add(new OntologyChangeEvent(ec, OntologyChangeEvent.ChangeType.ADD));
			
			ObjectProperty counter = KAON2Manager.factory().objectProperty(ns + "counter");
			ObjectPropertyDomain domain = KAON2Manager.factory().objectPropertyDomain(counter, proxy);
			verbose("Adding: " + domain.toString());
	        changes.add(new OntologyChangeEvent(domain, OntologyChangeEvent.ChangeType.ADD));
	        
	        ObjectProperty invcounter = KAON2Manager.factory().objectProperty(ns + "invcounter");
	        InverseObjectProperties invcounterinverse = KAON2Manager.factory().inverseObjectProperties(counter, invcounter);
			verbose("Adding: " + invcounterinverse.toString());
	        changes.add(new OntologyChangeEvent(invcounterinverse, OntologyChangeEvent.ChangeType.ADD));
	        
			ObjectCardinality cardinv = KAON2Manager.factory().objectCardinality(1, 1, invcounter, proxy);
			Description[] invdescs = new Description[2];
			invdescs[0] = cardinv;
			invdescs[1] = owlClass;			
			EquivalentClasses invequi = KAON2Manager.factory().equivalentClasses(invdescs);
			verbose("Adding: " + invequi.toString());
	        changes.add(new OntologyChangeEvent(invequi, OntologyChangeEvent.ChangeType.ADD));
	        
			ObjectPropertyRange range = KAON2Manager.factory().objectPropertyRange(counter, owlClass);
			verbose("Adding: " + range.toString());
	        changes.add(new OntologyChangeEvent(range, OntologyChangeEvent.ChangeType.ADD));
	        
			ObjectCardinality card = KAON2Manager.factory().objectCardinality(number, number, counter, owlClass);
			Description[] descs = new Description[2];
			descs[0] = card;
			descs[1] = proxy;			
			EquivalentClasses equi = KAON2Manager.factory().equivalentClasses(descs);
			verbose("Adding: " + equi.toString());
	        changes.add(new OntologyChangeEvent(equi, OntologyChangeEvent.ChangeType.ADD));
	        
	        Individual petrus = KAON2Manager.factory().individual(ns + "petrus");
	        ClassMember petrusMember = KAON2Manager.factory().classMember(proxy, petrus);
			verbose("Adding: " + petrusMember.toString());
	        changes.add(new OntologyChangeEvent(petrusMember, OntologyChangeEvent.ChangeType.ADD));

	        Individual t = KAON2Manager.factory().individual(ns + "t");
	        ClassMember tMember = KAON2Manager.factory().classMember(proxy, t);
			verbose("Adding: " + tMember.toString());
	        changes.add(new OntologyChangeEvent(tMember, OntologyChangeEvent.ChangeType.ADD));

	        Individual[] i2 = new Individual[3];
	        int j = 0;
	        for (Individual i : individuals)
	        	i2[j++] = i;
	        //i2[j++] = t;
	        i2[j] = petrus;
	        DifferentIndividuals allunique = KAON2Manager.factory().differentIndividuals(i2);
			verbose("Adding: " + allunique.toString());
	        changes.add(new OntologyChangeEvent(allunique, OntologyChangeEvent.ChangeType.ADD));
	        
	        // Add appropriate rule: same(petrus, X) :- Proxy(X)
	        Variable X = KAON2Manager.factory().variable("X");
	        Term[] Xarray = new Term[1];
	        Xarray[0] = X;
	        Term[] petrusAndX = new Term[2];
	        petrusAndX[0] = petrus;
	        petrusAndX[1] = X;
	        Predicate eq = KAON2Manager.factory().equal();
	        Literal head = KAON2Manager.factory().literal(true, eq, petrusAndX);
	        Literal body[] = new Literal[1];
	        Literal bodyLit = KAON2Manager.factory().literal(true, proxy, Xarray);
	        body[0] = bodyLit;
	        Rule fixRule = KAON2Manager.factory().rule(head, body);
			verbose("Adding: " + fixRule.toString());
	        changes.add(new OntologyChangeEvent(fixRule, OntologyChangeEvent.ChangeType.ADD));
	        
	        //Variable X = KAON2Manager.factory().variable("X");
	        //Term[] Xarray = new Term[1];
	        //Xarray[0] = X;
	        //Term[] petrusAndX = new Term[2];
	        //petrusAndX[0] = petrus;
	        //petrusAndX[1] = X;
	        //Predicate eq = KAON2Manager.factory().equal();
	        //Literal head = KAON2Manager.factory().literal(true, eq, petrusAndX);
	        Literal body2[] = new Literal[1];
	        Literal bodyLit2 = KAON2Manager.factory().literal(true, thing, Xarray);
	        body2[0] = bodyLit2;
	        Rule fixRule2 = KAON2Manager.factory().rule(head, body2);
			verbose("Adding: " + fixRule2.toString());
	        changes.add(new OntologyChangeEvent(fixRule2, OntologyChangeEvent.ChangeType.ADD));

	        for (Individual individual : individuals) {
				ClassMember classMember = KAON2Manager.factory().classMember(owlClass, individual);
		        ObjectPropertyMember counterMember = KAON2Manager.factory().objectPropertyMember(counter, petrus, individual);
	        	if (!ontology.containsAxiom(classMember, false)) {
					verbose("Adding: " + counterMember.toString());
			        changes.add(new OntologyChangeEvent(counterMember, OntologyChangeEvent.ChangeType.ADD));
					verbose("Adding: " + classMember.toString());
			        changes.add(new OntologyChangeEvent(classMember, OntologyChangeEvent.ChangeType.ADD));
	        	} else {
	        		//verbose("Already in: " + classMember.toString());
	        	}
			}
			result = owlClass;
		}
		
		if (ObjectHasValue.class.isInstance(description)) {
			ObjectHasValue hasValue = (ObjectHasValue)description;

			owlClass = KAON2Manager.factory().owlClass(ns + "Nominal");
			Individual individual = hasValue.getIndividual();
			
			ClassMember classMember = KAON2Manager.factory().classMember(owlClass, individual);
        	if (!ontology.containsAxiom(classMember, false)) {
				verbose("Adding: " + classMember.toString());
		        changes.add(new OntologyChangeEvent(classMember, OntologyChangeEvent.ChangeType.ADD));
        	} else {
        		verbose("Already in: " + classMember.toString());
	        }

	        ObjectAll objectAll = KAON2Manager.factory().objectAll(hasValue.getObjectProperty(), owlClass);
	        result = objectAll;
		}
		
		if (ObjectAll.class.isInstance(description)) {
			ObjectAll all = (ObjectAll)description;
			
			Description allFrom = processDescription(all.getDescription(), ontology, changes, null);
			if (allFrom!=null) {
				ObjectAll objectAll = KAON2Manager.factory().objectAll(all.getObjectProperty(), allFrom);
				result = objectAll;
			}
		}
		
		if (ObjectSome.class.isInstance(description)) {
			ObjectSome some = (ObjectSome)description;
			
			Description someFrom = processDescription(some.getDescription(), ontology, changes, null);
			if (someFrom!=null) {
				ObjectSome objectSome = KAON2Manager.factory().objectSome(some.getObjectProperty(), someFrom);
				result = objectSome;
			}
		}
		
		if (ObjectCardinality.class.isInstance(description)) {
			ObjectCardinality cardinality = (ObjectCardinality)description;
			
			Description cardinalityOn = processDescription(cardinality.getDescription(), ontology, changes, null);
			if (cardinalityOn!=null) {
				ObjectCardinality objectCardinality = KAON2Manager.factory().objectCardinality(cardinality.getCardinalityType(), cardinality.getCardinality(), cardinality.getObjectProperty(), cardinalityOn);
				result = objectCardinality;
			}
		}
		
		if (ObjectAnd.class.isInstance(description)) {
			ObjectAnd and = (ObjectAnd)description;
			
			Set<Description> descriptions = and.getDescriptions();
			Set<Description> newDescriptions = new HashSet<Description>();
			boolean unchanged = true;
			for (Description element : descriptions) {
				Description processed = processDescription(element, ontology, changes, null);
				if (processed == null) {
					newDescriptions.add(element);
				} else {
					newDescriptions.add(processed);
					unchanged = false;
				}
			}
			ObjectAnd objectAnd = KAON2Manager.factory().objectAnd(newDescriptions);
			if (!unchanged)
				result = objectAnd;
		}
		
		if (ObjectOr.class.isInstance(description)) {
			ObjectOr or = (ObjectOr)description;
			
			Set<Description> descriptions = or.getDescriptions();
			Set<Description> newDescriptions = new HashSet<Description>();
			boolean unchanged = true;
			for (Description element : descriptions) {
				Description processed = processDescription(element, ontology, changes, null);
				if (processed == null) {
					newDescriptions.add(element);
				} else {
					newDescriptions.add(processed);
					unchanged = false;
				}
			}
			ObjectOr objectOr = KAON2Manager.factory().objectOr(newDescriptions);
			if (!unchanged)
				result = objectOr;
		}
		
		if (ObjectNot.class.isInstance(description)) {
			ObjectNot not = (ObjectNot)description;
			
			Description isNot = processDescription(not.getDescription(), ontology, changes, null);
			if (isNot!=null) {
				ObjectNot objectNot = KAON2Manager.factory().objectNot(isNot);
				result = objectNot;
			}
		}
		
		
		return result;
	}
	
	static public Ontology deo(Ontology ontology) throws KAON2Exception {
		// look for everything that has a description
		
  		List<OntologyChangeEvent> changes=new ArrayList<OntologyChangeEvent>();
  		
  		Cursor<ClassMember> classCursor = ontology.createAxiomRequest(ClassMember.class).openCursor();
  		while (classCursor.hasNext()) {
  			ClassMember a = classCursor.next();
  			Description c = a.getDescription();
	  		Description result = processDescription(c, ontology, changes, null);
	  		if (result!=null) {
				ClassMember classMember = KAON2Manager.factory().classMember(result, a.getIndividual());
		        changes.add(new OntologyChangeEvent(classMember, OntologyChangeEvent.ChangeType.ADD));
		        changes.add(new OntologyChangeEvent(a, OntologyChangeEvent.ChangeType.REMOVE));
		        verbose("Adding: " + classMember.toString());
		        verbose("Deleting: " + a.toString());
	  		}
  		}
  		classCursor.close();

  		Cursor<DataPropertyDomain> dataPropertyDomainCursor = ontology.createAxiomRequest(DataPropertyDomain.class).openCursor();
  		while (dataPropertyDomainCursor.hasNext()) {
  			DataPropertyDomain a = dataPropertyDomainCursor.next();
  			Description c = a.getDomain();
	  		Description result = processDescription(c, ontology, changes, null);
	  		if (result!=null) {
				DataPropertyDomain dataPropertyDomain = KAON2Manager.factory().dataPropertyDomain(a.getDataProperty(), result);
		        changes.add(new OntologyChangeEvent(dataPropertyDomain, OntologyChangeEvent.ChangeType.ADD));
		        changes.add(new OntologyChangeEvent(a, OntologyChangeEvent.ChangeType.REMOVE));
		        verbose("Adding: " + dataPropertyDomain.toString());
		        verbose("Deleting: " + a.toString());
	  		}
  		}
  		dataPropertyDomainCursor.close();

  		Cursor<ObjectPropertyDomain> objectPropertyDomainCursor = ontology.createAxiomRequest(ObjectPropertyDomain.class).openCursor();
  		while (objectPropertyDomainCursor.hasNext()) {
  			ObjectPropertyDomain a = objectPropertyDomainCursor.next();
  			Description c = a.getDomain();
	  		Description result = processDescription(c, ontology, changes, null);
	  		if (result!=null) {
				ObjectPropertyDomain objectPropertyDomain = KAON2Manager.factory().objectPropertyDomain(a.getObjectProperty(), result);
		        changes.add(new OntologyChangeEvent(objectPropertyDomain, OntologyChangeEvent.ChangeType.ADD));
		        changes.add(new OntologyChangeEvent(a, OntologyChangeEvent.ChangeType.REMOVE));
		        verbose("Adding: " + objectPropertyDomain.toString());
		        verbose("Deleting: " + a.toString());
	  		}
  		}
  		objectPropertyDomainCursor.close();

  		Cursor<ObjectPropertyRange> objectPropertyRangeCursor = ontology.createAxiomRequest(ObjectPropertyRange.class).openCursor();
  		while (objectPropertyRangeCursor.hasNext()) {
  			ObjectPropertyRange a = objectPropertyRangeCursor.next();
  			Description c = a.getRange();
	  		Description result = processDescription(c, ontology, changes, null);
	  		if (result!=null) {
				ObjectPropertyRange objectPropertyRange = KAON2Manager.factory().objectPropertyRange(a.getObjectProperty(), result);
		        changes.add(new OntologyChangeEvent(objectPropertyRange, OntologyChangeEvent.ChangeType.ADD));
		        changes.add(new OntologyChangeEvent(a, OntologyChangeEvent.ChangeType.REMOVE));
		        verbose("Adding: " + objectPropertyRange.toString());
		        verbose("Deleting: " + a.toString());
	  		}
  		}
  		objectPropertyRangeCursor.close();

  		Cursor<SubClassOf> subClassOfCursor = ontology.createAxiomRequest(SubClassOf.class).openCursor();
  		while (subClassOfCursor.hasNext()) {
  			SubClassOf a = subClassOfCursor.next();
  			Description c = a.getSubDescription();
	  		Description sub = processDescription(c, ontology, changes, null);
  			c = a.getSuperDescription();
	  		Description sup = processDescription(c, ontology, changes, null);
	  		if ((sub!=null)||(sup!=null)) {
	  			if (sub==null) sub = a.getSubDescription();
	  			if (sup==null) sup = a.getSuperDescription();
				SubClassOf subClassOf = KAON2Manager.factory().subClassOf(sub, sup);
		        changes.add(new OntologyChangeEvent(subClassOf, OntologyChangeEvent.ChangeType.ADD));
		        changes.add(new OntologyChangeEvent(a, OntologyChangeEvent.ChangeType.REMOVE));
		        verbose("Adding: " + subClassOf.toString());
		        verbose("Deleting: " + a.toString());
	  		}
  		}
  		subClassOfCursor.close();

  		Cursor<DisjointClasses> disjointClassesCursor = ontology.createAxiomRequest(DisjointClasses.class).openCursor();
  		while (disjointClassesCursor.hasNext()) {
  			DisjointClasses a = disjointClassesCursor.next();
  			Set<Description> cs = a.getDescriptions();
  			Set<Description> newCs = new HashSet<Description>();
  			boolean hasNominal = false;
  			for (Description c : cs) {
		  		Description result = processDescription(c, ontology, changes, null);
		  		if (result==null) {
		  			newCs.add(c);
		  		} else {
		  			newCs.add(result);
		  			hasNominal = true;
		  		}
  			}
  			if (hasNominal) {
				DisjointClasses disjointClasses = KAON2Manager.factory().disjointClasses(newCs);
		        changes.add(new OntologyChangeEvent(disjointClasses, OntologyChangeEvent.ChangeType.ADD));
		        changes.add(new OntologyChangeEvent(a, OntologyChangeEvent.ChangeType.REMOVE));
		        verbose("Adding: " + disjointClasses.toString());
		        verbose("Deleting: " + a.toString());
  			}
  		}
  		disjointClassesCursor.close();

  		Cursor<EquivalentClasses> equivalentClassesCursor = ontology.createAxiomRequest(EquivalentClasses.class).openCursor();
  		while (equivalentClassesCursor.hasNext()) {
  			EquivalentClasses a = equivalentClassesCursor.next();
  			Set<Description> cs = a.getDescriptions();
  			Set<Description> newCs = new HashSet<Description>();
  			boolean hasNominal = false;
  			OWLClass simpleClass = null;

  			//for (Description c : cs) {
  			//	if (OWLClass.class.isInstance(c))
  			//		simpleClass = (OWLClass)c;
  			//}
  			
  			for (Description c : cs) {
		  		Description result = processDescription(c, ontology, changes, simpleClass);
		  		if (result==null) {
		  			newCs.add(c);
		  		} else {
		  			if (simpleClass == null)
		  				newCs.add(result);
		  			hasNominal = true;
		  		}
  			}
  			if (hasNominal) {
  				if (newCs.size() > 1) {
					EquivalentClasses equivalentClasses = KAON2Manager.factory().equivalentClasses(newCs);
			        changes.add(new OntologyChangeEvent(equivalentClasses, OntologyChangeEvent.ChangeType.ADD));
			        verbose("Adding: " + equivalentClasses.toString());
  				}
		        changes.add(new OntologyChangeEvent(a, OntologyChangeEvent.ChangeType.REMOVE));
		        verbose("Deleting: " + a.toString());
  			}
  		}
  		equivalentClassesCursor.close();
    	
        ontology.applyChanges(changes);
 	
	  	return ontology;
	}

	static void copy( InputStream in, OutputStream out ) throws IOException 
	  { 
	    byte[] buffer = new byte[ 0xFFFF ]; 
	 
	    for ( int len; (len = in.read(buffer)) != -1; ) 
	      out.write( buffer, 0, len ); 
	  } 
	 
	  static void copyFile( String src, String dest ) 
	  { 
	    FileInputStream  fis = null; 
	    FileOutputStream fos = null; 
	 
	    try 
	    { 
	      fis = new FileInputStream( src ); 
	      fos = new FileOutputStream( dest ); 
	 
	      copy( fis, fos ); 
	    } 
	    catch ( IOException e ) { 
	      e.printStackTrace(); 
	    } 
	    finally { 
	      if ( fis != null ) 
	        try { fis.close(); } catch ( IOException e ) { } 
	      if ( fos != null ) 
	        try { fos.close(); } catch ( IOException e ) { } 
	    } 
	  } 	
	
	int ok = 0;
	int notok = 0;
	int allcount = 0;
	int s1 = 0;
	int s2 = 0;
	int s3 = 0;
	int s4 = 0;
	int s5 = 0;
	int max = 0;
	int min = 10;
	int ecount = 0;
	int emax = 0;
	int icount = 0;
	int imax = 0;
	int tbox = 0;
	int satcount = 0;
	int unsatcount = 0;
	int undecidedcount = 0;
	
	Set<Axiom> results;
	
	public void filevisitor(File f) {
		if ( f.isFile() ) {
			try {
				String filename = f.getName();
				String name = f.toURI().toString();
				printer.print(f.getCanonicalPath());
				Ontology ontology = loadOntology( name );
				int count = api.count(ontology);
				int entitycount = api.countEntities(ontology);
				int individualcount = api.countEntities(ontology, Individual.class);
				TimedReasoner t = new TimedReasoner();
				t.set(ontology, printer);
				t.start();
				TimeUnit.SECONDS.sleep(10);
				if (!t.decided()) {
					t.interrupt();
					undecidedcount++;
					printer.print("Stopped satisfiability check");
				} else {
					boolean sat = t.satisfiable();
					if (sat) {
						printer.print("yes, ");
						satcount++;
					} else {
						printer.print("no, ");
						unsatcount++;
					}
					printer.print(t.time() + " nanoseconds");
				}
				endConnection();
				if (t.time() > 0) {
			        DataProperty p = KAON2Manager.factory().dataProperty("http://owltools.ontoware.org/measure#sattime");
			        Individual o = KAON2Manager.factory().individual("http://owltools.ontoware.org/watson/" + filename);
				    Constant v = KAON2Manager.factory().constant(t.time());
					Axiom axiom = KAON2Manager.factory().dataPropertyMember(p, o, v);
					results.add(axiom);
				}
				printer.print("#" + ok);
				printer.print("Satisfiable count: " + satcount);
				printer.print("Unsatisfiable count: " + unsatcount);
				printer.print("Undecided: " + undecidedcount);
				printer.print( "Count: " + count );
				printer.print( "Entity count: " + entitycount );
				printer.print( "Individual count: " + individualcount );
				
				if ((individualcount == 0) && (entitycount > 0)) tbox = tbox+1;
				ecount = ecount + entitycount;
				icount = icount + individualcount;
				allcount = allcount + count;
				ok++;
				if (count > max) max = count;
				if (count < min) min = count;
				//String newname = "E:\\ontologies\\watson\\save\\" + f.getName() + ".owl";
				//copyFile(f.getCanonicalPath(), newname);
				printer.print("All Axiom #: " + allcount);
				printer.print("Avg Axiom #: " + allcount / ok);
				if (count<10) s1++;
				else if (count<100) s2++;
				else if (count<1000) s3++;
				else if (count<10000) s4++;
				else s5 = s5+1;
				printer.print ( "Sizedistr: " + s1 + "/" + s2 + "/" + s3 + "/" + s4 + "/" + s5 );
				printer.print ( "Sizerange: " + min + "-" + max);
				printer.print ("Pure TBoxes: " + tbox);
				printer.print("# Entities: " + ecount);
				printer.print("# Individuals: " + icount);
			} catch (Exception e) {
				printer.error(e.toString());
			}
			return;
		}
		File[] entries = f.listFiles();
		for ( File entry : entries ) {
			filevisitor( entry );
		}
	}
	
	private long takeMeasure(Ontology o) throws KAON2Exception {
		return api.count(o, EquivalentClasses.class);
	}
	
	private long takeSatTimeMeasure(Ontology o) throws KAON2Exception, InterruptedException {
		TimedReasoner t = new TimedReasoner();
		t.set(o, printer);
		t.start();
		int loop = 0;
		while (!t.decided() && (loop++<30))
			TimeUnit.SECONDS.sleep(1);
		if (!t.decided()) {
			t.interrupt();
			printer.error("Stopped satisfiability check");
		} 
		return t.time();
	}
	
	private long takePredicateTimeMeasure(Ontology o) throws KAON2Exception, InterruptedException {
		TimedReasoner t = new TimedReasoner();
		t.set(o, printer);
		t.start();
		int loop = 0;
		while (!t.decided() && (loop++<30))
			TimeUnit.SECONDS.sleep(1);
		if (!t.decided()) {
			t.interrupt();
			printer.error("Stopped satisfiability check");
		} 
		return t.time();
	}
	
	String rv;
	Ontology o;
	void add(int i) {
		rv = rv + " " + i;
	}
	void addcount(Class c) throws KAON2Exception {
		add(api.count(o, c));
	}
	void addcounte(Class c) throws KAON2Exception {
		add(api.countEntities(o, c));
	}
	
	public void compile() {
		rv = "";
		try {
			//id axiom entity annotation classmember datapropertymember datapropertydomain objectpropertymember objectpropertydomain objectpropertyrange disjointclasses equivalentclasses subclassof inverseobjectproperties class individual objectproperty dataproperty sameindividual subobjectproperty subdataproperty 
        	o = loadOntology(m_fileuri);
            add(api.count(o));
            add(api.countEntities(o));
            addcount(EntityAnnotation.class);
            addcount(ClassMember.class);
            addcount(DataPropertyMember.class);
            addcount(DataPropertyDomain.class);
            addcount(ObjectPropertyMember.class); 
            addcount(ObjectPropertyDomain.class);
            addcount(ObjectPropertyRange.class);
            addcount(DisjointClasses.class);
            addcount(EquivalentClasses.class);
            addcount(SubClassOf.class);
            addcount(InverseObjectProperties.class);
            addcounte(OWLClass.class);
            addcounte(Individual.class);
            addcounte(ObjectProperty.class);
            addcounte(DataProperty.class);
            addcount(SameIndividual.class);
            addcount(SubObjectPropertyOf.class);
            addcount(SubDataPropertyOf.class);
        	FileWriter f = new FileWriter("features.txt", true);
        	f.write(m_fileuri + rv + '\n');
        	f.flush();
		} catch (IOException e) {
        	printer.error(e.toString());
        } catch (InterruptedException e) {
        	printer.error(e.toString());
        } catch (KAON2Exception e) {
        	printer.error(e.toString());
		}        
	}
	
	public void compile4() {
        try {
        	Ontology ontology = loadOntology(m_fileuri);   	
        	long res = takeSatTimeMeasure(ontology);
        	printer.print(res);
        	if (res > 0) {
            	Ontology result = loadOntology("result.rdf");
    	        DataProperty p = KAON2Manager.factory().dataProperty("http://owltools.ontoware.org/measure#sattime");
    	        Individual o = KAON2Manager.factory().individual("http://owltools.ontoware.org/watson/" + m_fileuri);
    		    Constant v = KAON2Manager.factory().constant(res);
    			Axiom axiom = KAON2Manager.factory().dataPropertyMember(p, o, v);
    			result.addAxiom(axiom);
        		printer.setFile("result.rdf");
       			result.saveOntology(m_format, printer.writer, "UTF-8");
        	}
        	System.exit(0);
        } catch (IOException e) {
        	printer.error(e.toString());
        } catch (InterruptedException e) {
        	printer.error(e.toString());
        } catch (KAON2Exception e) {
        	printer.error(e.toString());
        }
		
	}
	
	public void compile3() {
		results = new HashSet<Axiom>();
		File f = new File( "." );
		try {
	        filevisitor(f);
			KAON2Connection resultconnection = KAON2Manager.newConnection();
			OntologyResolver resolver = resultconnection.getOntologyResolver();
			if (DefaultOntologyResolver.class.isInstance(resolver)) {
				DefaultOntologyResolver res = (DefaultOntologyResolver)resolver;
				res.registerReplacement("http://owltools.ontoware.org/watsonmeasure/result", "file:result.owl");
		        resolver = res;
			}
			resultconnection.setOntologyResolver(resolver);

	        Ontology result = resultconnection.createOntology("http://owltools.ontoware.org/watsonmeasure/result", new HashMap<String, Object>());
	        result.addAxioms(results);

    		printer.setFile("result.owl");
	        result.saveOntology(m_format, printer.writer, "UTF-8");
			resultconnection.close();
		} catch (KAON2Exception e) {
			printer.error(e.toString());
		} catch (InterruptedException e) {
			printer.error(e.toString());
		} catch (IOException e) {
			printer.error(e.toString());
		}
		printer.print("OK:     " + ok);
		printer.print("Not OK: " + notok);
		printer.print("Count:  " + allcount);
	}
	
	public void compile2() {
		try {
	        if (!m_output.equals("")) printer.setFile(m_output);

	        
	        
	        try {

	        	Ontology ontology = loadOntology(m_fileuri);
	        	
/*	        	for (int i=0; i<2000; i++) {
	        		printer.putline("Pair " + i);
		        	while (!getPair(ontology));
	        		printer.putline("");
	        	}
	        	for (int i=0; i<20; i++) {
	        		if (m_level[i]>0) printer.putline("Occurence level " + i + ": " + m_level[i]);
	        	}
	        	printer.putline("");
	        	for (int i=0; i<20; i++) {
	        		if (m_distance[i]>0) printer.putline("Occurence distance " + i + ": " + m_distance[i]);
	        	}
	        	printer.putline("");
	        	System.out.println(m_siblings);
	        	printer.putline("# siblings      : " + m_siblings);
	        	printer.putline("# non-siblings  : " + m_nonsiblings);
	        	printer.putline("# ancestors     : " + m_ancestors);
	        	printer.putline("# non-ancestors : " + m_nonancestors);
*/
	
	        	ontology.saveOntology(m_format, printer.writer, "UTF-8");
	        	
/*	        	if (m_screech)
	        		ontology = script1(ontology);
	        	else
	        		ontology = script(ontology);
*/
	        	
	        	// ontology = scriptNominal(ontology);
	        	
	        	//queryTop(ontology);
	        	
	        	//Evaluation e = new AllLabelled(ontology);
	        	//System.out.println(e.getEvaluation());
	        	
	        	
//	        	Literal[] body = new Literal[5];
//	        	Term[] headarg = new Term[1];
//	        	Term[] bodyargs0 = new Term[2];
//	        	Term[] bodyargs1 = new Term[2];
//	        	Term[] bodyargs2 = new Term[2];
//	        	Term[] bodyargs3 = new Term[2];
//	        	Term[] bodyargs4 = new Term[2];
//	        	Variable g = KAON2Manager.factory().variable("G");
//	        	Variable s = KAON2Manager.factory().variable("S");
//	        	Variable x = KAON2Manager.factory().variable("X");
//	        	Variable n = KAON2Manager.factory().variable("N");
//	        	Predicate nationality = KAON2Manager.factory().objectProperty("http://www.aifb.uni-karlsruhe.de/WBS/dvr/rove/rove#nationality");
//	        	Predicate hasMember = KAON2Manager.factory().objectProperty("http://www.aifb.uni-karlsruhe.de/WBS/dvr/rove/rove#hasMember");
//	        	Predicate differentFrom = KAON2Manager.factory().objectProperty("http://www.w3.org/2002/07/owl#differentFrom");
//	        	Predicate badgroup = KAON2Manager.factory().owlClass("http://www.aifb.uni-karlsruhe.de/WBS/dvr/rove/rove#BadGroupByNationality");
//	        	bodyargs0[0] = g;
//	        	bodyargs0[1] = s;
//	        	body[0] = KAON2Manager.factory().literal(true, hasMember, bodyargs0);
//	        	bodyargs1[0] = g;
//	        	bodyargs1[1] = x;
//	        	body[1] = KAON2Manager.factory().literal(true, hasMember, bodyargs1);
//	        	bodyargs2[0] = s;
//	        	bodyargs2[1] = n;
//	        	body[2] = KAON2Manager.factory().literal(true, nationality, bodyargs2);
//	        	bodyargs3[0] = x;
//	        	bodyargs3[1] = n;
//	        	body[3] = KAON2Manager.factory().literal(true, nationality, bodyargs3);
//	        	bodyargs4[0] = s;
//	        	bodyargs4[1] = x;
//	        	headarg[0] = g;
//	        	body[4] = KAON2Manager.factory().literal(true, differentFrom, bodyargs4);
//	        	Literal head = KAON2Manager.factory().literal(true, badgroup, headarg);
//	        	Rule rule = KAON2Manager.factory().rule(head, body);
//	        	
//	      		List<OntologyChangeEvent> changes=new ArrayList<OntologyChangeEvent>();
//	        	changes.add(new OntologyChangeEvent(rule, OntologyChangeEvent.ChangeType.ADD));
//	        	ontology.applyChanges(changes);
	        	
	        	//if (m_result)
//	        		ontology.saveOntology(m_format, printer.writer, "UTF-8");

	        } catch (KAON2Exception e) {
				printer.error(e.toString());
			} catch (IOException io) {
				printer.error(io.toString());
			} catch (InterruptedException i) {
				printer.error(i.toString());				
			}
			finally {
				endConnection();
			}
		} catch (KAON2Exception e) {
			printer.error(e.toString());
		}
	}
	
	/**
	 * prints the command line parameters and a short info
	 */
	public static void printUsage() {
		System.out.println("Usage: owl x <input_url> [<output_filename>] [-verb] [-rdf|-xml] [-noresult]");
		System.out.println("");
		System.out.println("That's for scripting and internal usage. Effect of this command may change");
		System.out.println("without notice. This is for testing for me, really.");
	}
}
