package edu.unika.aifb.owltools.tools;

import org.semanticweb.kaon2.api.*;
import org.semanticweb.kaon2.api.formatting.OntologyFileFormat;
import org.semanticweb.kaon2.api.owl.elements.Individual;
import org.semanticweb.kaon2.api.reasoner.Query;
import org.semanticweb.kaon2.api.reasoner.Reasoner;
import org.semanticweb.kaon2.api.logic.Variable;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.Stack;
import java.util.Properties;
import java.util.Set;

/**
 * @author Vetal
 *         Date: 26.04.2006
 *         Time: 19:48:28
 *         Translating  between two ontologies
 */

public class Translate extends KAON2Tool {

    String fileUriSource = "";
    String fileUriDestination = "";
    String fileUriMapping = "";
    String data4Translation = "";
    String m_output = "";
    String m_resolver = "resolver.xml";
    String m_format = OntologyFileFormat.OWL_XML;

    private static KAON2Connection connection;

    boolean m_result = true;

    public static void main(String[] args) {

        Translate t = new Translate();
        if (!t.processArguments(args))          //if command arguments are wrong - exit
            return;

        t.compile();
        t.printer.close();
    }

    public String name() {
        return "translate";
    }

    /**
     * Processing command line arguments
     * args[] : baseOntology destinationOntology mappingOntology dataForTranslation resultFile
     *
     * @return true (everything is ok) | false (printUsage)
     */
    public boolean processArguments(String[] args) {

        int argc = args.length;
        if (argc == 0) {
            printUsage();
            return false;
        }
        String s = args[0].toLowerCase();
        if (argc == 1) {
            if ((s.equalsIgnoreCase("-?")) || (s.equalsIgnoreCase("?")) ||
                    (s.equalsIgnoreCase("-h")) || (s.equalsIgnoreCase("--help")) ||
                    (s.equalsIgnoreCase("-help")) || (s.equalsIgnoreCase("help")) || (s.equalsIgnoreCase("translate")))
            {
                printUsage();
                return false;
            }
        }

        boolean hasFileName1 = false;
        boolean hasFileName2 = false;
        boolean hasFileName3 = false;
        boolean hasTranslateData = false;
        boolean hasOutput = false;
        boolean hasVerbose = false;
        boolean hasFormat = false;
        boolean hasResult = false;
        boolean hasResolver = false;
        boolean hasR = false;

        for (int i = 0; i < argc; ++i) {
            boolean unknownArg = true;

            String arg = args[i];
            if ((arg.equalsIgnoreCase("-?")) || (arg.equalsIgnoreCase("?")) || (arg.equalsIgnoreCase("-h")) || (arg.equalsIgnoreCase("--help")) || (arg.equalsIgnoreCase("-help")) || (arg.equalsIgnoreCase("help")))
            {
                printUsage();
                return false;
            }
            if (arg.startsWith("-")) {
                if ((arg.equalsIgnoreCase("-verbose")) || (arg.equalsIgnoreCase("-verb"))) {
                    if (hasVerbose) {
                        printUsageError();
                        return false;
                    }
                    printer.verbose = true;
                    hasVerbose = true;
                    unknownArg = false;
                }
                if ((arg.equalsIgnoreCase("-noresult")) || (arg.equalsIgnoreCase("-no"))) {
                    if (hasResult) {
                        printUsageError();
                        return false;
                    }
                    m_result = false;
                    hasResult = true;
                    unknownArg = false;
                }
                if (arg.equalsIgnoreCase("-rdf")) {
                    if (hasFormat) {
                        printUsageError();
                        return false;
                    }
                    m_format = OntologyFileFormat.OWL_RDF;
                    hasFormat = true;
                    unknownArg = false;
                }
                if (arg.equalsIgnoreCase("-xml")) {
                    if (hasFormat) {
                        printUsageError();
                        return false;
                    }
                    m_format = OntologyFileFormat.OWL_XML;
                    hasFormat = true;
                    unknownArg = false;
                }
                if (arg.equalsIgnoreCase("-r")) {
                    if (hasResolver) {
                        printUsageError();
                        return false;
                    }
                    m_resolver = "resolver.xml";
                    hasR = true;

                    unknownArg = false;
                }
            } else {
                if (arg.equalsIgnoreCase("translate")) {
                    unknownArg = false;
                } else {
                    if (hasFileName1) {
                        if (hasFileName2) {
                            if (hasFileName3) {
                                if (hasTranslateData) {
                                    if (hasOutput) {
                                        if (hasR) {
                                            if (hasResolver) {
                                                printUsageError();
                                                return false;
                                            }
                                            m_resolver = arg;
                                            hasResolver = true;
                                            unknownArg = false;

                                        }


                                    }
                                    if (!hasR) {
                                        System.out.println("result ontology:" + arg);
                                        m_output = arg;
                                        hasOutput = true;
                                        unknownArg = false;
                                    } else {
                                        System.out.println("resolver file:" + arg);
                                        m_resolver = arg;
                                        unknownArg = false;
                                    }
                                } else {
                                    data4Translation = arg;
                                    System.out.println("data for translation:" + arg);

                                    hasTranslateData = true;
                                    unknownArg = false;
                                }
                            } else {
                                fileUriMapping = arg;
                                System.out.println("mapping ontology:" + arg);

                                hasFileName3 = true;
                                unknownArg = false;
                            }
                        } else {
                            fileUriDestination = arg;
                            System.out.println("destination ontology:" + arg);

                            hasFileName2 = true;
                            unknownArg = false;
                        }
                    } else {
                        fileUriSource = arg;
                        System.out.println("source ontology:" + arg);

                        hasFileName1 = true;
                        unknownArg = false;
                    }
                }
            }
            if (unknownArg) {
                printUsageError();
                return false;
            }
        }
        if (!hasFileName2 || !hasFileName3 || !hasTranslateData) {
            printUsageError();
            return false;
        }
        return true;
    }

    /**
     * Setting KAON2Connection, loading ontologies from file(resolver) and opening ontologies for working.
     * This method uses translation method in order to create the resulting ontology
     */
    public void compile() {
        try {
            if (!m_output.equals("")) printer.setFile(m_output);

            try {

                connection = KAON2Manager.newConnection();
                DefaultOntologyResolver resolver = new DefaultOntologyResolver();

                //Properties props = new Properties();
                //ClassLoader loader = Translate.class.getClassLoader();
                //props.load(loader.getResourceAsStream("props.properties"));

                //String resolverFile = (String) props.get("resolver");
                String resolverFile = m_resolver;

                //System.out.println(resolverFile);
                File f = new File(resolverFile);
                if (f.exists()) {
                    resolver.loadFromProperties(f);
                }
                connection.setOntologyResolver(resolver);

                resolver.registerOntology("file:" + fileUriSource);
                resolver.registerReplacement(fileUriSource, "file:" + fileUriSource);
                System.out.print("Opening ontology " + fileUriSource + " ...");
                Ontology base = connection.openOntology(fileUriSource, new HashMap<String, Object>());
                System.out.println("Done");

                resolver.registerOntology("file:" + fileUriDestination);
                resolver.registerReplacement(fileUriDestination, "file:" + fileUriDestination);
                System.out.print("Opening ontology " + fileUriDestination + " ...");
                Ontology diff = connection.openOntology(fileUriDestination, new HashMap<String, Object>());
                System.out.println("Done");

                resolver.registerOntology("file:" + fileUriMapping);
                resolver.registerReplacement(fileUriMapping, "file:" + fileUriMapping);
                System.out.print("Opening ontology " + fileUriMapping + " ...");
                Ontology mapp = connection.openOntology(fileUriMapping, new HashMap<String, Object>());
                System.out.println("Done");

                resolver.registerOntology("file:" + data4Translation);
                resolver.registerReplacement(data4Translation, "file:" + data4Translation);
                System.out.print("Opening ontology " + data4Translation + " ...");
                Ontology translate = connection.openOntology(data4Translation, new HashMap<String, Object>());
                System.out.println("Done");

                Ontology result = translation(base, diff, mapp, translate, resolver);

                if (m_result && result != null)
                    result.saveOntology(m_format, printer.writer, "UTF-8");

            } catch (KAON2Exception e) {
                printer.error("Could not open ontology.");
                e.printStackTrace();
            } catch (IOException io) {
                printer.error("Could not save resulting ontology " + m_output);
                printer.error("Probably file for the resolver registration needed ");
                io.printStackTrace();
            } catch (InterruptedException i) {
                printer.error("Could not save resulting ontology.");
                printer.error(i.toString());
            }
            finally {
                endConnection();
            }
        } catch (KAON2Exception e) {
            printer.error("Could not open or close KAON2 connection.");
            e.printStackTrace();
        }
    }

    /**
     * Main logic of programm. Creating temporary ontology und stack with resulting queries
     *
     * @return resulting ontology
     */
    public Ontology translation(Ontology base, Ontology diff, Ontology mapp, Ontology translation, DefaultOntologyResolver resolver) throws KAON2Exception {

        resolver.registerReplacement("http://kaon2.semanticweb.org/temp", "file:temp.owl");
        Ontology tmp = connection.createOntology("http://kaon2.semanticweb.org/temp", new HashMap<String, Object>());

        try {
            connection.openOntology("http://kaon2.semanticweb.org/temp", new HashMap<String, Object>());

            tmp.addToImports(base);
            tmp.addToImports(diff);
            tmp.addToImports(mapp);
            tmp.addToImports(translation);

        } catch (InterruptedException ex) {
            ex.printStackTrace();
        }

        Reasoner reasoner = tmp.createReasoner();

        resolver.registerReplacement("http://www.aifb.uni-karlsruhe.de/WBS/dvr/owltools/translate/result", "file:result.rdf");
        Ontology result = connection.createOntology("http://www.aifb.uni-karlsruhe.de/WBS/dvr/owltools/translate/result", new HashMap<String, Object>());

        result.addToImports(tmp);

        Stack v = createQuery(base, mapp);

        try {
            if (!m_output.equals(""))
                System.out.print("Writing result to the file ...");

            String value = "";
            while (v.size() != 0) {
                value = (String) v.pop();
                String tempQuery = "construct {?x rdf:type <" + value + "> }\n" +
                        "where\n" +
                        "{ ?x rdf:type <" + value + "> }";

                Query query = reasoner.createQuery(Namespaces.INSTANCE, tempQuery);
                //Query query = reasoner.createQuerySPARQL(Namespaces.INSTANCE, tempQuery);
                if (!m_output.equals("")) {
                    query.executeStoreResults(result);

                    query.dispose();
                } else {
                    query.open();
                    Object[] tupleBuffer = query.tupleBuffer();
                    while (!query.afterLast()) {
                        System.out.print("[ ");
                        for (int i = 0; i < tupleBuffer.length; i++) {
                            if (i != 0)
                                System.out.print(", ");
                            System.out.println("[ " + value + " ]");
                            System.out.print(tupleBuffer[i]);
                        }
                        System.out.println(" ]");
                        query.next();
                    }
                    System.out.println(" ");
                    query.dispose();
                }
            }
            reasoner.dispose();

            if (!m_output.equals(""))
                System.out.println("Done");

            result.removeFromImports(tmp);
        } catch (InterruptedException ex) {
            ex.printStackTrace();
        }

        connection.close();
        if (!m_output.equals(""))
            return result;
        else
            return null;
    }

    /**
     * Creating stack with axioms
     *
     * @return stack with necessary axioms
     */
    public static Stack createQuery(Ontology base, Ontology map) {

        Stack v = new Stack();
        try {

            Cursor<Axiom> cursorMap = map.createAxiomRequest().openCursor();

            StringBuffer axiomBuf = new StringBuffer();
            StringBuffer buf = new StringBuffer();

            while (cursorMap.hasNext()) {

                Axiom mapAxiom = cursorMap.next();

                mapAxiom.toString(axiomBuf, Namespaces.INSTANCE);

                String axioms = axiomBuf.substring(axiomBuf.indexOf(" ") + 1, axiomBuf.lastIndexOf("]"));
                String firstElement = axioms.substring(axioms.indexOf("http://"), axioms.indexOf(" "));
                String lastElement = axioms.substring(axioms.indexOf(" ") + 1);

                System.out.println("Axiom: " + firstElement);
                System.out.println("Equivalent: " + lastElement);

                axiomBuf.setLength(0);
                Cursor<Axiom> cursorBase = base.createAxiomRequest().openCursor();

                while (cursorBase.hasNext()) {
                    buf.setLength(0);

                    Axiom baseAxiom = cursorBase.next();
                    baseAxiom.toString(buf, Namespaces.INSTANCE);

                    if (buf.lastIndexOf(firstElement) > 0) {
                        v.push(firstElement);
                        break;
                    } else {
                        if (buf.lastIndexOf(lastElement) > 0) {
                            v.push(lastElement);
                            break;
                        }
                    }
                }
            }
            buf.setLength(0);
            cursorMap.close();
        } catch (KAON2Exception kex) {
            kex.printStackTrace();
        }
        return v;
    }

    /**
     * Printing usage if command line arguments aren't correct
     */
    public static void printUsage() {
        System.out.println("Usage: owl translate <url1> <url2> <url3> <url4> [<output_filename>] [-verbose] [-rdf|-xml] [-no]");
        System.out.println("");
        System.out.println("translate creates one ontology with the translation for two given ontologies");
        System.out.println("");
        System.out.println(" <url1>");
        System.out.println("     The location of the source ontology");
        System.out.println(" <url2>");
        System.out.println("     The location of the destination ontology");
        System.out.println(" <url3>");
        System.out.println("     The location of the mapping ontology");
        System.out.println("  <url4>");
        System.out.println("     The location of the date for the translation");
        System.out.println(" <output_filename>");
        System.out.println("     The name of the file where the merged ontology will be written to.");
        System.out.println("     If omitted, the result is sent to the console.");
        System.out.println("");
        System.out.println(" -verbose or simply -verb");
        System.out.println("     If this switch is set, you will get additional output while processing.");
        System.out.println("");
        System.out.println(" -rdf / -xml");
        System.out.println("     These switches decide the format of the output ontology: either");
        System.out.println("     RDF/XML-serialisation or OWL/XML Presentation Syntax (default)");
        System.out.println("");
        System.out.println(" -noresult or simply -no");
        System.out.println("     Doesn't save or show the result.");
        System.out.println("");
        System.out.println(" User must have file for the resolver registration, which consists all ontologies and uris ");
        System.out.println("     Example");
        System.out.println("        <?xml version=\"1.0\" encoding=\"UTF-8\"?>");
        System.out.println("        <!DOCTYPE properties SYSTEM \"http://java.sun.com/dtd/properties.dtd\">");
        System.out.println("                                                                                 ");
        System.out.println("        <properties>                                                             ");
        System.out.println("            <entry key=\"http://swrc.ontoware.org/ontology\">file:/C:/projects/owltranslate/owltools/owltools-src-2005-11-22/owltools/src/swrc.owl</entry>       ");
        System.out.println("            <entry key=\"http://proton.semanticweb.org/2005/04/protont2\">file:/C:/projects/owltranslate/owltools/owltools-src-2005-11-22/owltools/src/protont2.rdf</entry>  ");
        System.out.println("            <entry key=\"http://my.mapping.org/bt_swrc\">file:/C:/projects/owltranslate/owltools/owltools-src-2005-11-22/owltools/src/mapping.owl</entry>              ");
        System.out.println("            <entry key=\"http://my.aifb.uni-karlsruhe.de/viewAIFB_OWL.owl\">file:/C:/projects/owltranslate/owltools/owltools-src-2005-11-22/owltools/src/viewAIFB_OWL.xml</entry>   ");
        System.out.println("            <entry key=\"http://proton.semanticweb.org/2005/04/protons\">file:/C:/projects/owltranslate/owltools/owltools-src-2005-11-22/owltools/src/protons.rdf</entry>   ");
        System.out.println("            <entry key=\"http://proton.semanticweb.org/2005/04/protonu\">file:/C:/projects/owltranslate/owltools/owltools-src-2005-11-22/owltools/src/protonu.rdf</entry> ");
        System.out.println("            <entry key=\"http://proton.semanticweb.org/2005/04/protont\">file:/C:/projects/owltranslate/owltools/owltools-src-2005-11-22/owltools/src/protont.rdf</entry>  ");
        System.out.println("                                                                                                                                                                   ");
        System.out.println("        </properties>                                                                                                                                              ");
    }

    /**
     * WebService
     * This method is used from JSP-Page (Web version of tool)
     */
    public static String Web(String _fileUriSource, String _fileUriDestination, String _fileUriMapping, String _data4Translation, String _resolverStr) {
        String resultat = "res";

        try {
            if (connection == null)
                connection = KAON2Manager.newConnection();
            DefaultOntologyResolver resolver = new DefaultOntologyResolver();

            File f = new File(_resolverStr);
            resolver.loadFromProperties(f);
            connection.setOntologyResolver(resolver);

            resolver.registerOntology(_fileUriSource);
            resolver.registerReplacement(_fileUriSource, _fileUriSource);

            connection.setOntologyResolver(resolver);

            Ontology base = connection.openOntology(_fileUriSource, new HashMap<String, Object>());

            resolver.registerOntology(_fileUriDestination);
            resolver.registerReplacement(_fileUriDestination, _fileUriDestination);
            Ontology diff = connection.openOntology(_fileUriDestination, new HashMap<String, Object>());

            resolver.registerOntology(_fileUriMapping);
            resolver.registerReplacement(_fileUriMapping, _fileUriMapping);
            Ontology mapp = connection.openOntology(_fileUriMapping, new HashMap<String, Object>());

            resolver.registerOntology(_data4Translation);
            resolver.registerReplacement(_data4Translation, _data4Translation);
            Ontology translate = connection.openOntology(_data4Translation, new HashMap<String, Object>());

            StringBuffer resStr = new StringBuffer();
            String tempQuery = "";

            resolver.registerReplacement("http://kaon2.semanticweb.org/temp", "file:temp.owl");
            Ontology tmp = connection.createOntology("http://kaon2.semanticweb.org/temp", new HashMap<String, Object>());
            try {
                connection.openOntology("http://kaon2.semanticweb.org/temp", new HashMap<String, Object>());

                tmp.addToImports(base);
                tmp.addToImports(diff);
                tmp.addToImports(mapp);
                tmp.addToImports(translate);

                Reasoner reasoner = tmp.createReasoner();

                resolver.registerReplacement("http://www.aifb.uni-karlsruhe.de/WBS/dvr/owltools/translate/result", "file:result.owl");
                Ontology result = connection.createOntology("http://www.aifb.uni-karlsruhe.de/WBS/dvr/owltools/translate/result", new HashMap<String, Object>());

                result.addToImports(tmp);

                Stack v = createQuery(base, mapp);

                String value = "";
                boolean ff = false;
                Query query;

                while (v.size() != 0) {
                    if (!ff) {

                        resStr.append("<table align=\"center\" border=\"1\"> <tr> <th>Translation</th></tr>");

                        ff = true;
                    }
                    value = (String) v.pop();
                    tempQuery =
                            "construct {?x rdf:type <" + value + "> }\n" +
                                    "where\n" +
                                    "{ ?x rdf:type <" + value + "> }";

                    //query = reasoner.createQuerySPARQL(Namespaces.INSTANCE, tempQuery);
                    query = reasoner.createQuery(Namespaces.INSTANCE, tempQuery);
                    query.open();

                    while (!query.afterLast()) {
                        resStr.append("<tr><td align=\"left\">" + value + "</td><td align=\"left\">");
                        Object[] res = query.tupleBuffer().clone();
                        for (int i = 0; i < res.length; i++) {
                            resStr.append(encode(res[i].toString()));
                            resStr.append("</td></tr>");
                        }
                        query.next();
                    }
                    query.close();

                }
                resStr.append("</table>");
                reasoner.dispose();
            } catch (InterruptedException ex) {
                ex.printStackTrace();
            }
            resultat = resStr.toString();

            connection.close();
            return resultat;

        } catch (KAON2Exception e) {
            e.printStackTrace();
        } catch (InterruptedException i) {
            i.printStackTrace();
        } catch (Exception ex) {
            ex.printStackTrace();
        }

        return resultat;
    }

    /**
     * @param input
     * @return an XML encoded string
     */
    public static String encode(String input) {

        if (input == null) {
            return "";
        }

        String data = input;

        data = data.replaceAll("&", "&amp;");
        data = data.replaceAll("<", "&lt;");
        data = data.replaceAll(">", "&gt;");
        data = data.replaceAll("\"", "&quot;");
        data = data.replaceAll("'", "&apos;");

        return data;
    }
}