/*
 * Created on 24.11.2004
 *
 * This package is just an utility package for the output.
 */
package edu.unika.aifb.owltools.utils;

import java.io.*;
import java.nio.charset.*;

/**
 * Printer offers static sinks for different kind of messages. Using this
 * in a central place we can easily send the different kinds of output to
 * different places.
 */
public class Printer {

	/**
	 * Should verbose output be shown or not?
	 */
	public boolean verbose = false;
	
	/**
	 * Should warnings be shown or not?
	 */
	public boolean warning = true;
	
	/**
	 * Saves the writer for the printer
	 */
	public PrintWriter writer = new PrintWriter(System.out);
	
	/**
	 * Is the output to be HTML-escaped?
	 */
	public boolean html =false;
	
	/**
	 * Place to save error output (for example in HTML'ed output)
	 */
	public String error = "";
	
	/**
	 * If the output needs to be saved in an internal string, this is the place
	 */
	private StringWriter output = new StringWriter();
	
	/**
	 * The output will be saved to a file
	 * @param filename The name of the file where the output is to be saved
	 */
	public void setFile(String filename) {
		try {
			writer = new PrintWriter(new OutputStreamWriter(new FileOutputStream(new File(filename)), Charset.forName("UTF-8")));
		} catch (IOException e) {
			error("Could not open file " + filename);
			error(e.toString());
		}
	}

	/**
	 * the output needs to be saved internally
	 *
	 */
	public void setString() {
		writer = new PrintWriter(output);
	}
	
	/**
	 * get the internally saved output
	 * @return The saved output
	 */
	public String getOutput() {
		writer.flush();
		return output.toString();
	}
	
	/**
	 * closes the writer. If the output was sent to a file, the files is saved now.
	 * Otherwise, the output is lost
	 *
	 */
	public void close() {
		writer.close();
	}
	
	/**
	 * Used for usual messages, that actually are not part of the output.
	 * Is always sent to the System.out
	 * 
	 * @param m the message
	 */
	public void print(Object m) {
		System.out.println(m);
	}
	
	/**
	 * Used for detailed messages. Sent to System.out, if verbose is switched on
	 * 
	 * @param m the message
	 */
	public void verbose(String m) {
		if (verbose) print(m);
	}

	/**
	 * Used for detailed messages. Sent to System.out, if verbose is switched on
	 * 
	 * @param m the message
	 */
	public void verbose(int m) {
		if (verbose) print(m);
	}

	/**
	 * Used for warnings. Sent to System.out, if warning is switched on
	 * 
	 * @param m the message
	 */
	public void warning(String m) {
		if (warning) print(m);
	}

	/**
	 * Used for warnings. Sent to System.out, if warning is switched on
	 * 
	 * @param m the message
	 */
	public void warning(int m) {
		if (warning) print(m);
	}

	/**
	 * Used for development messages
	 * 
	 * @param m the message
	 */
	public void debug(String m) {
		if (edu.unika.aifb.owltools.OWLTools.developmentVersion()) print(m);
	}

	/**
	 * Used for detailed messages. Sent to System.out, if verbose is switched on
	 * 
	 * @param m the message
	 */
	public void debug(int m) {
		if (edu.unika.aifb.owltools.OWLTools.developmentVersion()) print(m);
	}

	/**
	 * Used for the output of the program (ending the output with a new line).
	 * 
	 * @param m the message
	 */
	public void putline(Object m) {
		put(m);
		if (html) put("<br />");
		writer.println();
	}

	/**
	 * Used for the output of the program.
	 * 
	 * @param m the message
	 */
	public void put(Object m) {
		writer.print(m);
	}

	/**
	 * Used for error messages.
	 * 
	 * @param m the message
	 */
	public void error(String m) {
		if (html)
			error += m + "<br \\>";
		else
			System.err.println(m);
	}

}
