package edu.unika.aifb.owltools.utils;

import java.util.*;

import org.semanticweb.kaon2.api.logic.*;


/**
 * Compare two rules, ordering them lexicographically according to their
 * serialization.
 */
public class RuleComparator implements Comparator<Rule> {

	public int compare(Rule r1, Rule r2) {
		int hl1 = r1.getHeadLength();
		int hl2 = r2.getHeadLength();
		if (hl1 > hl2) return 1;
		if (hl1 < hl2) return -1;
		// head lengths equal
		if (hl1 == 0) return 0;
		String head1 = hl1 + ")"; // sort via head length
		String head2 = hl2 + ")";
		for (int i=0; i<hl1; i++) { // sort via head predicates
			head1 += r1.getHeadLiteral(i).getPredicate().toString() + "(" + r1.getHeadLiteral(i).getArity();
			head2 += r2.getHeadLiteral(i).getPredicate().toString() + "(" + r2.getHeadLiteral(i).getArity();
		}
		head1 += "(" + r1.getBodyLength(); // sort via body length (facts first)
		head2 += "(" + r2.getBodyLength();
		return head1.compareTo(head2);
	}

	public boolean equals(Object obj) {
		return (obj instanceof RuleComparator);
	}
	
}